/* $Id: isaac_pfits.c,v 1.19 2013-03-12 08:06:48 llundin Exp $
 *
 * This file is part of the ISAAC Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2013-03-12 08:06:48 $
 * $Revision: 1.19 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
   								Includes
 -----------------------------------------------------------------------------*/

#include <string.h>
#include <cpl.h>

#include "isaac_pfits.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup isaac_pfits     FITS header protected access
 *
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*-----------------------------------------------------------------------------
  							Function codes
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    find out airmass start
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
double isaac_pfits_get_airmass_start(const cpl_propertylist * plist)
{
    const double val = cpl_propertylist_get_double(plist, "ESO TEL AIRM START");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out airmass end
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
double isaac_pfits_get_airmass_end(const cpl_propertylist * plist)
{
    const double val = cpl_propertylist_get_double(plist, "ESO TEL AIRM END");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the arcfile   
  @param    plist       property list to read from
  @return   Pointer to string in the propertylist, or NULL on error
 */
/*----------------------------------------------------------------------------*/
const char * isaac_pfits_get_arcfile(const cpl_propertylist * plist)
{
    const char * val = cpl_propertylist_get_string(plist, "ARCFILE");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the arm which is active
  @param    plist       property list to read from
  @return   Pointer to string in the propertylist, or NULL on error
 */
/*----------------------------------------------------------------------------*/
const char * isaac_pfits_get_arm(const cpl_propertylist * plist)
{
    const char * val = cpl_propertylist_get_string(plist, "ESO OCS SELECT-ARM");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/** 
  @brief    find out the shutter status
  @param    plist       property list to read from
  @return   Pointer to string in the propertylist, or NULL on error
 */
/*----------------------------------------------------------------------------*/
int isaac_pfits_get_calshut_status(const cpl_propertylist * plist)
{
    const int val = cpl_propertylist_get_bool(plist, "ESO INS CALSHUT ST");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the chip name
  @param    plist       property list to read from
  @return   Pointer to string in the propertylist, or NULL on error
 */
/*----------------------------------------------------------------------------*/
const char * isaac_pfits_get_chip(const cpl_propertylist * plist)
{
    const char * val = cpl_propertylist_get_string(plist, "ESO DET CHIP NAME");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the number of chopping cycles
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int isaac_pfits_get_chopping_cycle(const cpl_propertylist * plist)
{
    const int val = cpl_propertylist_get_int(plist, "ESO DET CHOP NCYCLES");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the chopping frequency
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
double isaac_pfits_get_chopping_frequency(const cpl_propertylist * plist)
{
    const double val = cpl_propertylist_get_double(plist, "ESO DET CHOP FREQ");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the chopping status
  @param    plist       property list to read from
  @return   Pointer to string in the propertylist, or NULL on error
 */
/*----------------------------------------------------------------------------*/
int isaac_pfits_get_chopping_status(const cpl_propertylist * plist)
{
    const int val = cpl_propertylist_get_bool(plist, "ESO TEL CHOP ST");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the cumulative offset in X
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
double isaac_pfits_get_cumoffsetx(const cpl_propertylist * plist)
{
    const double val = cpl_propertylist_get_double(plist, "ESO SEQ CUMOFFSETX");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the cumulative offset in Y
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
double isaac_pfits_get_cumoffsety(const cpl_propertylist * plist)
{
    const double val = cpl_propertylist_get_double(plist, "ESO SEQ CUMOFFSETY");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the current exposure number
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int isaac_pfits_get_current_exp_nb(const cpl_propertylist * plist)
{
    const int val = cpl_propertylist_get_int(plist, "ESO TPL EXPNO");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the date of the file creation
  @param    plist       property list to read from
  @return   Pointer to string in the propertylist, or NULL on error
 */
/*----------------------------------------------------------------------------*/
const char * isaac_pfits_get_date(const cpl_propertylist * plist)
{
    const char * val = cpl_propertylist_get_string(plist, "DATE");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the date of observation 
  @param    plist       property list to read from
  @return   Pointer to string in the propertylist, or NULL on error
 */
/*----------------------------------------------------------------------------*/
const char * isaac_pfits_get_date_obs(const cpl_propertylist * plist)
{
    const char * val = cpl_propertylist_get_string(plist, "DATE-OBS");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the DEC
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
double isaac_pfits_get_dec(const cpl_propertylist * plist)
{
    const double val = cpl_propertylist_get_double(plist, "DEC");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the frame type
  @param    plist       property list to read from
  @return   Pointer to string in the propertylist, or NULL on error
 */
/*----------------------------------------------------------------------------*/
const char * isaac_pfits_get_frame_type(const cpl_propertylist * plist)
{
    /* DET.FRAME moved to DET.FRAM in DICB on 28-07-2001 */
    const char * val
        = cpl_propertylist_has(plist, "ESO DET FRAM TYPE")
        ? cpl_propertylist_get_string(plist, "ESO DET FRAM TYPE")
        : cpl_propertylist_get_string(plist, "ESO DET FRAME TYPE");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the DIT value 
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
double isaac_pfits_get_dit(const cpl_propertylist * plist)
{
    const double val = cpl_propertylist_get_double(plist, "ESO DET DIT");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the data category
  @param    plist       property list to read from
  @return   Pointer to string in the propertylist, or NULL on error
 */
/*----------------------------------------------------------------------------*/
const char * isaac_pfits_get_dpr_catg(const cpl_propertylist * plist)
{
    const char * val = cpl_propertylist_get_string(plist, "ESO DPR CATG");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the data technique
  @param    plist       property list to read from
  @return   Pointer to string in the propertylist, or NULL on error
 */
/*----------------------------------------------------------------------------*/
const char * isaac_pfits_get_dpr_tech(const cpl_propertylist * plist)
{
    const char * val = cpl_propertylist_get_string(plist, "ESO DPR TECH");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/** 
  @brief    find out the data type
  @param    plist       property list to read from
  @return   Pointer to string in the propertylist, or NULL on error
 */
/*----------------------------------------------------------------------------*/
const char * isaac_pfits_get_dpr_type(const cpl_propertylist * plist)
{
    const char * val = cpl_propertylist_get_string(plist, "ESO DPR TYPE");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/** 
  @brief    find out which wave band is active in long wavelength
  @param    plist       property list to read from
  @return   Pointer to string in the propertylist, or NULL on error
 */
/*----------------------------------------------------------------------------*/
const char * isaac_pfits_get_filter_lw(const cpl_propertylist * plist)
{
    cpl_errorstate prestate = cpl_errorstate_get();
    const char * val = cpl_propertylist_get_string(plist, "ESO INS FILT3 ID");

    if (val == NULL || !strcmp(val, "open")) {
        /* FILT3 is missing or open, try FILT4 */
        val = cpl_propertylist_get_string(plist, "ESO INS FILT4 ID");
        if (val == NULL) {
            (void)cpl_error_set(cpl_func, cpl_error_get_code());
        } else {
            cpl_errorstate_set(prestate);
            if (!strcmp(val, "open")) val = NULL;
        }
    }

    return val;
}

/*----------------------------------------------------------------------------*/
/** 
  @brief    find out which wave band is active in short wavelength
  @param    plist       property list to read from
  @return   Pointer to string in the propertylist, or NULL on error
 */
/*----------------------------------------------------------------------------*/
const char * isaac_pfits_get_filter_sw(const cpl_propertylist * plist)
{
    cpl_errorstate prestate = cpl_errorstate_get();
    const char * val = cpl_propertylist_get_string(plist, "ESO INS FILT1 ID");

    if (val == NULL || !strcmp(val, "open")) {
        /* FILT1 is missing or open, try FILT2 */
        val = cpl_propertylist_get_string(plist, "ESO INS FILT2 ID");
        if (val == NULL) {
            (void)cpl_error_set(cpl_func, cpl_error_get_code());
        } else {
            cpl_errorstate_set(prestate);
            if (!strcmp(val, "open")) val = NULL;
        }
    }

    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the humidity level
  @param    plist       property list to read from
  @return   the requested value 
 */
/*----------------------------------------------------------------------------*/
double isaac_pfits_get_humidity_level(const cpl_propertylist * plist)
{
    const double val = cpl_propertylist_get_double(plist, "ESO TEL AMBI RHUM");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/** 
  @brief    find out the instrument
  @param    plist       property list to read from
  @return   Pointer to string in the propertylist, or NULL on error
 */
/*----------------------------------------------------------------------------*/
const char * isaac_pfits_get_instrument(const cpl_propertylist * plist)
{
    const char * val = cpl_propertylist_get_string(plist, "INSTRUME");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/** 
  @brief    find out the lamp1 name
  @param    plist       property list to read from
  @return   Pointer to string in the propertylist, or NULL on error
 */
/*----------------------------------------------------------------------------*/
const char * isaac_pfits_get_lamp1_name(const cpl_propertylist * plist)
{
    const char * val = cpl_propertylist_get_string(plist, "ESO INS LAMP1 NAME");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/** 
  @brief    find out the lamp1 status
  @param    plist       property list to read from
  @return   Pointer to string in the propertylist, or NULL on error
 */
/*----------------------------------------------------------------------------*/
int isaac_pfits_get_lamp1_status(const cpl_propertylist * plist)
{
    const int val = cpl_propertylist_get_bool(plist, "ESO INS LAMP1 ST");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/** 
  @brief    find out the lamp2 name
  @param    plist       property list to read from
  @return   Pointer to string in the propertylist, or NULL on error
 */
/*----------------------------------------------------------------------------*/
const char * isaac_pfits_get_lamp2_name(const cpl_propertylist * plist)
{
    const char * val = cpl_propertylist_get_string(plist, "ESO INS LAMP2 NAME");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/** 
  @brief    find out the lamp2 status
  @param    plist       property list to read from
  @return   Pointer to string in the propertylist, or NULL on error
 */
/*----------------------------------------------------------------------------*/
int isaac_pfits_get_lamp2_status(const cpl_propertylist * plist)
{
    const int val = cpl_propertylist_get_bool(plist, "ESO INS LAMP2 ST");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the lamp3 intensity
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int isaac_pfits_get_lamp3_intensity(const cpl_propertylist * plist)
{
    const int val = cpl_propertylist_get_int(plist, "ESO INS LAMP3 SET");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the MJD-OBS
  @param    plist       property list to read from
  @return   the requested value 
 */
/*----------------------------------------------------------------------------*/
double isaac_pfits_get_mjdobs(const cpl_propertylist * plist)
{
    const double val = cpl_propertylist_get_double(plist, "MJD-OBS");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/** 
  @brief    find out the instrument mode
  @param    plist       property list to read from
  @return   Pointer to string in the propertylist, or NULL on error
 */
/*----------------------------------------------------------------------------*/
const char * isaac_pfits_get_mode(const cpl_propertylist * plist)
{
    const char * val = cpl_propertylist_get_string(plist, "ESO INS MODE");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the WIN STARTX keyword
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
double isaac_pfits_get_win_startx(const cpl_propertylist * plist)
{
    const double val = cpl_propertylist_get_double(plist, "ESO DET WIN STARTX");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the WIN STARTY keyword
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
double isaac_pfits_get_win_starty(const cpl_propertylist * plist)
{
    const double val = cpl_propertylist_get_double(plist, "ESO DET WIN STARTY");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the NAXIS1 keyword
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int isaac_pfits_get_naxis1(const cpl_propertylist * plist)
{
    const int val = cpl_propertylist_get_int(plist, "NAXIS1");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the NDIT keyword
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int isaac_pfits_get_ndit(const cpl_propertylist * plist)
{
    const int val = cpl_propertylist_get_int(plist, "ESO DET NDIT");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the NDSAMPLES keyword
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int isaac_pfits_get_ndsamples(const cpl_propertylist * plist)
{
    const int val = cpl_propertylist_get_int(plist, "ESO DET NDSAMPLES");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the number of exposures
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int isaac_pfits_get_numbexp(const cpl_propertylist * plist)
{
    const int val = cpl_propertylist_get_int(plist, "ESO TPL NEXP");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/** 
  @brief    find out the objective
  @param    plist       property list to read from
  @return   Pointer to string in the propertylist, or NULL on error
 */
/*----------------------------------------------------------------------------*/
const char * isaac_pfits_get_objective(const cpl_propertylist * plist)
{
    const char * mode = isaac_pfits_get_mode(plist);
    const char * val = NULL;

    /* The keyword used to store the objective is different for LW and SW */
    if (mode == NULL) {
        (void)cpl_error_set(cpl_func, cpl_error_get_code());
    } else if (mode[0] == 'S') {
        val = cpl_propertylist_get_string(plist, "ESO INS OPTI2 NAME");
        (void)cpl_error_set(cpl_func, cpl_error_get_code());
    } else if (mode[0] == 'L') {
        val = cpl_propertylist_get_string(plist, "ESO INS OPTI3 NAME");
        (void)cpl_error_set(cpl_func, cpl_error_get_code());
    } else {
        (void)cpl_error_set(cpl_func, CPL_ERROR_UNSUPPORTED_MODE);
    }

    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the OBS ID
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int isaac_pfits_get_obs_id(const cpl_propertylist * plist)
{
    const int val = cpl_propertylist_get_int(plist, "ESO OBS ID");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/** 
  @brief    find out the target name
  @param    plist       property list to read from
  @return   Pointer to string in the propertylist, or NULL on error
 */
/*----------------------------------------------------------------------------*/
const char * isaac_pfits_get_targ_name(const cpl_propertylist * plist)
{
    const char * val = cpl_propertylist_get_string(plist, "ESO OBS TARG NAME");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the OPTI1.ID key
  @param    plist       property list to read from
  @return   Pointer to string in the propertylist, or NULL on error
 */
/*----------------------------------------------------------------------------*/
const char * isaac_pfits_get_opti1_id(const cpl_propertylist * plist)
{
    const char * val = cpl_propertylist_get_string(plist, "ESO INS OPTI1 ID");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the order
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int isaac_pfits_get_order(const cpl_propertylist * plist)
{
    const int val = cpl_propertylist_get_int(plist, "ESO INS GRAT ORDER");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the original file name
  @param    plist       property list to read from
  @return   Pointer to string in the propertylist, or NULL on error
 */
/*----------------------------------------------------------------------------*/
const char * isaac_pfits_get_origfile(const cpl_propertylist * plist)
{
    const char * val = cpl_propertylist_get_string(plist, "ORIGFILE");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the pixel scale
  @param    plist       property list to read from
  @return   the requested value 
 */
/*----------------------------------------------------------------------------*/
double isaac_pfits_get_pixscale(const cpl_propertylist * plist)
{
    const double val = cpl_propertylist_get_double(plist, "ESO INS PIXSCALE");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the RA
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
double isaac_pfits_get_ra(const cpl_propertylist * plist)
{
    const double val = cpl_propertylist_get_double(plist, "RA");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the resolution
  @param    plist       property list to read from
  @return   Pointer to string in the propertylist, or NULL on error
 */
/*----------------------------------------------------------------------------*/
const char * isaac_pfits_get_resolution(const cpl_propertylist * plist)
{
    const char * val = cpl_propertylist_get_string(plist, "ESO INS GRAT NAME");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the read out mode
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int isaac_pfits_get_rom(const cpl_propertylist * plist)
{
    const int val = cpl_propertylist_get_int(plist, "ESO DET NCORRS");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the read out mode name
  @param    plist       property list to read from
  @return   Pointer to string in the propertylist, or NULL on error
 */         
/*----------------------------------------------------------------------------*/
const char * isaac_pfits_get_romode_name(const cpl_propertylist * plist)
{
    const char * val = cpl_propertylist_get_string(plist, "ESO DET MODE NAME");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the read out mode name
  @param    plist       property list to read from
  @return   Pointer to string in the propertylist, or NULL on error
 */         
/*----------------------------------------------------------------------------*/
const char * isaac_pfits_get_rom_name(const cpl_propertylist * plist)
{
    const char * val = cpl_propertylist_get_string(plist, "ESO DET NCORRS NAME");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the read speed
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
int isaac_pfits_get_rspeed(const cpl_propertylist * plist)
{
    const int val = cpl_propertylist_get_int(plist, "ESO DET RSPEED");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the template ID
  @param    plist       property list to read from
  @return   Pointer to string in the propertylist, or NULL on error
 */
/*----------------------------------------------------------------------------*/
const char * isaac_pfits_get_templateid(const cpl_propertylist * plist)
{
    const char * val = cpl_propertylist_get_string(plist, "ESO TPL ID");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the central wavelength
  @param    plist       property list to read from
  @return   the requested value
 */
/*----------------------------------------------------------------------------*/
double isaac_pfits_get_wlen(const cpl_propertylist * plist)
{
    const double val = cpl_propertylist_get_double(plist, "ESO INS GRAT WLEN");

    (void)cpl_error_set(cpl_func, cpl_error_get_code());
    return val;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out the filter
  @param    plist       property list to read from
  @return   Pointer to string in the propertylist, or NULL on error
 */
/*----------------------------------------------------------------------------*/
const char * isaac_pfits_get_filter(const cpl_propertylist * plist)
{
    const char * val = isaac_pfits_get_arm(plist);
    const char * retval = val;
        
    if (val != NULL) {
        if (val[0]=='S')        retval = isaac_pfits_get_filter_sw(plist);
        else if (val[0]=='L')   retval = isaac_pfits_get_filter_lw(plist);
    } else                      retval = isaac_pfits_get_filter_sw(plist);

    return retval;
}

/**@}*/
