/* $Id: isaac_img_jitter.c,v 1.75 2013-03-12 08:06:48 llundin Exp $
 *
 * This file is part of the ISAAC Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2013-03-12 08:06:48 $
 * $Revision: 1.75 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <string.h>
#include <math.h>
#include <cpl.h>

#include "irplib_plugin.h"
#include "irplib_utils.h"
#include "irplib_calib.h"

#include "isaac_utils.h"
#include "isaac_pfits.h"
#include "isaac_dfs.h"

/*-----------------------------------------------------------------------------
                                Define
 -----------------------------------------------------------------------------*/

#define RECIPE_STRING "isaac_img_jitter"

#ifndef NEGLIG_OFF_DIFF
#define NEGLIG_OFF_DIFF     0.1
#endif

#define SQR(x) ((x)*(x))

#ifdef ISAAC_IMG_ERR_ESTIMATE
#ifndef ISAAC_IMG_GAIN
/* FIXME: Empirically determined. Use header info (arm dependent) */
#define ISAAC_IMG_GAIN 4.5
#endif

#ifndef ISAAC_IMG_SATURATION_LEVEL
/* FIXME: Empirically determined */
#define ISAAC_IMG_SATURATION_LEVEL 40000.0
#endif
#endif


/*-----------------------------------------------------------------------------
                            Private Function prototypes
 -----------------------------------------------------------------------------*/

static cpl_image ** isaac_img_jitter_reduce(cpl_frameset *,
                                            const cpl_parameterlist *,
                                            const cpl_frameset *,
                                            const cpl_frameset *, 
                                            const char *, const char *,
                                            const char *, cpl_vector **);
static cpl_imagelist ** isaac_img_jitter_load(const cpl_frameset *,
                                              const cpl_frameset *);
static cpl_vector * isaac_img_jitter_sky(cpl_imagelist **, cpl_imagelist *);
static cpl_vector * isaac_img_jitter_sky_running(cpl_imagelist **); 
static cpl_image ** isaac_img_jitter_saa_nochop(cpl_imagelist *,
                                                const cpl_frameset *);
static cpl_image ** isaac_img_jitter_saa_chop(cpl_imagelist *,
                                              const cpl_frameset *);
static int isaac_img_jitter_chopping_classif(cpl_bivector *, int **, int **);
static cpl_error_code isaac_img_jitter_sub_row_median(cpl_image *);
static cpl_table * isaac_img_jitter_qc(cpl_image *);
static double isaac_img_jitter_get_mode(cpl_vector *);
static cpl_error_code isaac_img_jitter_save(cpl_frameset *, const cpl_image *,
                                            const cpl_image *,
                                            const cpl_table *,
                                            const cpl_vector *, 
                                            const cpl_parameterlist *);

cpl_recipe_define(isaac_img_jitter, ISAAC_BINARY_VERSION,
                  "Lars Lundin", PACKAGE_BUGREPORT, "2008", 
                  "ISAAC imaging jitter recipe",
                  RECIPE_STRING " -- ISAAC imaging jitter recipe.\n"
                  "The files listed in the Set Of Frames (sof-file) " 
                  "must be tagged:\n"
                  "raw-file.fits "  ISAAC_IMG_JITTER_OBJ_RAW  " or\n"
                  "raw-file.fits "  ISAAC_IMG_JITTER_SKY_RAW  " or\n"
                  "raw-file.fits "  ISAAC_IMG_JITTER_CHOP_RAW " or\n"
                  "flat-file.fits " ISAAC_CALIB_FLAT          " or\n"
                  "bpm-file.fits "  ISAAC_CALIB_BPM           " or\n"
                  "dark-file.fits " ISAAC_CALIB_DARK          "\n");

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs */
    const char      *   offsets;
    const char      *   objects;
    int                 oddeven;
    int                 sky_minnb;
    int                 sky_halfw;
    int                 sky_rejmin;
    int                 sky_rejmax;
    int                 sx;
    int                 sy;
    int                 mx;
    int                 my;
    cpl_geom_combine    comb_meth;
    int                 saa_refine;
    int                 rej_low;
    int                 rej_high;
    int                 row_med;
    int                 chopping;
    /* Outputs */
    double          pixscale;
    double          dit;
    int             nb_obj_frames;
    int             nb_sky_frames;
    int             nb_rej_frames;
    double          iq;
    int             nbobjs;
    double          fwhm_pix;
    double          fwhm_arcsec;
    double          fwhm_mode;
    double          angle_med;
    double          ellip_med;
} isaac_img_jitter_config;

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Fill the recipe parameterlist
  @param    self  The parameterlist
  @return   0 iff everything is ok

 */
/*----------------------------------------------------------------------------*/
static
cpl_error_code isaac_img_jitter_fill_parameterlist(cpl_parameterlist * self)
{
    const char * context = PACKAGE "." RECIPE_STRING;
    cpl_error_code err;

    cpl_ensure_code(self, CPL_ERROR_NULL_INPUT);

    /* Fill the parameters list */

    /* --off */
    err = irplib_parameterlist_set_string(self, PACKAGE, RECIPE_STRING,
                                          "offsets", NULL, "off", context,
                       "An optional ASCII specification of the offsets "
                       "in case those in FITS-headers are missing or wrong. "
                       "The file must consist of one line per object FITS-"
                       "file and each line must consist of two "
                       "numbers which represent the shift in pixels of that "
                       "image relative to the first image. The first line "
                       "should thus comprise two zeros. Correct FITS-header "
                       "offsets mean that the i'th X offset can be gotten "
                       "from Xoffset_0 - Xoffset_i, where Xoffset_i is the "
                       "value of ESO SEQ CUMOFFSETX and likewise for Y.");
    cpl_ensure_code(!err, err);

    /* --objs */
    err = irplib_parameterlist_set_string(self, PACKAGE, RECIPE_STRING,
                                          "objects", NULL, "objs", context,
                                          "objects file");
    cpl_ensure_code(!err, err);

    /* --oddeven */
    err = irplib_parameterlist_set_bool(self, PACKAGE, RECIPE_STRING, "oddeven",
                                        CPL_FALSE, NULL, context, "Flag to "
                                        "correct the oddeven column effect");
    cpl_ensure_code(!err, err);

    /* --sky_par */
    err = irplib_parameterlist_set_string(self, PACKAGE, RECIPE_STRING,
                                          "sky_par", "10,7,3,3", NULL, context,
                                          "Rejection parameters for sky "
                                          "filtering");
    cpl_ensure_code(!err, err);

    /* --xcorr */
    err = irplib_parameterlist_set_string(self, PACKAGE, RECIPE_STRING, "xcorr",
                                          "40,40,65,65", NULL, context, "Cross "
                                          "correlation search and measure "
                                          "sizes");
    cpl_ensure_code(!err, err);

    /* --comb_meth */
    err = irplib_parameterlist_set_string(self, PACKAGE, RECIPE_STRING,
                                          "comb_meth", "union", NULL, context,
                                          "union / inter / first");
    cpl_ensure_code(!err, err);
 
    /* --saa_refine */
    err = irplib_parameterlist_set_bool(self, PACKAGE, RECIPE_STRING,
                                        "saa_refine", CPL_TRUE, NULL, context,
                                        "Flag to refine the offsets");
    cpl_ensure_code(!err, err);

    /* --rej */
    err = irplib_parameterlist_set_string(self, PACKAGE, RECIPE_STRING, "rej",
                                          "2,2", NULL, context, "Low and high "
                                          "number of rejected values");
    cpl_ensure_code(!err, err);

    /* --row_med */
    err = irplib_parameterlist_set_bool(self, PACKAGE, RECIPE_STRING,
                                        "row_med", CPL_TRUE, NULL, context,
                                        "Flag to subtract the median of each "
                                        "row");
    cpl_ensure_code(!err, err);

    return CPL_ERROR_NONE;
}



/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    The recipe data reduction part is implemented here 
  @param    framelist   the frames list
  @param    parlist     the parameters list
  @return   0 iff everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_jitter(cpl_frameset            * framelist,
                            const cpl_parameterlist * parlist)
{
    const char   * sval;
    const char   * badpix;
    const char   * flat;
    const char   * dark;
    cpl_frameset * objframes = NULL;
    cpl_frameset * skyframes = NULL;
    cpl_image   ** combined = NULL;
    cpl_table    * objs_stats = NULL;
    cpl_vector   * sky_bg = NULL;

    bug_if(0);

    /* Initialise */
    isaac_img_jitter_config.pixscale = -1.0;
    isaac_img_jitter_config.dit = -1.0;
    isaac_img_jitter_config.iq = -1.0;
    isaac_img_jitter_config.nbobjs = -1;
    isaac_img_jitter_config.fwhm_pix = -1.0;
    isaac_img_jitter_config.fwhm_arcsec = -1.0;
    isaac_img_jitter_config.fwhm_mode = -1.0;
    isaac_img_jitter_config.offsets = NULL;
    isaac_img_jitter_config.objects = NULL;
    isaac_img_jitter_config.nb_obj_frames = 0;
    isaac_img_jitter_config.nb_rej_frames = 0;
    isaac_img_jitter_config.nb_sky_frames = 0;

    /* Retrieve input parameters */
    /* Offsets */
    isaac_img_jitter_config.offsets
        = irplib_parameterlist_get_string(parlist, PACKAGE, RECIPE_STRING,
                                       "offsets");

    /* Objects */
    isaac_img_jitter_config.objects
        = irplib_parameterlist_get_string(parlist, PACKAGE, RECIPE_STRING,
                                       "objects");

    /* Oddeven */
    isaac_img_jitter_config.oddeven
        = irplib_parameterlist_get_bool(parlist, PACKAGE, RECIPE_STRING,
                                       "oddeven");

    /* Rejection parameters for sky filtering */
    sval = irplib_parameterlist_get_string(parlist, PACKAGE, RECIPE_STRING,
                                       "sky_par");

    bug_if(sval == NULL);

    skip_if (sscanf(sval, "%d,%d,%d,%d",
                    &isaac_img_jitter_config.sky_minnb,
                    &isaac_img_jitter_config.sky_halfw,
                    &isaac_img_jitter_config.sky_rejmin,
                    &isaac_img_jitter_config.sky_rejmax) != 4);

    /* Cross correlation windows parameters */
    sval = irplib_parameterlist_get_string(parlist, PACKAGE, RECIPE_STRING,
                                       "xcorr");
    bug_if(sval == NULL);

    skip_if (sscanf(sval, "%d,%d,%d,%d",
                    &isaac_img_jitter_config.sx,
                    &isaac_img_jitter_config.sy,
                    &isaac_img_jitter_config.mx,
                    &isaac_img_jitter_config.my) != 4);

    /* --comb_meth */
    sval = irplib_parameterlist_get_string(parlist, PACKAGE, RECIPE_STRING,
                                       "comb_meth");
    bug_if(sval == NULL);
    if (!strcmp(sval, "union")) 
        isaac_img_jitter_config.comb_meth = CPL_GEOM_UNION;
    else if (!strcmp(sval, "inter"))
        isaac_img_jitter_config.comb_meth = CPL_GEOM_INTERSECT;
    else if (!strcmp(sval, "first"))
        isaac_img_jitter_config.comb_meth = CPL_GEOM_FIRST;
    else {
        cpl_msg_error(cpl_func, "Invalid combine method specified");
        skip_if(1);
    }
    
    /* Refine of offsets */
    isaac_img_jitter_config.saa_refine
        = irplib_parameterlist_get_bool(parlist, PACKAGE, RECIPE_STRING,
                                        "saa_refine");

     /* Number of rejected values in stacking */
    sval = irplib_parameterlist_get_string(parlist, PACKAGE, RECIPE_STRING,
                                       "rej");
    bug_if(sval == NULL);
    skip_if (sscanf(sval, "%d,%d",
                    &isaac_img_jitter_config.rej_low,
                    &isaac_img_jitter_config.rej_high) != 2);

    /* Row median */
    isaac_img_jitter_config.row_med
        = irplib_parameterlist_get_bool(parlist, PACKAGE, RECIPE_STRING,
                                        "row_med");

    /* Identify the RAW and CALIB frames in the input frameset */
    skip_if (isaac_dfs_set_groups(framelist));

    /* Retrieve calibration data */
    flat   = isaac_extract_filename(framelist, ISAAC_CALIB_FLAT);
    dark   = isaac_extract_filename(framelist, ISAAC_CALIB_DARK);
    badpix = isaac_extract_filename(framelist, ISAAC_CALIB_BPM);
    
    /* Retrieve raw frames */
    if ((objframes = isaac_extract_frameset(framelist,
                    ISAAC_IMG_JITTER_OBJ_RAW)) != NULL) {
        isaac_img_jitter_config.chopping = 0;
    } else if ((objframes = isaac_extract_frameset(framelist,
                    ISAAC_IMG_JITTER_CHOP_RAW)) != NULL) {
        isaac_img_jitter_config.chopping = 1;
    } else {
        cpl_msg_error(cpl_func, "Cannot find objs frames in the input list");
        skip_if(1);
    }
    skyframes = isaac_extract_frameset(framelist, ISAAC_IMG_JITTER_SKY_RAW);
    
    /* Apply the reduction */
    cpl_msg_info(cpl_func, "Apply the data recombination");
    cpl_msg_indent_more();
    if ((combined = isaac_img_jitter_reduce(framelist, parlist, objframes,
                                            skyframes, flat,
                                            dark, badpix, &sky_bg)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot recombine the data");
        cpl_msg_indent_less();
        skip_if(1);
    }
    cpl_msg_indent_less();
   
    /* Compute QC parameters from the combined image */
    cpl_msg_info(cpl_func, "Compute QC parameters from the combined image");
    cpl_msg_indent_more();
    if ((objs_stats = isaac_img_jitter_qc(combined[0])) == NULL) {
        cpl_msg_warning(cpl_func, "Cannot compute all parameters");
    }
    cpl_msg_indent_less();
    
    /* Save the products */
    cpl_msg_info(cpl_func, "Save the products");
    skip_if (isaac_img_jitter_save(framelist, combined[0], combined[1],
                                   objs_stats, sky_bg, parlist));
    
    end_skip;

    cpl_frameset_delete(objframes);
    cpl_frameset_delete(skyframes);
    if (combined != NULL) {
        cpl_image_delete(combined[0]);
        cpl_image_delete(combined[1]);
        cpl_free(combined);
    }
    cpl_table_delete(objs_stats);
    cpl_vector_delete(sky_bg);

    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part is implemented here 
  @param    obj         the objects frames
  @param    sky         the sky frames or NULL
  @param    flat        the flat field or NULL
  @param    dark        the dark or NULL
  @param    bpm         the bad pixels map or NULL
  @param    skybg       the computed sky background values
  @return   the combined image and the contribution map  or NULL in error case
 */
/*----------------------------------------------------------------------------*/
static
cpl_image ** isaac_img_jitter_reduce(cpl_frameset            * frameset,
                                     const cpl_parameterlist * parlist,
                                     const cpl_frameset    *   obj,
                                     const cpl_frameset    *   sky,
                                     const char      *   flat,
                                     const char      *   dark,
                                     const char      *   bpm,
                                     cpl_vector      **  skybg)
{
#ifdef ISAAC_IMG_ERR_ESTIMATE
    cpl_propertylist * qclist;
    cpl_imagelist    * err;
#endif
    cpl_imagelist   **  in;
    cpl_image       **  combined;
    cpl_imagelist   *   corrected;
    cpl_image       *   cur_im;
    int                 i;

    cpl_ensure(frameset != NULL, CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(parlist  != NULL, CPL_ERROR_NULL_INPUT, NULL);

    /* Initialise */
    *skybg = NULL;
    combined = NULL;
    
    /* Load the input data */
    cpl_msg_info(cpl_func, "Load the input data");
    cpl_msg_indent_more();
    if ((in = isaac_img_jitter_load(obj, sky)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot load input data");
        cpl_msg_indent_less();
        return NULL;
    }
    cpl_msg_indent_less();

    /* Apply the odd-even correction */
    if (isaac_img_jitter_config.oddeven) {
        cpl_msg_info(cpl_func, "Apply the odd-even effect correction");
        cpl_msg_indent_more();
        corrected = cpl_imagelist_new();
        for (i=0; i<cpl_imagelist_get_size(in[0]); i++) {
            cpl_msg_info(cpl_func, "Correct object frame nb %d", i+1);
            if ((cur_im = isaac_oddeven_correct(
                            cpl_imagelist_get(in[0], i))) == NULL) {
                cpl_msg_warning(cpl_func,"Problem in odd-even corr. %d",i+1);
                cpl_imagelist_delete(corrected);
                corrected = NULL;
                break;
            } 
            cpl_imagelist_set(corrected, cur_im, i);
        }
        /* Odd/even correction successfull */
        if (corrected != NULL) {
            cpl_imagelist_delete(in[0]);
            in[0] = corrected;
        }
        cpl_msg_indent_less();
    }

#ifdef ISAAC_IMG_ERR_ESTIMATE
    /*----------------------------------------------------------------------------*/
    /* Save the imagelist before applying the correction */

    err = cpl_imagelist_duplicate(in[0]);

    /* Dark correction */
    if (dark != NULL) {
        cpl_image * dark_image;

        cpl_msg_info(cpl_func, "Error propagation: Subtract the dark to the images");
        /* Load the dark image */
        if ((dark_image = cpl_image_load(dark, CPL_TYPE_FLOAT, 0, 0)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot load the dark %s", dark);
            return NULL;
        }
        /* Apply the dark correction to the images */
        if (cpl_imagelist_subtract_image(err, dark_image)!=CPL_ERROR_NONE) {
            cpl_msg_error(cpl_func, "Cannot apply the dark to the images");
            cpl_image_delete(dark_image);
            return NULL;
        }
        cpl_image_delete(dark_image);
    }


    /* Apply the bad pixel mask requested */
    if (bpm != NULL) {
        cpl_image * bpm_im;
        cpl_msg_info(cpl_func, "Error propagation: Correct the bad pixels in the images");
        /* Load the bad pixels image - as double */
        if ((bpm_im = cpl_image_load(bpm, CPL_TYPE_DOUBLE, 0, 0)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot load the bad pixel map %s", bpm);
            return NULL;
        }
        /* Invert the map */
        cpl_image_threshold(bpm_im, 0.4, 0.6, 1, 0);

        cpl_imagelist_multiply_image(err, bpm_im);

        /*        cpl_image_save(bpm_im,"test.fits",CPL_BPP_IEEE_FLOAT,NULL,CPL_IO_CREATE);*/
        cpl_image_delete(bpm_im);
    }

    cpl_imagelist_threshold(err, 1, ISAAC_IMG_SATURATION_LEVEL,
                            FLT_MAX, FLT_MAX);
    cpl_imagelist_divide_scalar(err, ISAAC_IMG_GAIN);
    cpl_imagelist_power(err, 0.5);

    /* Flat-field correction */
    if (flat != NULL) {
        cpl_image * flat_image;
        cpl_msg_info(cpl_func, "Error propagation: Divide the images by the flatfield");
        /* Load the flat image */
        if ((flat_image = cpl_image_load(flat, CPL_TYPE_FLOAT, 0, 0)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot load the flat field %s", flat);
            return NULL;
        }
        /* Apply the flatfield correction to the images */
        if (cpl_imagelist_divide_image(err, flat_image)!=CPL_ERROR_NONE) {
            cpl_msg_error(cpl_func, "Cannot apply the flatfield to the images");
            cpl_image_delete(flat_image);
            return NULL;
        }
        cpl_image_delete(flat_image);
    }
    cpl_imagelist_threshold(err, 1, ISAAC_IMG_SATURATION_LEVEL,
                            FLT_MAX, FLT_MAX);

    /*----------------------------------------------------------------------------*/
#endif

    /* Apply the calibrations */
    if (flat || dark || bpm) {
        cpl_msg_info(cpl_func, "Apply the calibrations");
        cpl_msg_indent_more();
        if (irplib_flat_dark_bpm_calib(in[0], flat, dark, bpm) == -1) {
            /* Calibrate the objects */
            cpl_msg_error(cpl_func, "Cannot calibrate the objects");
            cpl_imagelist_delete(in[0]);
            if (in[1]) cpl_imagelist_delete(in[1]);
            cpl_free(in);
            cpl_msg_indent_less();
            return NULL;
        }
        if (in[1]) {
            if (irplib_flat_dark_bpm_calib(in[1], flat, dark, bpm) == -1) {
                /* Calibrate the sky */
                cpl_msg_error(cpl_func, "Cannot calibrate the sky");
                cpl_imagelist_delete(in[0]);
                if (in[1]) cpl_imagelist_delete(in[1]);
                cpl_free(in);
                cpl_msg_indent_less();
                return NULL;
            }
        }
        cpl_msg_indent_less();
    }


#ifdef ISAAC_IMG_ERR_ESTIMATE
    /* Save the imagelist as intermediate product */

    qclist = cpl_propertylist_new();

    cpl_propertylist_update_string(qclist, CPL_DFS_PRO_CATG, "ISAAC_IMAGELIST");
    if(cpl_dfs_save_imagelist(frameset,
                              NULL,
                              parlist,
                              frameset,
                              NULL,
                              in[0],
                              CPL_BPP_IEEE_FLOAT,
                              RECIPE_STRING,
                              qclist,
                              NULL,
                              PACKAGE "/" PACKAGE_VERSION,
                              "isaac_img_imagelist.fits")) {
        /* Propagate the error */
        (void)cpl_error_set_where(cpl_func);
    }

    cpl_propertylist_update_string(qclist, CPL_DFS_PRO_CATG, "ISAAC_IMAGELISTERR");
    if(cpl_dfs_save_imagelist(frameset,
                              NULL,
                              parlist,
                              frameset,
                              NULL,
                              err,
                              CPL_BPP_IEEE_FLOAT,
                              RECIPE_STRING,
                              qclist,
                              NULL,
                              PACKAGE "/" PACKAGE_VERSION,
                              "isaac_img_imagelist_error.fits")) {
        /* Propagate the error */
        (void)cpl_error_set_where(cpl_func);
    }

    cpl_propertylist_delete(qclist);
    cpl_imagelist_delete(err);
#endif

    /* Apply the sky correction only in jitter mode */
    if (isaac_img_jitter_config.chopping == 0) {
        cpl_msg_info(cpl_func, "Sky estimation and correction");
        cpl_msg_indent_more();
        if ((*skybg = isaac_img_jitter_sky(&(in[0]), in[1])) == NULL) {
            cpl_msg_error(cpl_func, "Cannot estimate the sky");
            cpl_imagelist_delete(in[0]);
            if (in[1]) cpl_imagelist_delete(in[1]);
            cpl_free(in);
            cpl_msg_indent_less();
            return NULL;
        }
        cpl_msg_indent_less();
    }
    isaac_img_jitter_config.nb_obj_frames = cpl_imagelist_get_size(in[0]);
    if (in[1] != NULL) 
        isaac_img_jitter_config.nb_sky_frames = cpl_imagelist_get_size(in[1]);
    if (in[1]) cpl_imagelist_delete(in[1]);
    in[1] = NULL;
    
    /* Apply the shift and add */
    cpl_msg_info(cpl_func, "Shift and add");
    cpl_msg_indent_more();
    if (isaac_img_jitter_config.chopping == 0) {
        combined = isaac_img_jitter_saa_nochop(in[0], obj);
    } else if (isaac_img_jitter_config.chopping == 1) {
        combined = isaac_img_jitter_saa_chop(in[0], obj);
    }
    if (combined == NULL) {
        cpl_msg_error(cpl_func, "Cannot apply the shift and add");
        cpl_imagelist_delete(in[0]);
        cpl_free(in);
        if (*skybg != NULL) cpl_vector_delete(*skybg);
        *skybg = NULL;
        cpl_msg_indent_less();
        return NULL;
    }
    cpl_imagelist_delete(in[0]);
    cpl_free(in);  
    cpl_msg_indent_less();
     
    /* Post processing on the combined image */
    if (isaac_img_jitter_config.row_med) {
        cpl_msg_info(cpl_func, "Subtract the median from each row");
        skip_if(isaac_img_jitter_sub_row_median(combined[0]));
    }

    end_skip;

    return combined;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Load the input data (sky and object) 
  @param    obj     the objects frames
  @param    sky     the sky frames or NULL
  @return   A newly allocated pointer to 2 newly allocated images sets, or NULL
 */
/*----------------------------------------------------------------------------*/
static cpl_imagelist ** isaac_img_jitter_load(
        const cpl_frameset    *   obj,
        const cpl_frameset    *   sky)
{
    cpl_imagelist       **  isets;
    cpl_imagelist       *   chop_b;
    const cpl_frame     *   frame;
    cpl_propertylist    *   plist;
    const char          *   sval;
    
    /* Test entries */
    if (obj == NULL) return NULL;

    /* Check if ok */
    cpl_ensure(!cpl_error_get_code(), cpl_error_get_code(), NULL);
   
    /* Get the frame type pixscale and DIT value */
    frame = cpl_frameset_get_position_const(obj, 0);
    cpl_ensure(frame != NULL, cpl_error_get_code(), NULL);

    plist=cpl_propertylist_load(cpl_frame_get_filename(frame), 0);
    cpl_ensure(!cpl_error_get_code(), cpl_error_get_code(), NULL);

    isaac_img_jitter_config.pixscale = isaac_pfits_get_pixscale(plist);
    isaac_img_jitter_config.dit = isaac_pfits_get_dit(plist);
    sval = isaac_pfits_get_frame_type(plist);
    if (sval == NULL) {
        cpl_msg_error(cpl_func, "Could not get frame type for jitter");
        cpl_propertylist_delete(plist);
        return NULL;
    } else if ((isaac_img_jitter_config.chopping==0) && strcmp(sval, "INT")) {
        cpl_msg_error(cpl_func, "Wrong type for jitter: %s - Should be INT", 
                sval);
        cpl_propertylist_delete(plist);
        return NULL;
    } else if ((isaac_img_jitter_config.chopping==1) && strcmp(sval, "CUBE1")) {
        cpl_msg_error(cpl_func, "Wrong type for chopping: %s - Should be CUBE1",
                sval);
        cpl_propertylist_delete(plist);
        return NULL;
    }
    cpl_propertylist_delete(plist);
    if (cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "Missing keyword in FITS header");
        return NULL;
    }

    /* Allocate the image sets */
    isets = cpl_malloc(2 * sizeof(cpl_imagelist *));
    
    /* Check if chop or nochop */
    if (isaac_img_jitter_config.chopping == 0) {
        isets[0] = cpl_imagelist_load_frameset(obj, CPL_TYPE_FLOAT, 1, 0);
        if (sky != NULL) 
            isets[1] = cpl_imagelist_load_frameset(sky, CPL_TYPE_FLOAT, 1, 0);
        else isets[1] = NULL;
    } else if (isaac_img_jitter_config.chopping == 1) {
        isets[1] = NULL;
        isets[0] = cpl_imagelist_load_frameset(obj, CPL_TYPE_FLOAT, 1, 0);
        chop_b = cpl_imagelist_load_frameset(obj, CPL_TYPE_FLOAT, 2, 0);
        cpl_imagelist_subtract(isets[0], chop_b);
        cpl_imagelist_delete(chop_b);
    } 
    
    /* Check if at least the objects are there */
    if (isets[0] == NULL) {
        cpl_msg_error(cpl_func, "The objects frames could not be loaded");
        if (isets[1] != NULL) cpl_imagelist_delete(isets[1]);
        cpl_free(isets);
        return NULL;
    }
    /* Return */
    return isets;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Estimate the sky signal and correct the images.
  @param    objs    the images with objects
  @param    skys    the images with the sky or NULL
  @return   the background values in a vector or NULL in error case
 */
/*----------------------------------------------------------------------------*/
static cpl_vector * isaac_img_jitter_sky(
        cpl_imagelist   **  objs,
        cpl_imagelist   *   skys)
{
    cpl_image       *   sky;
    cpl_vector      *   bg;
    int                 sky_method;
    int                 nframes, nskys;
    double              median;
    cpl_image       *   cur_ima;
    int                 i;

    /* Initialise */
    bg = NULL;
    nframes = cpl_imagelist_get_size(*objs);

    /* Decide the sky method */
    if (isaac_img_jitter_config.sky_minnb > nframes) sky_method = 1;
    else sky_method = 2;

    /* Compute the sky frame */
    if (skys != NULL) {
        cpl_msg_info(cpl_func, "Median of sky images");
        /* Get the median of the sky images */
        nskys = cpl_imagelist_get_size(skys);
        bg = cpl_vector_new(nskys);
        for (i=0; i<nskys; i++) {
            median = cpl_image_get_median(cpl_imagelist_get(skys, i));
            cpl_vector_set(bg, i, median);
        }
        /* Use sky images */
        if ((sky = cpl_imagelist_collapse_median_create(skys)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot compute the median of sky images");
            cpl_vector_delete(bg);
            return NULL;
        }
        /* Correct the objects images  */
        if (cpl_imagelist_subtract_image(*objs, sky) != CPL_ERROR_NONE) {
            cpl_msg_error(cpl_func, "Cannot corr. the obj images from the sky");
            cpl_image_delete(sky);
            cpl_vector_delete(bg);
            return NULL;
        }
        cpl_image_delete(sky);
        /* Normalise the object planes */
        for (i=0; i<nframes; i++) {
            cur_ima = cpl_imagelist_get(*objs, i);
            median = cpl_image_get_median(cur_ima);
            cpl_image_subtract_scalar(cur_ima, median);
        }
    } else if (sky_method == 1) {
        cpl_msg_info(cpl_func, "Median of object images");
         /* Use objs images */
        if ((sky = cpl_imagelist_collapse_median_create(*objs)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot compute the median of obj images");
            return NULL;
        }
        /* Correct the objects images  */
        if (cpl_imagelist_subtract_image(*objs, sky) != CPL_ERROR_NONE) {
            cpl_msg_error(cpl_func, "Cannot corr. the obj images from the sky");
            cpl_image_delete(sky);
            return NULL;
        }
        /* Create a 1 value vector */
        bg = cpl_vector_new(1);
        cpl_vector_set(bg, 0, cpl_image_get_median(sky));
        cpl_image_delete(sky);
        /* Normalise the object planes */
        for (i=0; i<nframes; i++) {
            cur_ima = cpl_imagelist_get(*objs, i);
            median = cpl_image_get_median(cur_ima);
            cpl_image_subtract_scalar(cur_ima, median);
        }
    } else if (sky_method == 2) {
        cpl_msg_info(cpl_func, "Running filter on object images");
        /* Use objects images */
        if ((bg = isaac_img_jitter_sky_running(objs)) == NULL) {
            cpl_msg_error(cpl_func, 
                    "Cannot apply the running filter for the sky");
            return NULL;
        }
    }
    /* Free and return */
    return bg;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Apply the running filter to subtract the sky
  @param    in          the images to correct
  @return   the vector containing the background values or NULL in error case

  This function performs a 3d rejection filter on the input cube.
  Each time line is extracted on +/-halfw. On this line of sight, all
  pixels are first normalized by subtracting the median value of the
  plane they belong to. The central, highest and lowest values are then
  removed and the rest is averaged to yield a value which is subtracted
  from the initial input pixel.

  The background array is returned.
  The array will contain the average value which has been subtracted
  to pixels in each plane. It is usually a good indicator of the
  average subtracted background value if you use this filter to subtract
  an infrared sky background.
 */
/*----------------------------------------------------------------------------*/
static cpl_vector * isaac_img_jitter_sky_running(cpl_imagelist ** in) 
{
    int                 rejmin, rejmax, halfw;
    cpl_imagelist   *   filtres;
    int                 ni, nx, ny, pos;
    cpl_vector      *   medians;
    cpl_image       *   cur_ima;
    float           *   pcur_ima;
    cpl_image       *   tmp_ima;
    float           *   ptmp_ima;
    cpl_vector      *   localwin;
    int                 fr_p, to_p, n_curp;
    cpl_vector      *   bg;
    double              bg_val, out;
    float               one_med;
    int                 i, j, k;

    /* Test entries */
    if (in==NULL || *in == NULL) return NULL;

    /* Initialise */
    rejmin = isaac_img_jitter_config.sky_rejmin;
    rejmax = isaac_img_jitter_config.sky_rejmax;
    halfw  = isaac_img_jitter_config.sky_halfw;
    ni = cpl_imagelist_get_size(*in);
    cur_ima = cpl_imagelist_get(*in, 0);
    nx = cpl_image_get_size_x(cur_ima);
    ny = cpl_image_get_size_y(cur_ima);
    
    /* Tests on validity of rejection parameters */
    if (((rejmin+rejmax)>=halfw) || (halfw<1) || (rejmin<0) || (rejmax<0)) {
        cpl_msg_error(cpl_func, "cannot run filter with rej parms %d (%d-%d)",
                halfw, rejmin, rejmax);
        return NULL;
    }   
    /* Pre-compute median value in each plane */
    medians = cpl_vector_new(ni);
    for (i=0; i<ni; i++) {
        cur_ima = cpl_imagelist_get(*in, i);
        cpl_vector_set(medians, i, cpl_image_get_median(cur_ima));
    }
    /* Allocate output cube */
    filtres = cpl_imagelist_new();

    /* Allocate output bg */
    bg = cpl_vector_new(ni);
    
    /* Main loop over input planes */
    for (k=0; k<ni; k++) {
        /* Allocate output plane */
        tmp_ima = cpl_image_new(nx, ny, CPL_TYPE_FLOAT);
        ptmp_ima = cpl_image_get_data_float(tmp_ima);
        /* Compute border indices */
        fr_p = k - halfw;
        to_p = k + halfw;
        if (fr_p<0) fr_p=0;
        if (to_p>(ni-1)) to_p=ni-1;

        /* Number of valid planes to consider after edge effects */
        n_curp = to_p - fr_p;

        /* Allocate local window */
        localwin = cpl_vector_new(n_curp);
        
        bg_val = 0.0;
        /* Loop over all pixels */
        for (pos=0; pos<nx*ny; pos++) {
            /* Fill up local window */
            j=0;
            for (i=fr_p; i<=to_p; i++) {
                if (i!=k) {
                    cur_ima = cpl_imagelist_get(*in, i);
                    pcur_ima = cpl_image_get_data_float(cur_ima);
                    cpl_vector_set(localwin, j,
                            (double)pcur_ima[pos]-cpl_vector_get(medians, i));
                    j++;
                }
            }
            /* Sort window */
            cpl_vector_sort(localwin, 1);
            /* Reject min and max, accumulate other pixels */
            out = 0.0;
            for (i=rejmin; i<(n_curp-rejmax); i++) {
                out += cpl_vector_get(localwin, i);
            }
            /* Take the mean */
            out /= (double)(n_curp - rejmin - rejmax);
            /* Assign value */
            cur_ima = cpl_imagelist_get(*in, k);
            pcur_ima = cpl_image_get_data_float(cur_ima);
            ptmp_ima[pos] = pcur_ima[pos] - 
                (float)(out + cpl_vector_get(medians, k));

            bg_val += (out+cpl_vector_get(medians, k));
        }
        cpl_vector_delete(localwin);
        cpl_vector_set(bg, k, bg_val/(nx*ny));
        cpl_imagelist_set(filtres, tmp_ima, k);
    }
    cpl_imagelist_delete(*in);
    cpl_vector_delete(medians);

    /* Subtract median from each frame */
    for (i=0; i<ni; i++) {
        cur_ima = cpl_imagelist_get(filtres, i);
        one_med = cpl_image_get_median(cur_ima);
        cpl_image_subtract_scalar(cur_ima, one_med);
    }
    *in = filtres;
    return bg;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Apply the shift and add in jitter mode
  @param    in          the images to stack
  @param    objframes   the list of objects frames
  @return   the combined image and the contribution map or NULL in error case
 */
/*----------------------------------------------------------------------------*/
static cpl_image ** isaac_img_jitter_saa_nochop(
        cpl_imagelist   *   in,
        const cpl_frameset    *   objframes)
{
    const cpl_frame     *   frame;
    cpl_propertylist    *   plist;
    cpl_bivector        *   offsets_est;
    double              *   offsets_est_x;
    double              *   offsets_est_y;
    cpl_bivector        *   objs;
    double              *   objs_x;
    double              *   objs_y;
    cpl_apertures       *   aperts;
    cpl_image           **  combined;
    cpl_vector          *   thresh_vect;
    cpl_image           *   diff;
    int                     nfiles;
    int                     nima;
    int                     refine = 
    isaac_img_jitter_config.saa_refine ? 1 : 0;
    int                     i;

    /* Get the number of images */
    nfiles = cpl_imagelist_get_size(in);
    if (cpl_frameset_get_size(objframes) != nfiles) {
        cpl_msg_error(cpl_func, "Invalid input objects sizes"); 
        return NULL;
    }
    
    /* Get the offsets estimation of each input file pair */
    cpl_msg_info(cpl_func, "Get the offsets estimation");
    offsets_est = NULL;
    if (isaac_img_jitter_config.offsets &&
            isaac_img_jitter_config.offsets[0] != (char)0) {
        /* A file has been provided on the command line */
        offsets_est = cpl_bivector_read(isaac_img_jitter_config.offsets);
        if ((offsets_est==NULL)||(cpl_bivector_get_size(offsets_est)!=nfiles)) {
            cpl_msg_error(cpl_func, "Cannot get offsets from %s", 
                    isaac_img_jitter_config.offsets);
            return NULL;
        }
    } else {
        /* Get the offsets from the header */
        offsets_est = cpl_bivector_new(nfiles);
        offsets_est_x = cpl_bivector_get_x_data(offsets_est);
        offsets_est_y = cpl_bivector_get_y_data(offsets_est);
        for (i=0; i<nfiles; i++) {
            if (cpl_error_get_code()) {
                cpl_bivector_delete(offsets_est);
                offsets_est = NULL;
                break;
            }
            /* X and Y offsets */
            frame = cpl_frameset_get_position_const(objframes, i);
            plist=cpl_propertylist_load(cpl_frame_get_filename(frame),0);
            offsets_est_x[i] = -1.0 * isaac_pfits_get_cumoffsetx(plist);
            offsets_est_y[i] = -1.0 * isaac_pfits_get_cumoffsety(plist);
            cpl_propertylist_delete(plist);
            if (cpl_error_get_code()) {
                cpl_msg_warning(cpl_func, "Cannot get offsets from header");
                cpl_bivector_delete(offsets_est);
                offsets_est = NULL;
                break;
            }
            if (i > 0) {
                /* Subtract the first offset from all offsets */
                offsets_est_x[i] -= offsets_est_x[0];
                offsets_est_y[i] -= offsets_est_y[0];
            }
        }
        if (i == nfiles) {
            offsets_est_x[0] = offsets_est_y[0] = 0.0;
        }
    }

    /* Read the provided objects file if provided */
    objs = NULL;
    if (isaac_img_jitter_config.objects &&
            isaac_img_jitter_config.objects[0] != (char)0) {
        cpl_msg_info(cpl_func, "Get the user provided correlation objects");
        /* A file has been provided on the command line */
        objs = cpl_bivector_read(isaac_img_jitter_config.objects);
        if (objs==NULL) {
            cpl_msg_error(cpl_func, "Cannot get objects from %s",
                    isaac_img_jitter_config.objects);
            if (offsets_est) cpl_bivector_delete(offsets_est);
            return NULL;
        }
    }

    /* Get a correlation point from the difference of the first images */
    if (objs == NULL) {
        cpl_msg_info(cpl_func, "Get a cross-correlation point");
        thresh_vect = cpl_vector_new(4);
        cpl_vector_set(thresh_vect, 0, 5.0);
        cpl_vector_set(thresh_vect, 1, 2.0);
        cpl_vector_set(thresh_vect, 2, 1.0);
        cpl_vector_set(thresh_vect, 3, 0.5);
        diff = cpl_image_subtract_create(cpl_imagelist_get(in, 0),
                cpl_imagelist_get(in, 1));
        if ((aperts = cpl_apertures_extract_window(diff, thresh_vect, 
                        200, 200, 800, 800, NULL)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot find any cross-correlation point");
            if (offsets_est) cpl_bivector_delete(offsets_est);
            cpl_vector_delete(thresh_vect);
            cpl_image_delete(diff);
            return NULL;
        }
        cpl_image_delete(diff);
        cpl_vector_delete(thresh_vect);
        cpl_apertures_sort_by_npix(aperts);
        objs = cpl_bivector_new(1);
        objs_x = cpl_bivector_get_x_data(objs);
        objs_y = cpl_bivector_get_y_data(objs);
        objs_x[0] = cpl_apertures_get_pos_x(aperts, 1);
        objs_y[0] = cpl_apertures_get_pos_y(aperts, 1);
        cpl_apertures_delete(aperts);
        if (objs == NULL) {
            cpl_msg_error(cpl_func, "Cannot find any cross-correlation point");
            if (offsets_est) cpl_bivector_delete(offsets_est);
            return NULL;
        }
        cpl_msg_info(cpl_func, 
                "Correlation point: %g %g\n", objs_x[0], objs_y[0]);
    }
    
    /* Recombine the images */
    cpl_msg_info(cpl_func, "Recombine the images set");
    cpl_msg_indent_more();
    if ((combined = cpl_geom_img_offset_combine(in, offsets_est, refine, objs,
                    NULL, NULL,
                    isaac_img_jitter_config.sx, 
                    isaac_img_jitter_config.sy,
                    isaac_img_jitter_config.mx, 
                    isaac_img_jitter_config.my,
                    isaac_img_jitter_config.rej_low,
                    isaac_img_jitter_config.rej_high,
                    isaac_img_jitter_config.comb_meth)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot recombine the images");
        if (offsets_est) cpl_bivector_delete(offsets_est);
        cpl_bivector_delete(objs);
        cpl_msg_indent_less();
        return NULL;
    }
    /* Update QC params */
    i = (int)(cpl_image_get_max(combined[1]));
    nima = cpl_imagelist_get_size(in);
    if ((nima > 3) && (nima>2*
            (isaac_img_jitter_config.rej_low+isaac_img_jitter_config.rej_high)))
        i += isaac_img_jitter_config.rej_low+isaac_img_jitter_config.rej_high;
    isaac_img_jitter_config.nb_rej_frames = 
        isaac_img_jitter_config.nb_obj_frames - i;
    isaac_img_jitter_config.nb_obj_frames = i;
    cpl_msg_indent_less();

    /* Free and return */
    if (offsets_est) cpl_bivector_delete(offsets_est);
    cpl_bivector_delete(objs);
    return combined;
}


/*----------------------------------------------------------------------------*/
/**
  @brief    Apply the shift and add in chopping mode
  @param    in          the images to stack
  @param    objframes   the list of objects frames
  @return   the combined image and the contribution map or NULL in error case
 */
/*----------------------------------------------------------------------------*/
static cpl_image ** isaac_img_jitter_saa_chop(
        cpl_imagelist   *   in,
        const cpl_frameset    *   objframes)
{
    const cpl_frame     *   frame;
    cpl_propertylist    *   plist;
    cpl_bivector        *   offsets_est;
    cpl_bivector        *   offsets_est_nod;
    double              *   offsets_est_x;
    double              *   offsets_est_y;
    double              *   offsets_est_nod_x;
    double              *   offsets_est_nod_y;
    cpl_bivector        *   objs;
    cpl_image           **  combined;
    cpl_vector          *   thresh_vect;
    int                     nfiles, nb_chop;
    int                 *   chop_a;
    int                 *   chop_b;
    cpl_imagelist       *   nodded;
    cpl_image           *   tmp_ima;
    int                     refine = 
    isaac_img_jitter_config.saa_refine ? 1 : 0;
    int                     i;

    /* Get the number of images */
    nfiles = cpl_imagelist_get_size(in);
    if (cpl_frameset_get_size(objframes) != nfiles) {
        cpl_msg_error(cpl_func, "Invalid input objects sizes"); 
        return NULL;
    }
   
    /* Get the offsets estimation of each input file pair */
    cpl_msg_info(cpl_func, "Get the offsets estimation");
    offsets_est = NULL;
    if (isaac_img_jitter_config.offsets &&
            isaac_img_jitter_config.offsets[0] != (char)0) {
        /* A file has been provided on the command line */
        offsets_est = cpl_bivector_read(isaac_img_jitter_config.offsets);
        if ((offsets_est==NULL) || 
                (cpl_bivector_get_size(offsets_est) != nfiles)) {
            cpl_msg_error(cpl_func, "Cannot get offsets from %s", 
                    isaac_img_jitter_config.offsets);
            return NULL;
        }
        offsets_est_x = cpl_bivector_get_x_data(offsets_est);
        offsets_est_y = cpl_bivector_get_y_data(offsets_est);
    } else {
        /* Get the offsets from the header */
        offsets_est = cpl_bivector_new(nfiles);
        offsets_est_x = cpl_bivector_get_x_data(offsets_est);
        offsets_est_y = cpl_bivector_get_y_data(offsets_est);
        for (i=0; i<nfiles; i++) {
            if (cpl_error_get_code()) {
                cpl_bivector_delete(offsets_est);
                return NULL;
            }
            /* X and Y offsets */
            frame = cpl_frameset_get_position_const(objframes, i);
            plist=cpl_propertylist_load(cpl_frame_get_filename(frame),0);
            offsets_est_x[i] = -1.0 * isaac_pfits_get_cumoffsetx(plist);
            offsets_est_y[i] = -1.0 * isaac_pfits_get_cumoffsety(plist);
            cpl_propertylist_delete(plist);
            if (cpl_error_get_code()) {
                cpl_msg_warning(cpl_func, "Cannot get offsets from header");
                cpl_bivector_delete(offsets_est);
                return NULL;
            }
        }
        /* Subtract the first offset to all offsets */
        for (i=1; i<nfiles; i++) {
            offsets_est_x[i] -= offsets_est_x[0];
            offsets_est_y[i] -= offsets_est_y[0];
        }
        offsets_est_x[0] = offsets_est_y[0] = 0.00;
    }

    /* Classify the A and B nodding positions using the offsets */
    if ((nb_chop=isaac_img_jitter_chopping_classif(offsets_est, 
                    &chop_a, &chop_b)) == -1) {
        cpl_msg_error(cpl_func, "cannot classify chopped frames");
        cpl_bivector_delete(offsets_est);
        return NULL;
    }
    
    /* Create the nodded frames chop_a-chop_b */
    nodded = cpl_imagelist_new();
    for (i=0; i<nb_chop; i++) {
        tmp_ima = cpl_image_subtract_create(
                cpl_imagelist_get(in, chop_a[i]),
                cpl_imagelist_get(in, chop_b[i]));
        cpl_imagelist_set(nodded, tmp_ima, i);
    }

    /* Update the offsets estimations between the nodded frames */
    offsets_est_nod = cpl_bivector_new(nb_chop);
    offsets_est_nod_x = cpl_bivector_get_x_data(offsets_est_nod);
    offsets_est_nod_y = cpl_bivector_get_y_data(offsets_est_nod);
    for (i=0; i<nb_chop; i++) {
        offsets_est_nod_x[i] = offsets_est_x[chop_a[i]];
        offsets_est_nod_y[i] = offsets_est_y[chop_a[i]];
    }
    cpl_bivector_delete(offsets_est);
    cpl_free(chop_a);
    cpl_free(chop_b);

    /* Read the provided objects file if provided */
    objs = NULL;
    if (isaac_img_jitter_config.objects &&
            isaac_img_jitter_config.objects[0] != (char)0) {
        cpl_msg_info(cpl_func, "Get the user provided correlation objects");
        /* A file has been provided on the command line */
        objs = cpl_bivector_read(isaac_img_jitter_config.objects);
        if (objs==NULL) {
            cpl_msg_error(cpl_func, "Cannot get objects from %s",
                    isaac_img_jitter_config.objects);
            cpl_bivector_delete(offsets_est_nod);
            cpl_imagelist_delete(nodded);
            return NULL;
        }
    }

    /* Create the vector for the detection thresholds */
    thresh_vect = cpl_vector_new(4);
    cpl_vector_set(thresh_vect, 0, 5.0);
    cpl_vector_set(thresh_vect, 1, 2.0);
    cpl_vector_set(thresh_vect, 2, 1.0);
    cpl_vector_set(thresh_vect, 3, 0.5);
    
    /* Recombine the images */
    cpl_msg_info(cpl_func, "Recombine the images set");
    cpl_msg_indent_more();
    if ((combined = cpl_geom_img_offset_combine(nodded, offsets_est_nod,
            refine, objs, thresh_vect, NULL,
                    isaac_img_jitter_config.sx, 
                    isaac_img_jitter_config.sy,
                    isaac_img_jitter_config.mx, 
                    isaac_img_jitter_config.my,
                    isaac_img_jitter_config.rej_low,
                    isaac_img_jitter_config.rej_high,
                    isaac_img_jitter_config.comb_meth)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot recombine the images");
        cpl_bivector_delete(offsets_est_nod);
        cpl_imagelist_delete(nodded);
        if (objs) cpl_bivector_delete(objs);
        cpl_vector_delete(thresh_vect);
        cpl_msg_indent_less();
        return NULL;
    }
    /* Update QC params */
    i = (int)(cpl_image_get_max(combined[1]));
    isaac_img_jitter_config.nb_rej_frames = 
        isaac_img_jitter_config.nb_obj_frames - i;
    isaac_img_jitter_config.nb_obj_frames = i;
    cpl_msg_indent_less();

    /* Free and return */
    cpl_vector_delete(thresh_vect);
    cpl_bivector_delete(offsets_est_nod);
    cpl_imagelist_delete(nodded);
    if (objs) cpl_bivector_delete(objs);
    return combined;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Classification of frames in chopping mode 
  @param    offsets offsets of the frames
  @param    chop_a  index array giving the positions of chopA frames
  @param    chop_b  index array giving the positions of chopB frames
  @return   Number of chop a frames (= nb of chop b frames) 
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_jitter_chopping_classif(
        cpl_bivector    *   offsets,
        int             **  chop_a,
        int             **  chop_b)
{
    int             nb_obj;
    double      *   offsets_x;
    double      *   offsets_y;
    double          throw_x,
                    throw_y;
    double      *   dist;
    int         *   dcount;
    int             max_count;
    int             i_max, j_max;
    int             classified;
    int             i, j, k, l;

    /* Test entries */
    if (offsets==NULL || chop_a==NULL || chop_b==NULL) return CPL_ERROR_UNSPECIFIED;

    /* Initialise */
    *chop_a = *chop_b = NULL;
    
    /* Find number of type_obj frames */
    nb_obj = cpl_bivector_get_size(offsets);
    offsets_x = cpl_bivector_get_x_data(offsets);
    offsets_y = cpl_bivector_get_y_data(offsets);

    /* If no object frame, exit */
    if (nb_obj == 0) return CPL_ERROR_UNSPECIFIED;

    /* Odd number of frames ? */
    if (nb_obj%2) {
        cpl_msg_error(cpl_func, "odd number of frames in input [%d]", nb_obj);
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Find all distances between offsets and count them */
    dist   = cpl_calloc(nb_obj*nb_obj, sizeof(double));
    dcount = cpl_calloc(nb_obj*nb_obj, sizeof(int));
    for (i=0; i<nb_obj; i++) {
         for (j=0; j<nb_obj; j++) {
             dist[i+j*nb_obj] = sqrt(SQR(offsets_x[i] - offsets_x[j]) +
                     SQR(offsets_y[i] - offsets_y[j]));
         }
    }

    for (i=0; i<nb_obj*nb_obj; i++) {
        for (j=0; j<nb_obj*nb_obj; j++)
            if (fabs(dist[i]-dist[j]) <= NEGLIG_OFF_DIFF) dcount[i] ++;
    }

    /* Find Chop offsets as the nonzero distance with maximal occurrence */
    max_count = 0;
    i_max     = 0;
    j_max     = 0;
    for (i=0; i<nb_obj; i++){
        for (j=0; j<nb_obj; j++) {
            if ((dcount[i+j*nb_obj]>max_count)&&(dist[i+j*nb_obj]>0.5)){
                max_count = dcount[i+j*nb_obj];
                i_max = i;
                j_max = j;
            }
        }
    }

    /* Compute the throw */
    throw_x = offsets_x[j_max] - offsets_x[i_max];
    throw_y = offsets_y[j_max] - offsets_y[i_max];

    cpl_msg_info(cpl_func, "Typical (%d) non-zero throw is (%g, %g) from %d "
                 "(%g, %g) to %d (%g, %g)", max_count, throw_x, throw_y,
                 i_max, offsets_x[i_max], offsets_y[i_max],
                 j_max, offsets_x[j_max], offsets_y[j_max]);

    /* Free */
    cpl_free(dist);
    cpl_free(dcount);

    /* Test the throw */
    if ((throw_x == 0) && (throw_y == 0)) {
        cpl_msg_error(cpl_func, "Throw is equal to 0 - cannot classify");
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Allocate maximal possible nb of chop frames (worst case: abab)*/
    *chop_a = cpl_calloc(nb_obj/2, sizeof(int));
    *chop_b = cpl_calloc(nb_obj/2, sizeof(int));

    k = l = 0;
    for (i=0; i<nb_obj-1; i++) {
        /* Consider the successive pairs of frames */
        /* AB ?  */
        if ((fabs(offsets_x[i]-offsets_x[i+1]+throw_x) < NEGLIG_OFF_DIFF) &&
            (fabs(offsets_y[i]-offsets_y[i+1]+throw_y) < NEGLIG_OFF_DIFF)) {
            (*chop_a)[k++] = i;
            (*chop_b)[l++] = i+1;
        /* BA ? */
        } else if ((fabs(offsets_x[i]-offsets_x[i+1]-throw_x) < 
                    NEGLIG_OFF_DIFF) &&
                (fabs(offsets_y[i]-offsets_y[i+1]-throw_y) < 
                 NEGLIG_OFF_DIFF)) {
            (*chop_b)[l++] = i;
            (*chop_a)[k++] = i+1;
        /* AA or BB ? -> do nothing */
        }
    }

    cpl_msg_info(cpl_func, "Found %d A- and %d B-frames (out of %d)", k, l,
                 nb_obj);

    /* Classify the frames not seen until here as rejected */
    for (i=0; i<nb_obj; i++) {
        classified = 0;
        /* For each frame, check if it has been classified */
        for (j=0; j < nb_obj/2; j++) {
            if ((*chop_a)[j] == i || (*chop_b)[j] == i) classified = 1;
        }
        if (classified == 0) {
            cpl_msg_error(cpl_func, "Frame %d cannot be classified", i+1);
            cpl_free(*chop_a);
            cpl_free(*chop_b);
            *chop_a = *chop_b = NULL;
            return CPL_ERROR_UNSPECIFIED;
        }
    }

    /* There should be as much chop a as chop b frames */
    if (k != l) {
        cpl_free(*chop_a);
        cpl_free(*chop_b);
        *chop_a = *chop_b = NULL;
        return CPL_ERROR_UNSPECIFIED;
    }
    return k;
}

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Subtract from each row the median of the row
  @param    self    the combined image of type float
  @return   0 iff everything is ok
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code isaac_img_jitter_sub_row_median(cpl_image * self)
{
    const int nx    = cpl_image_get_size_x(self);
    const int ny    = cpl_image_get_size_y(self);
    float *   pself = cpl_image_get_data_float(self);
    int       i, j;

    bug_if(self == NULL);
    bug_if(cpl_image_get_type(self) != CPL_TYPE_FLOAT);

 
    for (j = 0; j < ny; j++, pself += nx) {
        cpl_errorstate prestate = cpl_errorstate_get();
        const double median = cpl_image_get_median_window(self, 1, j+1, nx, j+1);

        if (cpl_errorstate_is_equal(prestate)) {
            for (i = 0; i < nx; i++) {
                pself[i] -= (float)median;
            }
        } else {
            cpl_msg_warning(cpl_func, "Could not subtract median from %d "
                            "pixel(s) in row %d/%d (%d bad pixel(s))", nx, 1+j,
                            ny, (int)cpl_image_count_rejected(self));
            cpl_errorstate_set(prestate);
        }
    }

    end_skip;

    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute some QC parameters from the combined image
  @param    combined    the combined image produced
  @return   a newly allocated table with the stars stats or NULL in error case
 */
/*----------------------------------------------------------------------------*/
static cpl_table * isaac_img_jitter_qc(cpl_image * combined)
{
    cpl_vector      *   thresh_vec;
    cpl_apertures   *   aperts;
    int                 nb_objs;
    double              angle;
    double          *   fwhms_x;
    double          *   fwhms_y;
    cpl_table       *   out_tab;
    cpl_bivector    *   iqe;
    int                 nb_good;
    cpl_vector      *   fwhms_good;
    double          *   fwhms_good_data;
    double              f_min, f_max, fr, fx, fy;
    int                 i, j;
    
    /* Initialise */
    double              seeing_min_arcsec = 0.1;
    double              seeing_max_arcsec = 5.0;
    double              seeing_fwhm_var   = 0.2;

    /* Check entries */
    if (combined == NULL) return NULL;
    
    /* Create the vector for the detection thresholds */
    thresh_vec = cpl_vector_new(4);
    cpl_vector_set(thresh_vec, 0, 5.0);
    cpl_vector_set(thresh_vec, 1, 2.0);
    cpl_vector_set(thresh_vec, 2, 1.0);
    cpl_vector_set(thresh_vec, 3, 0.5);

    /* Detect apertures */
    if ((aperts = cpl_apertures_extract(combined, thresh_vec, NULL)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot detect any aperture");
        cpl_vector_delete(thresh_vec);
        return NULL;
    }
    cpl_vector_delete(thresh_vec);

    /* Number of detected objects */
    nb_objs = cpl_apertures_get_size(aperts);
    isaac_img_jitter_config.nbobjs = nb_objs;
    fwhms_x = cpl_malloc(nb_objs * sizeof(double));
    fwhms_y = cpl_malloc(nb_objs * sizeof(double));

    /* Create the output table */
    out_tab = cpl_table_new(nb_objs);
    cpl_table_new_column(out_tab, "POS_X", CPL_TYPE_DOUBLE);
    cpl_table_new_column(out_tab, "POS_Y", CPL_TYPE_DOUBLE);
    cpl_table_new_column(out_tab, "ANGLE", CPL_TYPE_DOUBLE);
    cpl_table_new_column(out_tab, "FWHM_X", CPL_TYPE_DOUBLE);
    cpl_table_new_column(out_tab, "FWHM_Y", CPL_TYPE_DOUBLE);
    cpl_table_new_column(out_tab, "ELLIP", CPL_TYPE_DOUBLE);
    cpl_table_new_column(out_tab, "FLUX", CPL_TYPE_DOUBLE);
    for (i=0; i<nb_objs; i++) {
        /* Fill with the already known information */
        cpl_table_set_double(out_tab, "POS_X", i, 
                cpl_apertures_get_centroid_x(aperts, i+1));
        cpl_table_set_double(out_tab, "POS_Y", i, 
                cpl_apertures_get_centroid_y(aperts, i+1));
        cpl_table_set_double(out_tab, "FLUX", i, 
                cpl_apertures_get_flux(aperts, i+1));
        /* Compute the FWHM informations */
        if ((iqe = cpl_image_iqe(combined, 
                (int)cpl_apertures_get_centroid_x(aperts, i+1) - 10,
                (int)cpl_apertures_get_centroid_y(aperts, i+1) - 10,
                (int)cpl_apertures_get_centroid_x(aperts, i+1) + 10,
                (int)cpl_apertures_get_centroid_y(aperts, i+1) + 10))==NULL){
            cpl_error_reset();
            cpl_msg_warning(cpl_func, "Cannot get FWHM for obj at pos %g %g",
                    cpl_apertures_get_centroid_x(aperts, i+1),
                    cpl_apertures_get_centroid_y(aperts, i+1));
            fwhms_x[i] = -1.0;
            fwhms_y[i] = -1.0;
            angle = 0.0;
        } else {
            fwhms_x[i] = cpl_vector_get(cpl_bivector_get_x(iqe), 2);
            fwhms_y[i] = cpl_vector_get(cpl_bivector_get_x(iqe), 3);
            angle = cpl_vector_get(cpl_bivector_get_x(iqe), 4);
            cpl_bivector_delete(iqe);
        }
        cpl_table_set_double(out_tab, "ANGLE",  i, angle);
        cpl_table_set_double(out_tab, "FWHM_X", i, fwhms_x[i]);
        cpl_table_set_double(out_tab, "FWHM_Y", i, fwhms_y[i]);
    if (fwhms_x[i] == 0.0) {
        cpl_msg_warning(cpl_func, "Infinite ELLIP for obj at pos %g %g",
                    cpl_apertures_get_centroid_x(aperts, i+1),
                    cpl_apertures_get_centroid_y(aperts, i+1));
        cpl_table_set_double(out_tab, "ELLIP",  i, DBL_MAX);
    } else {
        cpl_table_set_double(out_tab, "ELLIP",  i,
                 fwhms_y[i] / fwhms_x[i]);
    }
    }
    cpl_apertures_delete(aperts);

    /* Get the number of good values */
    nb_good = 0;
    for (i=0; i<nb_objs; i++) {
        if ((fwhms_x[i] > 0.0) && (fwhms_y[i] > 0.0)) nb_good++;
    }
    if (nb_good == 0) {
        cpl_table_delete(out_tab);
        cpl_free(fwhms_x);
        cpl_free(fwhms_y);
        return NULL;
    }
    
    /* Get the good values */
    fwhms_good = cpl_vector_new(nb_good);
    fwhms_good_data = cpl_vector_get_data(fwhms_good);
    j=0;
    for (i=0; i<nb_objs; i++) {
        if ((fwhms_x[i] > 0.0) && (fwhms_y[i] > 0.0)) {
            fwhms_good_data[j] = (fwhms_x[i]+fwhms_y[i])/2.0;
            j++;
        }
    }
   
    /* Compute the fwhm */
    if (nb_good < 3) {
        /* Too few values to compute the median */
        isaac_img_jitter_config.fwhm_pix = fwhms_good_data[0];
    } else {
        /* Compute the median */
        isaac_img_jitter_config.fwhm_pix = cpl_vector_get_median_const(fwhms_good);
    }
    isaac_img_jitter_config.fwhm_arcsec = 
        isaac_img_jitter_config.fwhm_pix * isaac_img_jitter_config.pixscale;

    /* Compute the mode of the FWHMs */
    if (nb_good > 5) {
        isaac_img_jitter_config.fwhm_mode=isaac_img_jitter_get_mode(fwhms_good);
        isaac_img_jitter_config.fwhm_mode *= isaac_img_jitter_config.pixscale;
    }
    cpl_vector_delete(fwhms_good);
    
    /* IQ is the median of the (fwhm_x+fwhm_y/2) of the good stars */
    /* Compute f_min and f_max */
    f_min = seeing_min_arcsec / isaac_img_jitter_config.pixscale;
    f_max = seeing_max_arcsec / isaac_img_jitter_config.pixscale;

    /* Get the number of good values */
    nb_good = 0;
    for (i=0; i<nb_objs; i++) {
        fx = fwhms_x[i];
        fy = fwhms_y[i];
        fr = 2.0 * fabs(fx-fy) / (fx+fy);
        if ((fx > f_min) && (fx < f_max) && (fy > f_min) && (fy < f_max) &&
            (fr < seeing_fwhm_var)) nb_good++;
    }
    if (nb_good == 0) {
        cpl_table_delete(out_tab);
        cpl_free(fwhms_x);
        cpl_free(fwhms_y);
        return NULL;
    }

    /* Get the good values */
    fwhms_good = cpl_vector_new(nb_good);
    fwhms_good_data = cpl_vector_get_data(fwhms_good);
    j=0;
    for (i=0; i<nb_objs; i++) {
        fx = fwhms_x[i];
        fy = fwhms_y[i];
        fr = 2.0 * fabs(fx-fy) / (fx+fy);
        if ((fx > f_min) && (fx < f_max) && (fy > f_min) && (fy < f_max) &&
            (fr < seeing_fwhm_var)) {
            fwhms_good_data[j] = (fx + fy)/2.0;
            j++;
        }
    }
    cpl_free(fwhms_x);
    cpl_free(fwhms_y);
    
    /* Compute the fwhm */
    if (nb_good < 3) {
        /* Too few values to compute the median */
        isaac_img_jitter_config.iq = fwhms_good_data[0];
    } else {
        /* Compute the median */
        isaac_img_jitter_config.iq = cpl_vector_get_median_const(fwhms_good);
    }
    cpl_vector_delete(fwhms_good);
    isaac_img_jitter_config.iq *= isaac_img_jitter_config.pixscale;

    isaac_img_jitter_config.angle_med = cpl_table_get_column_median(out_tab,
                                    "ANGLE");
    isaac_img_jitter_config.ellip_med = cpl_table_get_column_median(out_tab,
                                    "ELLIP");
    return out_tab;
}
        
/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the histogram mode
  @param    vec         the vector 
  @return   the mode or -1 in error case
 */
/*----------------------------------------------------------------------------*/
static double isaac_img_jitter_get_mode(cpl_vector * vec)
{
    int                 nb;
    int                 nbins;
    double              min, max;
    double              bin_size;
    cpl_bivector    *   hist;
    cpl_vector      *   hist_x;
    cpl_vector      *   hist_y;
    double              cur_val;
    int                 cur_bin;
    double              max_val;
    int                 max_bin;
    double              mode;
    int                 i;

    /* Test entries  */
    if (vec == NULL) return -1.0;
    
    /* Initialise */
    nb = cpl_vector_get_size(vec);

    /* Create the histogram */
    nbins = 10;
    min = cpl_vector_get_min(vec);
    max = cpl_vector_get_max(vec);
    bin_size = (max-min)/nbins;
    hist = cpl_bivector_new(nbins);
    hist_x = cpl_bivector_get_x(hist);
    hist_y = cpl_bivector_get_y(hist);
    cpl_vector_fill(hist_x, 0.0);
    cpl_vector_fill(hist_y, 0.0);
    for (i=0; i<nbins; i++) {
        cpl_vector_set(hist_x, i, min + i * bin_size);
    }
    for (i=0; i<nb; i++) {
        cur_val = cpl_vector_get(vec, i);
        cur_bin = (int)((cur_val - min) / bin_size);
        if (cur_bin >= nbins) cur_bin -= 1.0;
        cur_val = cpl_vector_get(hist_y, cur_bin);
        cur_val += 1.0;
        cpl_vector_set(hist_y, cur_bin, cur_val);
    }
    
    /* Get the mode of the histogram */
    max_val = cpl_vector_get(hist_y, 0);
    max_bin = 0;
    for (i=0; i<nbins; i++) {
        cur_val = cpl_vector_get(hist_y, i);
        if (cur_val > max_val) {
            max_val = cur_val;
            max_bin = i;
        }
    }
    mode = cpl_vector_get(hist_x, max_bin);
    cpl_bivector_delete(hist);
    return mode;
}
        
/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Save the jitter recipe products on disk
  @param    set         the input frame set
  @param    combined    the combined image produced
  @param    contrib     the contribution map
  @param    objs_stats  the table with the detected objects statistics or NULL
  @param    sky_bg      the vector with the sky background values or NULL
  @param    parlist     the input list of parameters
  @return   0 iff everything is ok
 */
/*----------------------------------------------------------------------------*/
static
cpl_error_code isaac_img_jitter_save(cpl_frameset            * set,
                                     const cpl_image         * combined,
                                     const cpl_image         * contrib,
                                     const cpl_table         * objs_stats,
                                     const cpl_vector        * sky_bg,
                                     const cpl_parameterlist * parlist)
{
    const cpl_frame  * ref_frame
        = irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW);
    const char       * filename = cpl_frame_get_filename(ref_frame);
    const int          sz_skybg = sky_bg ? cpl_vector_get_size(sky_bg) : 0;
    cpl_propertylist * plist = NULL;
    cpl_propertylist * qclist = cpl_propertylist_new();
    const char       * wcs_reg = "(CRVAL|CRPIX|CTYPE)[0-9]+|(CD)[0-9]+_[0-9]+";
    cpl_table        * sky_bg_tab = NULL;
    int                i;

    bug_if(0);
    bug_if(combined   == NULL);
    bug_if(contrib    == NULL);
    bug_if(parlist    == NULL);

    /* Get the QC params in qclist */
    if (isaac_img_jitter_config.chopping == 0) {
        const char * key = "ESO QC BACKGD INSTMAG";
        const double pscale   = isaac_img_jitter_config.pixscale;
        const double dit      = isaac_img_jitter_config.dit;
        const double bg_mean  = cpl_vector_get_mean(sky_bg);
        const double bg_stdev = sz_skybg < 2 ? 0.0
            : cpl_vector_get_stdev(sky_bg);


        bug_if(cpl_propertylist_append_double(qclist, "ESO QC BACKGD MEAN",
                                              bg_mean));
        bug_if(cpl_propertylist_append_double(qclist, "ESO QC BACKGD STDEV",
                                              bg_stdev));

        if (pscale * dit != 0.0 && bg_mean / (pscale * pscale * dit) > 0.0) {
            const double bg_instmag = -2.5 * log10(bg_mean/(pscale*pscale*dit));

            bug_if(cpl_propertylist_append_double(qclist, key, bg_instmag));
        } else {
            cpl_msg_warning(cpl_func, "Could not compute %s: dit=%g, pscale=%g, "
                            "bg_mean=%g", key, dit, pscale, bg_mean);
        }
    }

    cpl_propertylist_append_int(qclist, "ESO QC NBOBJS",
                                isaac_img_jitter_config.nbobjs);
    cpl_propertylist_append_double(qclist, "ESO QC IQ",
                                   isaac_img_jitter_config.iq);
    cpl_propertylist_append_double(qclist, "ESO QC FWHM PIX",
                                   isaac_img_jitter_config.fwhm_pix);
    cpl_propertylist_append_double(qclist, "ESO QC FWHM ARCSEC",
                                   isaac_img_jitter_config.fwhm_arcsec);
    cpl_propertylist_append_double(qclist, "ESO QC FWHM MODE",
                                   isaac_img_jitter_config.fwhm_mode);
    cpl_propertylist_append_int(qclist, "ESO QC NB_OBJ_F",
                                isaac_img_jitter_config.nb_obj_frames);
    cpl_propertylist_append_int(qclist, "ESO QC NB_SKY_F",
                                isaac_img_jitter_config.nb_sky_frames);
    cpl_propertylist_append_int(qclist, "ESO QC NB_REJ_F",
                                isaac_img_jitter_config.nb_rej_frames);
    cpl_propertylist_append_double(qclist, "ESO QC ANGLE MED",
                                   isaac_img_jitter_config.angle_med);
    cpl_propertylist_append_double(qclist, "ESO QC ELLIP MED",
                                   isaac_img_jitter_config.ellip_med);
    bug_if(0);

    /* Get FITS header from reference file */
    plist = cpl_propertylist_load(filename, 0);
    skip_if(plist == NULL);

    /* Load the WCS keys */
    bug_if(cpl_propertylist_copy_property_regexp(qclist, plist, wcs_reg, 0));

    /* Get the keywords for the paf file */
    (void)cpl_propertylist_erase_regexp(plist,
                                        "^(ARCFILE|MJD-OBS|INSTRUME"
                                        "|ESO TPL ID|ESO TPL NEXP"
                                        "|ESO DPR CATG"
                                        "|ESO DPR TECH|ESO DPR TYPE"
                                        "|DATE-OBS|ESO OBS ID"
                                        "|ESO INS PIXSCALE)$", 1);

    /* Write the image */
    skip_if(irplib_dfs_save_image(set, parlist, set, combined,
                                  CPL_BPP_IEEE_FLOAT, RECIPE_STRING,
                                  ISAAC_IMG_JITTER_COMB, qclist, NULL,
                                  PACKAGE "/" PACKAGE_VERSION,
                                  RECIPE_STRING  CPL_DFS_FITS));
    (void)cpl_propertylist_erase_regexp(qclist, wcs_reg, 0);

    /* Append the contribution map */
    skip_if (cpl_image_save(contrib, RECIPE_STRING CPL_DFS_FITS,
                            CPL_BPP_16_UNSIGNED, NULL, CPL_IO_EXTEND));

    if (sky_bg != NULL) {
        /* Write the FITS background table */

        /* First create the table from the vector */
        sky_bg_tab = cpl_table_new(sz_skybg);

        cpl_table_new_column(sky_bg_tab, "SKY_BG", CPL_TYPE_DOUBLE);

        for (i = 0; i < sz_skybg; i++) {
            cpl_table_set_double(sky_bg_tab, "SKY_BG", i, 
                                 cpl_vector_get(sky_bg, i));
        } 

        skip_if(irplib_dfs_save_table(set, parlist, set, sky_bg_tab, NULL,
                                      RECIPE_STRING, ISAAC_IMG_JITTER_BG,
                                      qclist, NULL, PACKAGE "/" PACKAGE_VERSION,
                                      RECIPE_STRING "_bg" CPL_DFS_FITS));
        cpl_table_delete(sky_bg_tab);
        sky_bg_tab = NULL;
    }

    /* Write the FITS table */
    if (objs_stats) {
        skip_if(irplib_dfs_save_table(set, parlist, set, objs_stats, NULL,
                                      RECIPE_STRING, ISAAC_IMG_JITTER_STARS,
                                      qclist, NULL, PACKAGE "/" PACKAGE_VERSION,
                                      RECIPE_STRING "_stars" CPL_DFS_FITS));
    }

    bug_if(cpl_propertylist_append(plist, qclist));
    cpl_propertylist_empty(qclist);

    /* PRO.CATG */
    cpl_propertylist_update_string(plist, CPL_DFS_PRO_CATG,
                                   ISAAC_IMG_JITTER_COMB);

    /* Save the PAF file */
    skip_if (cpl_dfs_save_paf("ISAAC", RECIPE_STRING, plist,
                             RECIPE_STRING CPL_DFS_PAF));

    end_skip;

    cpl_propertylist_delete(plist);
    cpl_propertylist_delete(qclist);
    cpl_table_delete(sky_bg_tab);

    return cpl_error_get_code();
}
