/* $Id: isaac_spc_arc.c,v 1.58 2013-03-12 08:06:48 llundin Exp $
 *
 * This file is part of the ISAAC Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2013-03-12 08:06:48 $
 * $Revision: 1.58 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <string.h>
#include <math.h>
#include <float.h>
#include <cpl.h>

#include "irplib_plugin.h"
#include "irplib_utils.h"
#include "irplib_distortion.h"

#include "isaac_utils.h"
#include "isaac_wavelength.h"
#include "isaac_physicalmodel.h"
#include "isaac_pfits.h"
#include "isaac_dfs.h"

/*-----------------------------------------------------------------------------
                            Define
 -----------------------------------------------------------------------------*/

#define RECIPE_STRING "isaac_spc_arc"

#define ISAAC_ARC_SATURATION    20000

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int isaac_spc_arc_reduce_sw(cpl_frameset *, const cpl_parameterlist *, 
        const char *, const char *, cpl_frameset *);
static int isaac_spc_arc_reduce_lw(cpl_frameset *, const cpl_parameterlist *, 
        const char *, const char *, cpl_frameset *);
static cpl_table * isaac_spc_arc_compute(const cpl_image *, const char *,
        const char *, const char *, const char *, cpl_table **, cpl_image **);
static int isaac_spc_arc_save(const cpl_table *, const cpl_table *,
        const cpl_image *, const char *, cpl_frameset *,
        const cpl_parameterlist *,  cpl_frameset *);
static int isaac_spc_arc_compare(const cpl_frame *, const cpl_frame *); 
static int * isaac_spc_arc_find_lamps(cpl_frameset *);
static int isaac_is_xenon_lamp_active(const cpl_propertylist *);
static int isaac_is_argon_lamp_active(const cpl_propertylist *);


cpl_recipe_define(isaac_spc_arc, ISAAC_BINARY_VERSION,
                  "Lars Lundin", PACKAGE_BUGREPORT, "2002, 2003, 2008", 
                  "ISAAC Spectro arc recipe",
                  RECIPE_STRING " -- ISAAC Spectro arc recipe\n"
                  "The files listed in the Set Of Frames (sof-file) " 
                  "must be tagged:\n"
                  "raw-file.fits "ISAAC_SPC_ARC_RAW" or\n"
                  "xe-cat.fits   "ISAAC_CALPRO_XE_CAT" or\n"
                  "ar-cat.fits   "ISAAC_CALPRO_AR_CAT"\n");

 
/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs */
    int         rej_left;
    int         rej_right;
    int         rej_bottom;
    int         rej_top;
    int         sub_dark;
    int         arc_max_width;
    int         max_offset;
    double      arc_kappa;
    int         out_corr;
    /* Outputs */
    int         set_nb;
    int         pair_nb;
    int         nb_saturated;
    int         arm;
    char        resol;
    double      disprel_cc;
    int         disprel_clines;
    int         disprel_dlines;
    double      disprel_rms;
    double      disprel_offset;
    double      fwhm_med;
    int         fwhm_good;
} isaac_spc_arc_config;

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Fill the recipe parameterlist
  @param    self  The parameterlist
  @return   0 iff everything is ok

 */
/*----------------------------------------------------------------------------*/
static
cpl_error_code isaac_spc_arc_fill_parameterlist(cpl_parameterlist * self)
{
    const char * context = PACKAGE "." RECIPE_STRING;
    cpl_error_code err;

    cpl_ensure_code(self, CPL_ERROR_NULL_INPUT);

    /* Fill the parameters list */

    /* --rejected */
    err = irplib_parameterlist_set_string(self, PACKAGE, RECIPE_STRING,
                                          "rejected", "-1,-1,100,100", "rej",
                                          context,
                                          "left right bottom top rejections");
    cpl_ensure_code(!err, err);

    /* --subdark */
    err = irplib_parameterlist_set_bool(self, PACKAGE, RECIPE_STRING,
                                        "subdark", CPL_FALSE, NULL, context,
                                        "Flag to subtract the dark");
    cpl_ensure_code(!err, err);

    /* --arc_max_w */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING,
                                       "arc_max_w", 33, NULL, context,
                                       "Maximum arc width allowed in pixels");
    cpl_ensure_code(!err, err);

    /* --max_offset */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING,
                                       "max_offset", 50, NULL, context,
                                       "Maximum offset from the physical "
                                       "model allowed in pixels");
    cpl_ensure_code(!err, err);

    /* --arc_kappa */
    err = irplib_parameterlist_set_double(self, PACKAGE, RECIPE_STRING,
                                       "arc_kappa", 0.33, NULL, context,
                                       "kappa for the threshold used for "
                                       "arcs detection");
    cpl_ensure_code(!err, err);

    /* --out_corr */
    err = irplib_parameterlist_set_bool(self, PACKAGE, RECIPE_STRING,
                                        "out_corr", CPL_FALSE, NULL, context,
                                        "Flag to output corrected images");
    cpl_ensure_code(!err, err);

    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part is implemented here 
  @param    parlist     the parameters list
  @param    framelist   the frames list
  @return   0 iff everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_spc_arc(cpl_frameset            * framelist,
                         const cpl_parameterlist * parlist)
{
    const char          *   sval;
    cpl_size            *   labels = NULL;
    cpl_size                nlabels = 0;
    const char          *   xe;
    const char          *   ar;
    cpl_propertylist    *   plist = NULL;
    cpl_frameset        *   arcframes = NULL;
    cpl_frameset        *   arc_one = NULL;
    int                     i;
    
    /* Retrieve input parameters */
    /* Rejection parameters */
    sval = irplib_parameterlist_get_string(parlist, PACKAGE, RECIPE_STRING,
                                           "rejected");
    skip_if(sval == NULL);
    error_if (sscanf(sval, "%d,%d,%d,%d",
                     &isaac_spc_arc_config.rej_left,
                     &isaac_spc_arc_config.rej_right,
                     &isaac_spc_arc_config.rej_bottom,
                     &isaac_spc_arc_config.rej_top) != 4,
              CPL_ERROR_DATA_NOT_FOUND, "Parameter not in format %s: %s",
              "%d,%d,%d,%d", sval);

    /* Output corrected images */
    isaac_spc_arc_config.out_corr
        = irplib_parameterlist_get_bool(parlist, PACKAGE, RECIPE_STRING,
                                        "out_corr");
    /* Dark subtraction */
    isaac_spc_arc_config.sub_dark
        = irplib_parameterlist_get_bool(parlist, PACKAGE, RECIPE_STRING,
                                        "subdark");

    /* Arc max width in pixels */
    isaac_spc_arc_config.arc_max_width
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "arc_max_w");

    /* Max offset in pixels */
    isaac_spc_arc_config.max_offset
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "max_offset");
    /* Arc detection kappa */
    isaac_spc_arc_config.arc_kappa
        = irplib_parameterlist_get_double(parlist, PACKAGE, RECIPE_STRING,
                                          "arc_kappa");

    /* Identify the RAW and CALIB frames in the input frameset */
    skip_if (isaac_dfs_set_groups(framelist));

    /* Retrieve raw frames */
    arcframes = isaac_extract_frameset(framelist, ISAAC_SPC_ARC_RAW);
    error_if (arcframes == NULL, CPL_ERROR_DATA_NOT_FOUND,
              "No frames are tagged %s", ISAAC_SPC_ARC_RAW);
        
    /* Retrieve calibration data */
    xe = isaac_extract_filename(framelist, ISAAC_CALPRO_XE_CAT);
    error_if (xe == NULL, CPL_ERROR_DATA_NOT_FOUND,
              "Xe catalogue not found");

    ar = isaac_extract_filename(framelist, ISAAC_CALPRO_AR_CAT);
    error_if (ar == NULL, CPL_ERROR_DATA_NOT_FOUND,
              "Ar catalogue not found");

    /* Labelise all input frames */
    labels = cpl_frameset_labelise(arcframes, isaac_spc_arc_compare, &nlabels);
    any_if ("Could not labelise input frames (nlabels=%d)", (int)nlabels);
   
    /* Extract settings and reduce each of them */
    for (i=0; i < nlabels; i++) {
        const cpl_frame * cur_frame;
        const char      * filename;
        int error;

        /* Reduce data set nb i */
        cpl_msg_info(cpl_func, "Reducing data set %d of %d", i+1, (int)nlabels);
        isaac_spc_arc_config.set_nb = i+1;
        cpl_frameset_delete(arc_one);
        arc_one = cpl_frameset_extract(arcframes, labels, i);

        /* Get the arm used (SW or LW) */
        cur_frame = cpl_frameset_get_position(arc_one, 0);
        filename = cpl_frame_get_filename(cur_frame);
        cpl_propertylist_delete(plist);
        plist = cpl_propertylist_load(filename, 0);
        sval = isaac_pfits_get_arm(plist);
        any_if ("Could not get the arm from %s in set %d of %d", filename,
                i+1, (int)nlabels);

        if (sval[0] == 'S') {
            isaac_spc_arc_config.arm = 1;
        } else if (sval[0] == 'L') {
            isaac_spc_arc_config.arm = 2;
        } else {
            isaac_spc_arc_config.arm = 0;
            error_if(1, CPL_ERROR_UNSUPPORTED_MODE,
                     "Unsupported arm in %s in set %d of %d: %s", filename,
                     i+1, (int)nlabels, sval);
        }

        /* Get the resolution */
        sval = isaac_pfits_get_resolution(plist);
        any_if ("Could not get the resolution from %s in set %d of %d",
                filename, i+1, (int)nlabels);

        if (sval[0]=='L') isaac_spc_arc_config.resol = 'L';
        else if (sval[0]=='M') isaac_spc_arc_config.resol = 'M';
        else isaac_spc_arc_config.resol = 'U';
        cpl_propertylist_empty(plist);

        error = isaac_spc_arc_config.arm == 1 ?
            /* SW mode */
            isaac_spc_arc_reduce_sw(arc_one, parlist, ar, xe, framelist) :
            /* LW mode */
            isaac_spc_arc_reduce_lw(arc_one, parlist, ar, xe, framelist);

        error_if(error, cpl_error_get_code(), 
                 "Could not reduce set %d of %d", i+1, (int)nlabels);
    }
    
    end_skip;

    cpl_frameset_delete(arc_one);
    cpl_frameset_delete(arcframes);
    cpl_free(labels);
    cpl_propertylist_delete(plist);

    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part for SW mode 
  @param    arcframes   the frames list with arc frames
  @param    parlist     the input list of parameters
  @param    ar          the ar lines catalog
  @param    xe          the xe lines catalog
  @param    set_tot     the input frame set
  @return   0 if ok, -1 in error case
 */
/*----------------------------------------------------------------------------*/
static int isaac_spc_arc_reduce_sw(
        cpl_frameset        *   arcframes,
        const cpl_parameterlist * parlist,
        const char          *   ar,
        const char          *   xe,
        cpl_frameset        *   set_tot)
{
    int                 *   lamps;
    int                     nframes;
    char                    lines_table[16];
    cpl_frame           *   cur_frame;
    const char          *   cur_fname;
    cpl_image           *   xenon;
    cpl_image           *   argon;
    cpl_image           *   xe_ar;
    cpl_image           *   dark;
    cpl_image           *   to_compute;
    cpl_table           *   arcs_fwhm;
    cpl_image           *   out_corr;
    cpl_table           *   out_table;
    int                     i;

    /* Initialise */
    nframes = cpl_frameset_get_size(arcframes);
    isaac_spc_arc_config.pair_nb = 0;
    to_compute = NULL;
    
    /* Look for each file which lamp is activated */
    /* lamps[i]==0   --->   frame i : lamps off */
    /* lamps[i]==1   --->   frame i : Xenon lamp on */
    /* lamps[i]==2   --->   frame i : Argon lamp on */
    /* lamps[i]==3   --->   frame i : Arg and Xe lamps on  */
    if ((lamps=isaac_spc_arc_find_lamps(arcframes)) == NULL) {
        cpl_msg_error(cpl_func, "in finding the activated lamps");
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Get the images */
    xenon = argon = xe_ar = dark = NULL;
    for (i=0; i<nframes; i++) {
        cur_frame = cpl_frameset_get_position(arcframes, i);
        cur_fname = cpl_frame_get_filename(cur_frame);
        if ((lamps[i] == 0) && (dark == NULL)) {
            cpl_msg_info(cpl_func, "Dark image: [%s]", cur_fname);
            dark = cpl_image_load(cur_fname, CPL_TYPE_FLOAT, 0, 0);
        } else if ((lamps[i] == 1) && (xenon == NULL)) {
            cpl_msg_info(cpl_func, "Xenon lamp: [%s]", cur_fname);
            xenon = cpl_image_load(cur_fname, CPL_TYPE_FLOAT, 0, 0);
        } else if ((lamps[i] == 2) && (argon == NULL)) {
            cpl_msg_info(cpl_func, "Argon lamp: [%s]", cur_fname);
            argon = cpl_image_load(cur_fname, CPL_TYPE_FLOAT, 0, 0);
        } else if ((lamps[i] == 3) && (xe_ar == NULL)) {
            cpl_msg_info(cpl_func, "Xenon+Argon lamp: [%s]", cur_fname);
            xe_ar = cpl_image_load(cur_fname, CPL_TYPE_FLOAT, 0, 0);
        }
    }
    cpl_free(lamps);
   
    /* All lamps are switched off */
    if (!xenon && !argon && !xe_ar) {
        cpl_msg_error(cpl_func, "neither xenon nor argon lamp activated");
        if (dark) cpl_image_delete(dark);
        return CPL_ERROR_UNSPECIFIED;
    }
   
    /* Get the first file name */
    cur_frame = cpl_frameset_get_position(arcframes, 0);
    cur_fname = cpl_frame_get_filename(cur_frame);
    
    /* Reduction in LR  */
    if (isaac_spc_arc_config.resol == 'L') {
        cpl_msg_info(cpl_func, "Low resolution");
         
        /* Xenon lines */
        if (xenon) {
            if (dark) to_compute = cpl_image_subtract_create(xenon, dark);
            else to_compute = cpl_image_duplicate(xenon);
            cpl_image_delete(xenon);
            strcpy(lines_table, "Xe");

            /* Apply the reduction */
            cpl_msg_info(cpl_func, "Apply the reduction");
            isaac_spc_arc_config.pair_nb ++;
            cpl_msg_indent_more();
            if ((out_table = isaac_spc_arc_compute(to_compute, cur_fname, 
                            lines_table, ar, xe, &arcs_fwhm, 
                            &out_corr)) == NULL) {
                cpl_msg_error(cpl_func, "arc reduction computation failed");
                cpl_image_delete(to_compute);
            } else {
                /* Save the products */
                cpl_image_delete(to_compute);
                cpl_msg_info(cpl_func, "Save the products");
                isaac_spc_arc_save(out_table, arcs_fwhm, out_corr, lines_table, 
                        arcframes, parlist, set_tot);
                cpl_table_delete(out_table);
                if (arcs_fwhm) cpl_table_delete(arcs_fwhm);
                if (out_corr) cpl_image_delete(out_corr);
            }
            cpl_msg_indent_less();
        }

        /* Argon lines */
        if (argon) {
            if (dark) to_compute = cpl_image_subtract_create(argon, dark);
            else to_compute = cpl_image_duplicate(argon);
            cpl_image_delete(argon);
            strcpy(lines_table, "Ar");

            /* Apply the reduction */
            cpl_msg_info(cpl_func, "Apply the reduction");
            isaac_spc_arc_config.pair_nb ++;
            cpl_msg_indent_more();
            if ((out_table = isaac_spc_arc_compute(to_compute, cur_fname, 
                            lines_table, ar, xe, &arcs_fwhm, 
                            &out_corr)) == NULL) {
                cpl_msg_error(cpl_func, "arc reduction computation failed");
                cpl_image_delete(to_compute);
            } else {
                /* Save the products */
                cpl_image_delete(to_compute);
                cpl_msg_info(cpl_func, "Save the products");
                isaac_spc_arc_save(out_table, arcs_fwhm, out_corr, lines_table, 
                        arcframes, parlist, set_tot);
                cpl_table_delete(out_table);
                if (arcs_fwhm) cpl_table_delete(arcs_fwhm);
                if (out_corr) cpl_image_delete(out_corr);
            }
            cpl_msg_indent_less();
        }

        /* Xenon+Argon lines */
        if (xe_ar) {
            if (dark) to_compute = cpl_image_subtract_create(xe_ar, dark);
            else to_compute = cpl_image_duplicate(xe_ar);
            cpl_image_delete(xe_ar);
            strcpy(lines_table, "Xe+Ar");

            /* Apply the reduction */
            cpl_msg_info(cpl_func, "Apply the reduction");
            isaac_spc_arc_config.pair_nb ++;
            cpl_msg_indent_more();
            if ((out_table = isaac_spc_arc_compute(to_compute, cur_fname, 
                            lines_table, ar, xe, &arcs_fwhm, 
                            &out_corr)) == NULL) {
                cpl_msg_error(cpl_func, "arc reduction computation failed");
                cpl_image_delete(to_compute);
            } else {
                /* Save the products */
                cpl_image_delete(to_compute);
                cpl_msg_info(cpl_func, "Save the products");
                isaac_spc_arc_save(out_table, arcs_fwhm, out_corr, lines_table, 
                        arcframes, parlist, set_tot);
                cpl_table_delete(out_table);
                if (arcs_fwhm) cpl_table_delete(arcs_fwhm);
                if (out_corr) cpl_image_delete(out_corr);
            }
            cpl_msg_indent_less();
        }
        
    } else if (isaac_spc_arc_config.resol == 'M') {
        cpl_msg_info(cpl_func, "Medium resolution");
        /* Get the input image */
        if (xenon && argon) {
            if (dark) {
                cpl_image_subtract(xenon, dark);
                cpl_image_subtract(argon, dark);
                to_compute = cpl_image_add_create(xenon, argon);
            } else to_compute = cpl_image_add_create(xenon, argon);
            cpl_image_delete(xenon);
            cpl_image_delete(argon);
            strcpy(lines_table, "Xe+Ar");
        } else if (xenon) {
            if (dark) to_compute = cpl_image_subtract_create(xenon, dark);
            else to_compute = cpl_image_duplicate(xenon);
            cpl_image_delete(xenon);
            strcpy(lines_table, "Xe");
        } else if (argon) {
            if (dark) to_compute = cpl_image_subtract_create(argon, dark);
            else to_compute = cpl_image_duplicate(argon);
            cpl_image_delete(argon);
            strcpy(lines_table, "Ar");
        } else if (xe_ar) {
            if (dark) to_compute = cpl_image_subtract_create(xe_ar, dark);
            else to_compute = cpl_image_duplicate(xe_ar);
            cpl_image_delete(xe_ar);
            strcpy(lines_table, "Xe+Ar");
        }

        /* Apply the reduction */
        cpl_msg_info(cpl_func, "Apply the reduction");
        isaac_spc_arc_config.pair_nb ++;
        cpl_msg_indent_more();
        if ((out_table = isaac_spc_arc_compute(to_compute, cur_fname, 
                        lines_table, ar, xe, &arcs_fwhm, 
                        &out_corr)) == NULL) {
            cpl_msg_error(cpl_func, "arc reduction computation failed");
            cpl_image_delete(to_compute);
        } else {
            /* Save the products */
            cpl_image_delete(to_compute);
            cpl_msg_info(cpl_func, "Save the products");
            isaac_spc_arc_save(out_table, arcs_fwhm, out_corr, lines_table, 
                    arcframes, parlist, set_tot);
            cpl_table_delete(out_table);
            if (arcs_fwhm) cpl_table_delete(arcs_fwhm);
            if (out_corr) cpl_image_delete(out_corr);
        }
        cpl_msg_indent_less();
    }

    if (dark) cpl_image_delete(dark);
    
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part for LW mode 
  @param    arcframes   the frames list with arc frames
  @param    parlist     the input list of parameters
  @param    ar          the ar lines catalog
  @param    xe          the xe lines catalog
  @param    set_tot     the input frame set
  @return   0 if ok, -1 in error case
 */
/*----------------------------------------------------------------------------*/
static int isaac_spc_arc_reduce_lw(
        cpl_frameset        *   arcframes,
        const cpl_parameterlist * parlist,
        const char          *   ar,
        const char          *   xe,
        cpl_frameset        *   set_tot)
{
    int                 *   lamps;
    int                     nframes;
    char                    lines_table[16];
    cpl_frame           *   cur_frame;
    const char          *   cur_fname;
    cpl_image           *   to_compute;
    cpl_image           *   dark;
    cpl_table           *   arcs_fwhm;
    cpl_image           *   out_corr;
    cpl_table           *   out_table;
    cpl_boolean             did_reduce = CPL_FALSE;
    int                     i;

    /* Initialise */
    nframes = cpl_frameset_get_size(arcframes);
    
     /* The number of frames must be even */
    if (nframes % 2) {
        cpl_msg_error(cpl_func, "An even nb of frames expected in input %d", 
                nframes);
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Look for each file which lamp is activated */
    /* lamps[i]==0   --->   frame i : lamps off */
    /* lamps[i]==1   --->   frame i : Xenon lamp on */
    /* lamps[i]==2   --->   frame i : Argon lamp on */
    /* lamps[i]==3   --->   frame i : Arg and Xe lamps on  */
    if ((lamps=isaac_spc_arc_find_lamps(arcframes)) == NULL) {
        cpl_msg_error(cpl_func, "in finding the activated lamps");
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Loop on the pairs */
    for (i=0; i<nframes/2; i++) {
        isaac_spc_arc_config.pair_nb = i+1;
        /* Load the image */
        cur_frame = cpl_frameset_get_position(arcframes, 2*i);
        cur_fname = cpl_frame_get_filename(cur_frame);
        if ((to_compute = cpl_image_load(cur_fname,CPL_TYPE_FLOAT,0,0))==NULL){
            cpl_msg_error(cpl_func, "Could not load frame %s", cur_fname);
            cpl_free(lamps);
        }
        cpl_msg_info(cpl_func, "Pair %d: Lamp and dark identification", i+1);
        switch (lamps[2*i]) {
            /* Xenon lamp */
            case 1:
                strcpy(lines_table, "Xe");
                cpl_msg_info(cpl_func, "Xenon lamp: [%s]", cur_fname);
                break;
            /* Argon lamp */
            case 2:
                strcpy(lines_table, "Ar");
                cpl_msg_info(cpl_func, "Argon lamp: [%s]", cur_fname);
                break;
            /* Xenon+Argon lamp */
            case 3:
                strcpy(lines_table, "Xe+Ar");
                cpl_msg_info(cpl_func, "Xenon+Argon lamp: [%s]", cur_fname);
                break;
            default:
                cpl_image_delete(to_compute);
                cpl_msg_info(cpl_func, "Lamps are off. Next pair...");
                continue;
        }
        
        /* Get the dark if present */
        if (lamps[2*i+1] == 0) {
            cur_frame = cpl_frameset_get_position(arcframes, 2*i+1);
            cur_fname = cpl_frame_get_filename(cur_frame);
            dark = cpl_image_load(cur_fname, CPL_TYPE_FLOAT, 0, 0);
            cpl_image_subtract(to_compute, dark);
            cpl_image_delete(dark);
            cpl_msg_info(cpl_func, "Dark frame : [%s]", cur_fname);
        } else {
            cpl_msg_info(cpl_func, "No dark frame");
        }

        /* Get the file name */
        cur_frame = cpl_frameset_get_position(arcframes, 2*i);
        cur_fname = cpl_frame_get_filename(cur_frame);
        
        /* Apply the reduction */
        cpl_msg_info(cpl_func, "Apply the reduction");
        cpl_msg_indent_more();
        if ((out_table = isaac_spc_arc_compute(to_compute, cur_fname, 
                        lines_table, ar, xe, &arcs_fwhm, 
                        &out_corr)) == NULL) {
            cpl_msg_error(cpl_func, "arc reduction computation failed");
            cpl_image_delete(to_compute);
        } else {
            /* Save the products */
            cpl_image_delete(to_compute);
            cpl_msg_info(cpl_func, "Save the products");
            isaac_spc_arc_save(out_table, arcs_fwhm, out_corr, lines_table, 
                    arcframes, parlist, set_tot);
            cpl_table_delete(out_table);
            if (arcs_fwhm) cpl_table_delete(arcs_fwhm);
            if (out_corr) cpl_image_delete(out_corr);
            if (!cpl_error_get_code()) did_reduce = CPL_TRUE;
        }
        cpl_msg_indent_less();
    }

    /* Free and return */
    cpl_free(lamps);
    cpl_ensure_code(did_reduce, CPL_ERROR_ILLEGAL_INPUT);

    return cpl_error_set_where(cpl_func); /* Propagate error, if any */
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Determine the distortion, correct, and wl calibrate
  @param    in          input image
  @param    in_name     input file name
  @param    lines_table catalog
  @param    ar          the ar lines catalog
  @param    xe          the xe lines catalog
  @param    arcs_fwhm   table with the computed arcs FHWM
  @param    out_corr    the distortion corrected image
  @return   the result table with 3 columns with the 2d poly and the
  last column with the  1D polynomial describing the dispersion
 */
/*----------------------------------------------------------------------------*/
static cpl_table * isaac_spc_arc_compute(
        const cpl_image     *   in,
        const char          *   in_name,
        const char          *   lines_table,
        const char          *   ar,
        const char          *   xe,
        cpl_table           **  arcs_fwhm,
        cpl_image           **  out_corr)
{
    cpl_table           *   out_tab;
    int                     xmin, ymin, xmax, ymax;
    int                     nx, ny;
    cpl_mask            *   saturation_map;
    double                  slit_width;
    int                     order;
    cpl_polynomial      *   distor_poly;
    cpl_polynomial      *   poly2d_id;
    cpl_vector          *   profile;
    cpl_image           *   in_corr;
    cpl_apertures       *   arcs;
    double              *   phdisprel;
    computed_disprel    *   disprel;
    cpl_size                power[2];
    double                  fwhm_x, fwhm_y;
    cpl_vector          *   fwhm_valid;
    int                     i;

    /* Initialise */
    nx = cpl_image_get_size_x(in);
    ny = cpl_image_get_size_y(in);
    *arcs_fwhm = NULL;
    *out_corr = NULL;

    /* Define zone */
    xmin = 1;
    xmax = nx;
    ymin = isaac_spc_arc_config.rej_bottom + 1;
    ymax = ny - isaac_spc_arc_config.rej_top;

    /* Compute the number of saturated pixels */
    saturation_map = cpl_mask_threshold_image_create(in, ISAAC_ARC_SATURATION, 
            DBL_MAX);
    if (saturation_map != NULL) {
        isaac_spc_arc_config.nb_saturated = cpl_mask_count(saturation_map);
        cpl_mask_delete(saturation_map);
    } else {
        isaac_spc_arc_config.nb_saturated = 0;
    }

    /* Distortion estimation */
    cpl_msg_info(cpl_func, "Estimate the distortion");
    cpl_msg_indent_more();
    if ((distor_poly = irplib_distortion_estimate(in, xmin, ymin, xmax, ymax, 
                    isaac_spc_arc_config.sub_dark, ISAAC_ARC_SATURATION,
                    isaac_spc_arc_config.arc_max_width, 
                    isaac_spc_arc_config.arc_kappa, 2, &arcs)) == NULL) {
        cpl_msg_error(cpl_func, "Could not estimage distortion");
        cpl_msg_indent_less();
        return NULL;
    }
    cpl_msg_indent_less();
                    
    /* Distortion correction */
    cpl_msg_info(cpl_func, "Correct the distortion");
    cpl_msg_indent_more();
    in_corr = cpl_image_duplicate(in);

    /* Create the identity polynomial */
    poly2d_id = cpl_polynomial_new(2);
    power[0] = 0; power[1] = 1;
    cpl_polynomial_set_coeff(poly2d_id, power, 1.0);
    
    /* Create the kernel */
    profile = cpl_vector_new(CPL_KERNEL_DEF_SAMPLES);
    cpl_vector_fill_kernel_profile(profile, CPL_KERNEL_DEFAULT,
            CPL_KERNEL_DEF_WIDTH);

    /* Apply the distortion */
    if (cpl_image_warp_polynomial(in_corr, in, distor_poly,
            poly2d_id, profile, CPL_KERNEL_DEF_WIDTH, profile,
            CPL_KERNEL_DEF_WIDTH) != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "Could not correct the distortion");
        cpl_image_delete(in_corr);
        cpl_polynomial_delete(poly2d_id);
        cpl_polynomial_delete(distor_poly);
        cpl_vector_delete(profile);
        cpl_msg_indent_less();
        return NULL;
    }
    cpl_polynomial_delete(poly2d_id);
    cpl_vector_delete(profile);
    cpl_msg_indent_less();

    /* Compute the FWHM of the detected arcs */
    cpl_msg_info(cpl_func, "Compute the FWHM of the detected arcs");
    /* Create the fwhms table */
    *arcs_fwhm = cpl_table_new(cpl_apertures_get_size(arcs));
    cpl_table_new_column(*arcs_fwhm, "XPOS", CPL_TYPE_DOUBLE);
    cpl_table_new_column(*arcs_fwhm, "FWHM", CPL_TYPE_DOUBLE);
    cpl_table_new_column(*arcs_fwhm, "FLUX", CPL_TYPE_DOUBLE);
    /* Compute the fwhms */
    isaac_spc_arc_config.fwhm_good = 0;
    for (i=0; i<cpl_apertures_get_size(arcs); i++) {
        cpl_table_set_double(*arcs_fwhm, "XPOS", i,
                cpl_apertures_get_centroid_x(arcs, i+1));
        cpl_table_set_double(*arcs_fwhm, "FLUX", i,
                cpl_apertures_get_flux(arcs, i+1));
        if (cpl_image_get_fwhm(in_corr, cpl_apertures_get_centroid_x(arcs, i+1),
                512, &fwhm_x, &fwhm_y) != CPL_ERROR_NONE) {
            cpl_msg_warning(cpl_func, "Could not get the FWHM");
            cpl_error_reset();
        }
        cpl_table_set_double(*arcs_fwhm, "FWHM", i, fwhm_x);
        if (fwhm_x > 0) isaac_spc_arc_config.fwhm_good ++;
    }
    cpl_apertures_delete(arcs);
    
    /* Compute the median of the good fwhms */
    if (isaac_spc_arc_config.fwhm_good > 0) {
        fwhm_valid = cpl_vector_new(isaac_spc_arc_config.fwhm_good);
        isaac_spc_arc_config.fwhm_good = 0;
        for (i=0; i<cpl_table_get_nrow(*arcs_fwhm); i++) {
            fwhm_x = cpl_table_get_double(*arcs_fwhm, "FWHM", i, NULL);
            if (fwhm_x > 0) {
                cpl_vector_set(fwhm_valid, isaac_spc_arc_config.fwhm_good, 
                        fwhm_x);
                isaac_spc_arc_config.fwhm_good ++;
            }
        }
        isaac_spc_arc_config.fwhm_med = cpl_vector_get_median_const(fwhm_valid);
        cpl_vector_delete(fwhm_valid);
    } else isaac_spc_arc_config.fwhm_med = 0.0;
    
    /* Wavelength calibration */
 
    /* Compute the slit_width */
    if ((slit_width = isaac_get_slitwidth(in_name)) == -1) {
        cpl_msg_error(cpl_func, "Could not get the slit width");
        cpl_polynomial_delete(distor_poly);
        cpl_image_delete(in_corr);
        cpl_table_delete(*arcs_fwhm);
        *arcs_fwhm = NULL;
        return NULL;
    }

    /* Get the order */
    if ((order = isaac_find_order(in_name)) == -1) {
        cpl_msg_error(cpl_func, "Could not get the order");
        cpl_polynomial_delete(distor_poly);
        cpl_image_delete(in_corr);
        cpl_table_delete(*arcs_fwhm);
        *arcs_fwhm = NULL;
        return NULL;
    }

    /* First estimation using a physical model */
    cpl_msg_info(cpl_func, "Compute the physical model");
    cpl_msg_indent_more();
    if ((phdisprel = isaac_get_disprel_estimate(in_name, 3)) == NULL) {
        cpl_msg_error(cpl_func, "Could not estimate the dispersion relation");
        cpl_polynomial_delete(distor_poly);
        cpl_image_delete(in_corr);
        cpl_table_delete(*arcs_fwhm);
        *arcs_fwhm = NULL;
        cpl_msg_indent_less();
        return NULL;
    }
    cpl_msg_info(cpl_func, "f(x)=%g + %g*x + %g*x^2 + %g*x^3",
            phdisprel[0], phdisprel[1], phdisprel[2], phdisprel[3]);
    cpl_msg_indent_less();

    /* Compute the dispersion relation */
    cpl_msg_info(cpl_func, "Compute the dispersion relation");
    cpl_msg_indent_more();
    if ((disprel = spectro_compute_disprel(in_corr, 
                    isaac_spc_arc_config.rej_bottom,
                    isaac_spc_arc_config.rej_top,
                    isaac_spc_arc_config.rej_left,
                    isaac_spc_arc_config.rej_right,
                    isaac_spc_arc_config.max_offset,
                    isaac_has_thermal(in_name) > 0,
                    lines_table, NULL, ar, xe, slit_width, order, 
                    (int)(cpl_msg_get_level() == CPL_MSG_DEBUG), 
                    phdisprel)) == NULL) {
        cpl_msg_error(cpl_func, "Could not compute the dispersion relation");
        cpl_polynomial_delete(distor_poly);
        cpl_image_delete(in_corr);
        cpl_free(phdisprel);
        cpl_table_delete(*arcs_fwhm);
        *arcs_fwhm = NULL;
        cpl_msg_indent_less();
        return NULL;
    }
    if (isaac_spc_arc_config.out_corr) {
        *out_corr = in_corr;
        in_corr = NULL;
    } else {
        cpl_image_delete(in_corr);
    }
    cpl_free(phdisprel);
    cpl_msg_info(cpl_func, "Cross correlation factor: %g", disprel->cc);
    cpl_msg_info(cpl_func, "f(x)=%g + %g*x + %g*x^2 + %g*x^3",
            disprel->poly[0], disprel->poly[1], disprel->poly[2],
            disprel->poly[3]);
    cpl_msg_indent_less();
    
    /* Create the output table and store the relevant information */
    out_tab = cpl_table_new(6);
    
    /* Fill distortion polynomial */
    cpl_table_new_column(out_tab, "Degree_of_x", CPL_TYPE_INT);
    cpl_table_new_column(out_tab, "Degree_of_y", CPL_TYPE_INT);
    cpl_table_new_column(out_tab, "poly2d_coef", CPL_TYPE_DOUBLE);
    power[0] = 0; power[1] = 0;
    cpl_table_set_int(out_tab, "Degree_of_x", 0, power[0]);
    cpl_table_set_int(out_tab, "Degree_of_y", 0, power[1]);
    cpl_table_set_double(out_tab, "poly2d_coef", 0,
            cpl_polynomial_get_coeff(distor_poly, power));
    power[0] = 1; power[1] = 0;
    cpl_table_set_int(out_tab, "Degree_of_x", 1, power[0]);
    cpl_table_set_int(out_tab, "Degree_of_y", 1, power[1]);
    cpl_table_set_double(out_tab, "poly2d_coef", 1,
            cpl_polynomial_get_coeff(distor_poly, power));
    power[0] = 0; power[1] = 1;
    cpl_table_set_int(out_tab, "Degree_of_x", 2, power[0]);
    cpl_table_set_int(out_tab, "Degree_of_y", 2, power[1]);
    cpl_table_set_double(out_tab, "poly2d_coef", 2,
            cpl_polynomial_get_coeff(distor_poly, power));
    power[0] = 1; power[1] = 1;
    cpl_table_set_int(out_tab, "Degree_of_x", 3, power[0]);
    cpl_table_set_int(out_tab, "Degree_of_y", 3, power[1]);
    cpl_table_set_double(out_tab, "poly2d_coef", 3,
            cpl_polynomial_get_coeff(distor_poly, power));
    power[0] = 2; power[1] = 0;
    cpl_table_set_int(out_tab, "Degree_of_x", 4, power[0]);
    cpl_table_set_int(out_tab, "Degree_of_y", 4, power[1]);
    cpl_table_set_double(out_tab, "poly2d_coef", 4,
            cpl_polynomial_get_coeff(distor_poly, power));
    power[0] = 0; power[1] = 2;
    cpl_table_set_int(out_tab, "Degree_of_x", 5, power[0]);
    cpl_table_set_int(out_tab, "Degree_of_y", 5, power[1]);
    cpl_table_set_double(out_tab, "poly2d_coef", 5,
            cpl_polynomial_get_coeff(distor_poly, power));

    /* Fill dispersion relation */
    cpl_table_new_column(out_tab, "WL_coefficients", CPL_TYPE_DOUBLE);
    cpl_table_set_double(out_tab, "WL_coefficients", 0, disprel->poly[0]);
    cpl_table_set_double(out_tab, "WL_coefficients", 1, disprel->poly[1]);
    cpl_table_set_double(out_tab, "WL_coefficients", 2, disprel->poly[2]);
    cpl_table_set_double(out_tab, "WL_coefficients", 3, disprel->poly[3]);
    cpl_table_set_double(out_tab, "WL_coefficients", 4, 0.0);
    cpl_table_set_double(out_tab, "WL_coefficients", 5, 0.0);
    isaac_spc_arc_config.disprel_cc     = disprel->cc;
    isaac_spc_arc_config.disprel_clines = disprel->clines;
    isaac_spc_arc_config.disprel_dlines = disprel->dlines;
    isaac_spc_arc_config.disprel_rms    = disprel->rms;
    isaac_spc_arc_config.disprel_offset = disprel->offset;

    /* Free and return */
    cpl_polynomial_delete(distor_poly);
    if (disprel->poly != NULL) cpl_free(disprel->poly);
    cpl_free(disprel);
    return out_tab;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Save the recipe products on disk
  @param    tab     table to save
  @param    fwhms   table with arcs fwhm values
  @param    corr    distortion corrected image
  @param    lines_table catalog
  @param    set     the frames from the current setting
  @param    parlist     the input list of parameters
  @param    set_tot the input frame set
  @return   0 if everything is ok, -1 otherwise
 */
/*----------------------------------------------------------------------------*/
static int isaac_spc_arc_save(
        const cpl_table     *   tab,
        const cpl_table     *   fwhms,
        const cpl_image     *   corr,
        const char          *   lines_table,
        cpl_frameset        *   set,
        const cpl_parameterlist * parlist,
        cpl_frameset        *   set_tot)
{
    cpl_propertylist    *   plist;
    cpl_propertylist    *   qclist;
    cpl_propertylist    *   paflist;
    const cpl_frame           *   ref_frame;
    const char          *   sval;
    char                    qc_str[128];
    const char          *   procat = isaac_spc_arc_config.arm == 1
        ? ISAAC_SPC_ARC_SW_RES : ISAAC_SPC_ARC_LW_RES;
    char                *   filename;
    int                     i;


    /* Get the QC params in qclist */
    qclist = cpl_propertylist_new();

    /* Get the reference frame */
    ref_frame = irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW);
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(ref_frame),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        cpl_propertylist_delete(qclist);
        return CPL_ERROR_UNSPECIFIED;
    }
    /* Test the status */
    if (cpl_error_get_code()) {
        cpl_propertylist_delete(qclist);
        cpl_propertylist_delete(plist);
        return CPL_ERROR_UNSPECIFIED;
    }
    sval = isaac_pfits_get_filter(plist);
    if (cpl_error_get_code()) cpl_error_reset();
    else cpl_propertylist_append_string(qclist, "ESO QC FILTER OBS", sval);
    cpl_propertylist_delete(plist);
    cpl_propertylist_append_string(qclist, "ESO QC LAMP", lines_table);
    cpl_propertylist_append_double(qclist, "ESO QC DISP XCORR", 
            isaac_spc_arc_config.disprel_cc);
    cpl_propertylist_append_int(qclist, "ESO QC DISP NUMCAT", 
            isaac_spc_arc_config.disprel_clines);
    cpl_propertylist_append_int(qclist, "ESO QC DISP NUMMATCH", 
            isaac_spc_arc_config.disprel_dlines);
    cpl_propertylist_append_double(qclist, "ESO QC DISP STDEV", 
            isaac_spc_arc_config.disprel_rms);
    cpl_propertylist_append_double(qclist, "ESO QC DISP OFFSET", 
            isaac_spc_arc_config.disprel_offset);
    cpl_propertylist_append_double(qclist, "ESO QC DISPCO1", 
            cpl_table_get_double(tab, "WL_coefficients", 0, NULL));
    cpl_propertylist_append_double(qclist, "ESO QC DISPCO2", 
            cpl_table_get_double(tab, "WL_coefficients", 1, NULL));
    cpl_propertylist_append_double(qclist, "ESO QC DISPCO3", 
            cpl_table_get_double(tab, "WL_coefficients", 2, NULL));
    cpl_propertylist_append_double(qclist, "ESO QC DISPCO4", 
            cpl_table_get_double(tab, "WL_coefficients", 3, NULL));
    cpl_propertylist_append_double(qclist, "ESO QC DIST1", 
            cpl_table_get_double(tab, "poly2d_coef", 0, NULL));
    cpl_propertylist_append_double(qclist, "ESO QC DISTX", 
            cpl_table_get_double(tab, "poly2d_coef", 1, NULL));
    cpl_propertylist_append_double(qclist, "ESO QC DISTY", 
            cpl_table_get_double(tab, "poly2d_coef", 2, NULL));
    cpl_propertylist_append_double(qclist, "ESO QC DISTXY", 
            cpl_table_get_double(tab, "poly2d_coef", 3, NULL));
    cpl_propertylist_append_double(qclist, "ESO QC DISTXX", 
            cpl_table_get_double(tab, "poly2d_coef", 4, NULL));
    cpl_propertylist_append_double(qclist, "ESO QC DISTYY", 
            cpl_table_get_double(tab, "poly2d_coef", 5, NULL));
    cpl_propertylist_append_int(qclist, "ESO QC SATUR NBPIX", 
            isaac_spc_arc_config.nb_saturated);
    cpl_propertylist_append_double(qclist, "ESO QC WLEN",
        (double)(cpl_table_get_double(tab, "WL_coefficients", 0, NULL) +
             512*cpl_table_get_double(tab, "WL_coefficients", 1, NULL) +
         512*512*cpl_table_get_double(tab, "WL_coefficients", 2, NULL) +
     512*512*512*cpl_table_get_double(tab, "WL_coefficients", 3, NULL)));
   
    if (fwhms) {
        cpl_propertylist_append_int(qclist, "ESO QC ARCS NUM", 
                cpl_table_get_nrow(fwhms));
        for(i=0; i<cpl_table_get_nrow(fwhms); i++) {
            sprintf(qc_str, "ESO QC ARCS%d XPOS", i+1);
            cpl_propertylist_append_double(qclist, qc_str, 
                    cpl_table_get_double(fwhms, "XPOS", i, NULL));
            sprintf(qc_str, "ESO QC ARCS%d FWHM", i+1);
            cpl_propertylist_append_double(qclist, qc_str, 
                    cpl_table_get_double(fwhms, "FWHM", i, NULL));
            sprintf(qc_str, "ESO QC ARCS%d FLUX", i+1);
            cpl_propertylist_append_double(qclist, qc_str, 
                    cpl_table_get_double(fwhms, "FLUX", i, NULL));
        }
        cpl_propertylist_append_int(qclist, "ESO QC ARCS NUMGOOD", 
                isaac_spc_arc_config.fwhm_good);
        cpl_propertylist_append_double(qclist, "ESO QC FWHM MED", 
                isaac_spc_arc_config.fwhm_med);
    }
   
  
    /* Write the arc table  */
    filename = cpl_sprintf("isaac_spc_arc_set%d_pair%d.fits",
            isaac_spc_arc_config.set_nb, isaac_spc_arc_config.pair_nb);
    irplib_dfs_save_table(set_tot,
                parlist,
                set,
                tab,
                NULL,
                "isaac_spc_arc",
                procat,
                qclist,
                NULL,
                PACKAGE "/" PACKAGE_VERSION,
                filename);
    cpl_free(filename);

    /* Write the corrected image  */
    if (corr) {
        const char * procatc = isaac_spc_arc_config.arm == 1
            ? ISAAC_SPC_ARC_SW_CORR : ISAAC_SPC_ARC_LW_CORR;
        filename = cpl_sprintf("isaac_spc_arc_set%d_pair%d_corr.fits",
                isaac_spc_arc_config.set_nb, isaac_spc_arc_config.pair_nb);
        irplib_dfs_save_image(set_tot,
                parlist,
                set,
                corr,
                CPL_BPP_IEEE_FLOAT,
                "isaac_spc_arc",
                procatc,
                qclist,
                NULL,
                PACKAGE "/" PACKAGE_VERSION,
                filename);
        cpl_free(filename);
    }

    /* Get the reference frame */
    ref_frame = irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW);

    /* Get FITS header from reference file */
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(ref_frame),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        cpl_propertylist_delete(qclist);
        return CPL_ERROR_UNSPECIFIED;
    }
    
    /* Get the keywords for the paf file */
    paflist = cpl_propertylist_new();
    cpl_propertylist_copy_property_regexp(paflist, plist, 
        "^(ARCFILE|MJD-OBS|INSTRUME|ESO TPL ID|ESO TPL NEXP|ESO DPR CATG|"
        "ESO DPR TECH|ESO DPR TYPE|DATE-OBS|ESO INS GRAT NAME|"
        "ESO INS GRAT WLEN|ESO INS GRAT ORDER|ESO INS MODE|"
        "ESO INS OPTI1 ID)$", 0);
    cpl_propertylist_delete(plist);

    /* Copy the QC in paflist */
    cpl_propertylist_copy_property_regexp(paflist, qclist, ".", 0);
    cpl_propertylist_delete(qclist);

    /* PRO.CATG */
    cpl_propertylist_update_string(paflist, CPL_DFS_PRO_CATG, procat);

    /* Save the PAF file */
    filename = cpl_sprintf("isaac_spc_arc_set%d_pair%d.paf",
            isaac_spc_arc_config.set_nb, isaac_spc_arc_config.pair_nb);
    cpl_dfs_save_paf("ISAAC",
            "isaac_spc_arc",
            paflist,
            filename);
    cpl_free(filename);
    cpl_propertylist_delete(paflist);
    return  0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Comparison function to identify different settings
  @param    frame1  first frame 
  @param    frame2  second frame 
  @return   0 if frame1!=frame2, 1 if frame1==frame2, -1 on error
 */
/*----------------------------------------------------------------------------*/
static int isaac_spc_arc_compare(const cpl_frame * frame1, 
                                 const cpl_frame * frame2) 
{
    int                comparison = 1;
    const char       * file1 = cpl_frame_get_filename(frame1);
    const char       * file2 = cpl_frame_get_filename(frame2);
    cpl_propertylist * plist1 = NULL;
    cpl_propertylist * plist2 = NULL;
    const char       * sval1,
                     * sval2;
    double             dval1, dval2;

    bug_if(frame1 == NULL);
    bug_if(frame2 == NULL);

    /* Get property lists */
    plist1 = cpl_propertylist_load(file1, 0);
    any_if("Could not load header from 1st frame %s", file1);

    plist2 = cpl_propertylist_load(file2, 0);
    any_if("Could not load header from 2nd frame %s", file2);

    /* Compare the slit used */
    sval1 = isaac_pfits_get_opti1_id(plist1);
    any_if("Could not get slit from 1st frame %s", file1);

    sval2 = isaac_pfits_get_opti1_id(plist2);
    any_if("Could not get slit from 2nd frame %s", file2);

    if (strcmp(sval1, sval2)) comparison = 0;

    /* Compare the resolution */
    sval1 = isaac_pfits_get_resolution(plist1);
    any_if("Could not get resolution from 1st frame %s", file1);

    sval2 = isaac_pfits_get_resolution(plist2);
    any_if("Could not get resolution from 2nd frame %s", file2);

    if (strcmp(sval1, sval2)) comparison = 0;

    /* Compare the central wavelength */
    dval1 = isaac_pfits_get_wlen(plist1);
    any_if("Could not get wavelength from 1st frame %s", file1);

    dval2 = isaac_pfits_get_wlen(plist2);
    any_if("Could not get wavelength from 2nd frame %s", file2);

    if (fabs(dval1-dval2) > 1e-4) comparison = 0;

    end_skip;

    cpl_propertylist_delete(plist1);
    cpl_propertylist_delete(plist2);

    return cpl_error_get_code() ? -1 : comparison;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    identify the activated lamps in a frames list   
  @param    fset    list of frames
  @return   array: array[i]: 0 if no activated lamps, 1 if xenon lamp is 
  activated, 2 if argon lamp is activated, 3 if both lamps are activated
 */
/*----------------------------------------------------------------------------*/
static int * isaac_spc_arc_find_lamps(cpl_frameset * fset)
{
    int                 *   lamps;
    int                     nframes;
    cpl_frame           *   cur_frame;
    cpl_propertylist    *   plist;
    int                     xenon, argon;
    int                     i;

    /* Check entries */
    if (fset == NULL) return NULL;
    
    /* Initialise */
    nframes = cpl_frameset_get_size(fset);
    xenon = argon = 0;
    
    /* Allocate the output array */
    lamps = cpl_malloc(nframes * sizeof(int));

    for (i=0; i<nframes; i++) {
        cur_frame = cpl_frameset_get_position(fset, i);
        plist=cpl_propertylist_load(cpl_frame_get_filename(cur_frame), 0);
        xenon = isaac_is_xenon_lamp_active(plist);
        argon = isaac_is_argon_lamp_active(plist);
        cpl_propertylist_delete(plist);
        if ((argon != 0 && argon != 1) || (xenon != 0 && xenon != 1)) {
            cpl_msg_error(cpl_func, "Could not check which lamp is on");
            cpl_free(lamps);
            return NULL;
        }
        if (xenon == 1) {
            if (argon == 1)         lamps[i] = 3;
            else if (argon == 0)    lamps[i] = 1;
        } else if (xenon == 0 ) {
            if (argon == 1)         lamps[i] = 2;
            else if (argon == 0)    lamps[i] = 0;
        }
    }

    /* Free and return */
    return lamps;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out if the Argon lamp was active
  @param    plist   the property list
  @return   1 if the lamp is active, 0 if not, -1 in error case
 */
/*----------------------------------------------------------------------------*/
static int isaac_is_argon_lamp_active(const cpl_propertylist * plist)
{
    int             status1, status2;

    /* Test error */
    if (cpl_error_get_code()) return CPL_ERROR_UNSPECIFIED;
    status1 = isaac_pfits_get_lamp1_status(plist);
    if (cpl_error_get_code()) return CPL_ERROR_UNSPECIFIED;
    if (status1 == 1) {
        /* Still has to verify that shutter is open */
        status2 = isaac_pfits_get_calshut_status(plist);
        if (cpl_error_get_code()) {
            cpl_error_reset();
            return 1;
        }
        if (status2 == 1) return 1;
    }
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    find out if the Xenon lamp was active
  @param    plist   the property list
  @return   1 if the lamp is active, 0 if not, -1 in error case
 */
/*----------------------------------------------------------------------------*/
static int isaac_is_xenon_lamp_active(const cpl_propertylist * plist)
{
    int             status1, status2;

    /* Test error */
    if (cpl_error_get_code()) return CPL_ERROR_UNSPECIFIED;
    status1 = isaac_pfits_get_lamp2_status(plist);
    if (cpl_error_get_code()) return CPL_ERROR_UNSPECIFIED;
    if (status1 == 1) {
        /* Still has to verify that shutter is open */
        status2 = isaac_pfits_get_calshut_status(plist);
        if (cpl_error_get_code()) {
            cpl_error_reset();
            return 1;
        }
        if (status2 == 1) return 1;
    }
    return 0;
}


