/* $Id: isaac_spc_jitter.c,v 1.94 2013-03-12 08:06:48 llundin Exp $
 *
 * This file is part of the ISAAC Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2013-03-12 08:06:48 $
 * $Revision: 1.94 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>

#include "irplib_plugin.h"
#include "irplib_utils.h"
#include "irplib_stdstar.h"
#include "irplib_spectrum.h"

#include "isaac_utils.h"
#include "isaac_wavelength.h"
#include "isaac_physicalmodel.h"
#include "isaac_pfits.h"
#include "isaac_dfs.h"

#include <string.h>
#include <math.h>

/*-----------------------------------------------------------------------------
                                Define
 -----------------------------------------------------------------------------*/

#define RECIPE_STRING "isaac_spc_jitter"

#define ISAAC_SPC_JITTER_OFFSET_ERR     10

#define KEYSIZE 512

#define CENT_WL_BAND_Z  0.9
#define CENT_WL_BAND_SZ 1.06
#define CENT_WL_BAND_J  1.25
#define CENT_WL_BAND_H  1.65
#define CENT_WL_BAND_K  2.2
#define CENT_WL_BAND_SL 3.78
#define CENT_WL_BAND_M  4.78

#define ISAAC_MIN(A,B) ((A) < (B) ? (A) : (B))

#define isaac_plot_manpage                                      \
    "The recipe can produce a number of predefined plots. "     \
    "Zero means that none of the plots are produced, while "    \
    "increasing values (e.g. 1 or 2) increases the number "     \
    "of plots produced. If the plotting fails a warning is "    \
    "produced, and the recipe continues. "                      \
    "The default behaviour of the plotting is to use "          \
    "gnuplot (with option -persist). The recipe currently "     \
    "produces 1D-plots using gnuplot commands. The recipe "     \
    "user can control the actual plotting-command used by "     \
    "the recipe to create the plot by setting the "             \
    "environment variable CPL_PLOTTER. Currently, if "          \
    "CPL_PLOTTER "                                              \
    "is set it must contain the string 'gnuplot'. Setting "     \
    "it to 'cat > my_gnuplot_$$.txt' causes a number of "       \
    "ASCII-files to be created, which each produce a plot "     \
    "when given as standard input to gnuplot (e.g. later "      \
    "or on a different computer). A finer control of the "      \
    "plotting options can be obtained by writing an "           \
    "executable script, e.g. my_gnuplot.pl, that "              \
    "executes gnuplot after setting the desired gnuplot "       \
    "options (e.g. set terminal pslatex color) "                \
    "and then setting CPL_PLOTTER to my_gnuplot.pl. "           \
    "The predefined plots include plotting of images. "         \
    "Images can be plotted not only with gnuplot, but also "    \
    "using the pnm format. This is controlled with the "        \
    "environment variable CPL_IMAGER. If CPL_IMAGER "           \
    "is set to a string that does not contain the word "        \
    "gnuplot, the recipe will generate the plot in pnm "        \
    "format. E.g. setting CPL_IMAGER to "                       \
    "'display - &' will produce a gray-scale image "            \
    "using the image viewer display."



/*-----------------------------------------------------------------------------
                            Private Function prototypes
 -----------------------------------------------------------------------------*/

static cpl_image ** isaac_spc_jitter_combine(const cpl_frameset *, const char *, 
        const char *, const char *, const char *);
static cpl_imagelist * isaac_spc_jitter_load(const cpl_frameset *);
static cpl_vector * isaac_spc_jitter_get_offsets(const cpl_frameset *);
static int * isaac_spc_jitter_classif(const cpl_vector *, int *);
static int off_comp(double, double, double);
static cpl_imagelist * isaac_spc_jitter_saa_groups(cpl_imagelist *,
        cpl_vector *, int *, int, cpl_vector **);
static int isaac_spc_jitter_wavecal(const char *, const cpl_image *,
                                    const char *, const cpl_frameset *);
static cpl_imagelist * isaac_spc_jitter_nodded(cpl_imagelist *, cpl_vector *, 
        cpl_vector **);
static cpl_imagelist * isaac_spc_jitter_distor(cpl_imagelist *, const char *, 
        const char *);
static double isaac_spc_jitter_refine_offset(const cpl_image *,
                                             const cpl_image *);
static cpl_table * isaac_spc_jitter_extract(const cpl_image *);
static int isaac_spc_jitter_std(const char *, const char *, const cpl_frame *, 
        cpl_table *, double);
static cpl_error_code isaac_spc_jitter_flat(cpl_imagelist *, const char *);
static cpl_error_code isaac_spc_jitter_save(cpl_frameset *, const cpl_image *,
                                            const cpl_image *,
                                            const cpl_table *,
                                            const cpl_parameterlist *);

cpl_recipe_define(isaac_spc_jitter, ISAAC_BINARY_VERSION,
                  "Lars Lundin", PACKAGE_BUGREPORT, "2002,2003, 2008", 
                  "ISAAC Spectro jitter recipe",
                  RECIPE_STRING " -- ISAAC spectro jitter recipe\n"
                  "The files listed in the Set Of Frames (sof-file) " 
                  "must be tagged:\n"
                  "Observation in nodding mode:\n"
                  "\traw-file.fits "ISAAC_SPC_JITTER_NODOBJ_RAW" or\n"
                  "\traw-file.fits "ISAAC_SPC_JITTER_NODSKY_RAW"\n"
                  "Observation in chopping mode:\n"
                  "\traw-file.fits "ISAAC_SPC_JITTER_CHOP_RAW" or\n"
                  "Calibration (standard star) in nodding mode:\n"
                  "\traw-file.fits "ISAAC_SPC_RESPFUNC_RAW" or\n"
                  "\traw-file.fits "ISAAC_SPC_RESPFUNC_OBJ_RAW" or\n"
                  "\traw-file.fits "ISAAC_SPC_RESPFUNC_SKY_RAW"\n"
                  "\traw-file.fits "ISAAC_SPC_RESPFUNC_FLUX_RAW"\n"
                  "Calibration (standard star) in chopping mode:\n"
                  "\traw-file.fits "ISAAC_SPC_JITTER_CHOP_CAL_RAW"\n"
                  "Calibration files:\n"
                  "\toh.fits "ISAAC_CALPRO_OH_CAT" or\n"
                  "\tflat-file.fits "ISAAC_CALIB_SPFLAT" or\n"
                  "\tflat-file.fits "ISAAC_CALIB_LW_SPFLAT" or\n"
                  "\tarc-file.fits "ISAAC_CALIB_ARC" or\n"
                  "\tarc-file.fits "ISAAC_CALIB_LW_ARC" or\n"
                  "\tstartrace-file.fits "ISAAC_CALIB_STARTRACE" or\n"
                  "\tstartrace-file.fits "ISAAC_CALIB_LW_STARTRACE" or\n"
                  "\tsed-file.fits "ISAAC_CALIB_SED" or\n"
                  "\tstdstars-file.fits "ISAAC_CALIB_STDSTARS"\n");

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs */
    int             arm;
    int             plot;
    int             oddeven;
    /* wavecal_in : 0 for physical model, 1 for sky, 2 for arc table */
    int             wavecal_in;
    int             wavecal_rej_bottom;
    int             wavecal_rej_top;
    int             wavecal_rej_left;
    int             wavecal_rej_right;
    int             max_offset;
    int             saa_refine;
    double          saa_rej_high;
    double          saa_rej_low;
    int             extr_spec_pos;
    int             extr_spec_width;
    int             extr_sky_hi_width;
    int             extr_sky_lo_width;
    int             extr_sky_hi_dist;
    int             extr_sky_lo_dist;
    int             std_mode;
    double          std_magnitude;
    /* Outputs */
    /* wavecal_out : 0 for physical model, 1 for sky, 2 for arc table */
    int             chopping;
    cpl_vector  *   throws;
    cpl_vector  *   intensities;
    char            filter[KEYSIZE];
    char            filter_ref[KEYSIZE];
    char          * starname;
    char          * sptype;
    int             wavecal_out;
    double          wavecal_cc;
    double          wavecal_a0;
    double          wavecal_a1;
    double          wavecal_a2;
    double          wavecal_a3;
} isaac_spc_jitter_config;


/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Fill the recipe parameterlist
  @param    self  The parameterlist
  @return   0 iff everything is ok

 */
/*----------------------------------------------------------------------------*/
static
cpl_error_code isaac_spc_jitter_fill_parameterlist(cpl_parameterlist * self)
{
    const char * context = PACKAGE "." RECIPE_STRING;
    cpl_error_code err;

    cpl_ensure_code(self, CPL_ERROR_NULL_INPUT);

    /* Fill the parameters list */


    /* --oddeven */
    err = irplib_parameterlist_set_bool(self, PACKAGE, RECIPE_STRING,
                                        "oddeven", CPL_FALSE, NULL, context,
                                        "Flag to correct the oddeven column "
                                        "effect");
    cpl_ensure_code(!err, err);

    /* --wavecal */
    err = irplib_parameterlist_set_string(self, PACKAGE, RECIPE_STRING,
                                          "wavecal", "sky", NULL, context,
                                          "Wavelength calibration method: "
                                          "sky, phy or arc");
    cpl_ensure_code(!err, err);

    /* --wavecal_rej */
    err = irplib_parameterlist_set_string(self, PACKAGE, RECIPE_STRING,
                                          "wavecal_rej", "-1,-1,50,50",
                                          "wc_rej", context,
                                          "left right bottom top rejections");
    cpl_ensure_code(!err, err);

    /* --max_offset */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING,
                                       "max_offset", 50, NULL, context,
                                       "Maximum allowed offset from the "
                                       "physical model [pixel]");
    cpl_ensure_code(!err, err);

    /* --saa_refine */
    err = irplib_parameterlist_set_bool(self, PACKAGE, RECIPE_STRING,
                                        "saa_refine", CPL_TRUE, NULL, context,
                                        "Flag to refine the offsets");
    cpl_ensure_code(!err, err);

    /* --saa_rej */
    err = irplib_parameterlist_set_string(self, PACKAGE, RECIPE_STRING,
                                          "saa_rej", "0.1,0.1", NULL, context,
                                          "Low and high rejection fractions "
                                          "(rounded down)");
    cpl_ensure_code(!err, err);

    /* --spec_pos */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING,
                                       "spec_pos", -1, NULL, context,
                                       "Spectrum position");
    cpl_ensure_code(!err, err);

    /* --spec_width */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING,
                                       "spec_width", 10, NULL, context,
                                       "Spectrum width");
    cpl_ensure_code(!err, err);

    /* --sky_hi_width */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING,
                                       "sky_hi_width", 10, NULL, context,
                                       "Sky width above the spectrum");
    cpl_ensure_code(!err, err);

    /* --sky_lo_width */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING,
                                       "sky_lo_width", 10, NULL, context,
                                       "Sky width below the spectrum");
    cpl_ensure_code(!err, err);

    /* --sky_hi_dist */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING,
                                       "sky_hi_dist", -1, NULL, context,
                                       "Sky distance above the spectrum");
    cpl_ensure_code(!err, err);

    /* --sky_lo_dist */
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING,
                                       "sky_lo_dist", -1, NULL, context,
                                       "Sky distance below the spectrum");
    cpl_ensure_code(!err, err);

    /* --plot */ 
    err = irplib_parameterlist_set_bool(self, PACKAGE, RECIPE_STRING,
                                        "display", CPL_FALSE, NULL, context,
                                        "Ignored. Use --plot instead");
    cpl_ensure_code(!err, err);

    /* --plot */ 
    err = irplib_parameterlist_set_int(self, PACKAGE, RECIPE_STRING,
                                       "plot", 0, NULL, context,
                                       isaac_plot_manpage);
    cpl_ensure_code(!err, err);

    /* --std_mode */ 
    err = irplib_parameterlist_set_bool(self, PACKAGE, RECIPE_STRING,
                                        "std_mode", CPL_FALSE, "std", context,
                                        "Flag for standard star");
    cpl_ensure_code(!err, err);

    /* --std_magnitude */
    err = irplib_parameterlist_set_double(self, PACKAGE, RECIPE_STRING,
                                          "std_magnitude", -1.0, "mag",
                                          context, "Standard star magnitude");
    cpl_ensure_code(!err, err);

    return CPL_ERROR_NONE;
}


/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    The recipe data reduction part is implemented here 
  @param    framelist   The list of frames
  @param    parlist     The list of parameters
  @return   0 iff everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_spc_jitter(cpl_frameset            * framelist,
                            const cpl_parameterlist * parlist)
{
    cpl_errorstate prestate = cpl_errorstate_get();
    const char          *   sval;
    const char          *   flat;
    const char          *   arc;
    const char          *   startrace;
    const char          *   sed;
    const char          *   stdstars;
    const char          *   oh;
    const cpl_frame     *   cur_frame;
    cpl_frameset        *   rawframes = NULL;
    cpl_propertylist    *   plist = NULL;
    cpl_image           **  combined = NULL;
    cpl_table           *   extracted = NULL;
    
    bug_if(0);

    /* Initialise */
    isaac_spc_jitter_config.chopping = -1;
    isaac_spc_jitter_config.wavecal_out = -1;
    isaac_spc_jitter_config.wavecal_cc = -1.0;
    isaac_spc_jitter_config.throws = NULL;
    isaac_spc_jitter_config.intensities = NULL;
    isaac_spc_jitter_config.filter[0] = (char)0;
    isaac_spc_jitter_config.filter_ref[0] = (char)0;
    isaac_spc_jitter_config.starname = NULL;
    isaac_spc_jitter_config.sptype = NULL;
    isaac_spc_jitter_config.wavecal_in = -1;
    isaac_spc_jitter_config.arm = -1;

    /* Retrieve input parameters */
    /* --wavecal */
    sval = irplib_parameterlist_get_string(parlist, PACKAGE, RECIPE_STRING,
                                           "wavecal");
    bug_if(sval == NULL);

    if (!strcmp(sval, "phy")) isaac_spc_jitter_config.wavecal_in = 0;
    else if (!strcmp(sval, "sky")) isaac_spc_jitter_config.wavecal_in = 1;
    else if (!strcmp(sval, "arc")) isaac_spc_jitter_config.wavecal_in = 2;

    error_if(isaac_spc_jitter_config.wavecal_in < 0,
             CPL_ERROR_UNSUPPORTED_MODE, "Invalid value for wavecal option: "
             "%s", sval);

    /* Rejection parameters for wavelength calibration*/
    sval = irplib_parameterlist_get_string(parlist, PACKAGE, RECIPE_STRING,
                                           "wavecal_rej");
    bug_if(sval == NULL);

    skip_if (sscanf(sval, "%d,%d,%d,%d",
                    &isaac_spc_jitter_config.wavecal_rej_left,
                    &isaac_spc_jitter_config.wavecal_rej_right,
                    &isaac_spc_jitter_config.wavecal_rej_bottom,
                    &isaac_spc_jitter_config.wavecal_rej_top) != 4);

    /* Max offset in pixels */
    isaac_spc_jitter_config.max_offset
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "max_offset");

    /* Oddeven */
    isaac_spc_jitter_config.oddeven
        = irplib_parameterlist_get_bool(parlist, PACKAGE, RECIPE_STRING,
                                        "oddeven");

    /* Refine of offsets */
    isaac_spc_jitter_config.saa_refine
        = irplib_parameterlist_get_bool(parlist, PACKAGE, RECIPE_STRING,
                                        "saa_refine");
    
    /* Rejection parameters for SAA */
    sval = irplib_parameterlist_get_string(parlist, PACKAGE, RECIPE_STRING,
                                           "saa_rej");
    bug_if(sval == NULL);

    skip_if (sscanf(sval, "%lg,%lg",
                    &isaac_spc_jitter_config.saa_rej_low,
                    &isaac_spc_jitter_config.saa_rej_high) != 2);

    /* --spec_pos */
    isaac_spc_jitter_config.extr_spec_pos
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "spec_pos");
    /* --spec_width */
    isaac_spc_jitter_config.extr_spec_width
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "spec_width");
    /* --sky_hi_width */
    isaac_spc_jitter_config.extr_sky_hi_width
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "sky_hi_width");
    /* --sky_lo_width */
    isaac_spc_jitter_config.extr_sky_lo_width
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "sky_lo_width");
    /* --sky_hi_dist */
    isaac_spc_jitter_config.extr_sky_hi_dist
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "sky_hi_dist");
    /* --sky_lo_dist */
    isaac_spc_jitter_config.extr_sky_lo_dist
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "sky_lo_dist");
    /* --plot */
    isaac_spc_jitter_config.plot
        = irplib_parameterlist_get_int(parlist, PACKAGE, RECIPE_STRING,
                                       "plot");
    /* --std */
    isaac_spc_jitter_config.std_mode
        = irplib_parameterlist_get_bool(parlist, PACKAGE, RECIPE_STRING,
                                       "std_mode");
    /* --mag */
    isaac_spc_jitter_config.std_magnitude
        = irplib_parameterlist_get_double(parlist, PACKAGE, RECIPE_STRING,
                                       "std_magnitude");
    
    /* Identify the RAW and CALIB frames in the input frameset */
    skip_if (isaac_dfs_set_groups(framelist));

    /* Retrieve raw frames */
    if ((rawframes = isaac_extract_frameset(framelist,
                    ISAAC_SPC_JITTER_CHOP_RAW)) != NULL) {
    } else if ((rawframes = isaac_extract_frameset(framelist,
                    ISAAC_SPC_JITTER_CHOP_CAL_RAW)) != NULL) {
    } else if ((rawframes = isaac_extract_frameset(framelist,
                    ISAAC_SPC_JITTER_NODOBJ_RAW)) != NULL) {
    } else if ((rawframes = isaac_extract_frameset(framelist,
                    ISAAC_SPC_RESPFUNC_RAW)) != NULL) {
        isaac_spc_jitter_config.std_mode = 1;
    } else if ((rawframes = isaac_extract_frameset(framelist,
                    ISAAC_SPC_RESPFUNC_OBJ_RAW)) != NULL) {
        isaac_spc_jitter_config.std_mode = 1;
    } else if ((rawframes = isaac_extract_frameset(framelist,
                    ISAAC_SPC_RESPFUNC_FLUX_RAW)) != NULL) {
        isaac_spc_jitter_config.std_mode = 1;
    } else {

        /* These frame types alone are not supported */

        error_if (cpl_frameset_find_const(framelist,
                                          ISAAC_SPC_JITTER_NODSKY_RAW),
                  CPL_ERROR_UNSUPPORTED_MODE, "Cannot reduce "
                  ISAAC_SPC_JITTER_NODSKY_RAW "-frames without object frames");

        error_if (cpl_frameset_find_const(framelist,
                                          ISAAC_SPC_RESPFUNC_SKY_RAW),
                  CPL_ERROR_UNSUPPORTED_MODE, "Cannot reduce "
                  ISAAC_SPC_RESPFUNC_SKY_RAW "-frames without object frames");

    }

    error_if (rawframes == NULL, CPL_ERROR_DATA_NOT_FOUND,
              "Cannot find the raw frames in the input list");

    /* Get the arm used (SW or LW) */
    cur_frame = cpl_frameset_get_position_const(framelist, 0);
    skip_if(cur_frame == NULL);
    plist = cpl_propertylist_load(cpl_frame_get_filename(cur_frame), 0);
    skip_if(plist == NULL);
    sval = isaac_pfits_get_arm(plist);
    error_if (sval == NULL, CPL_ERROR_DATA_NOT_FOUND, "cannot get the arm");

    if (sval[0] == 'S')      isaac_spc_jitter_config.arm = 1;
    else if (sval[0] == 'L') isaac_spc_jitter_config.arm = 2;
    
    error_if (isaac_spc_jitter_config.arm < 0, CPL_ERROR_UNSUPPORTED_MODE,
              "Unsupported arm: %s", sval);

    /* Get the filter */
    sval = isaac_pfits_get_filter(plist);
    if (sval == NULL) {
        cpl_error_reset();
    } else {
        (void)strncpy(isaac_spc_jitter_config.filter, sval, KEYSIZE);
        isaac_spc_jitter_config.filter[KEYSIZE-1] = (char)0;
    }
    cpl_propertylist_empty(plist);

    /* Retrieve calibration data */
    arc = isaac_extract_filename(framelist, ISAAC_CALIB_ARC);
    if (arc == NULL)
        arc = isaac_extract_filename(framelist, ISAAC_CALIB_LW_ARC);
    flat = isaac_extract_filename(framelist, ISAAC_CALIB_SPFLAT);
    if (flat == NULL)
        flat = isaac_extract_filename(framelist, ISAAC_CALIB_LW_SPFLAT);
    startrace = isaac_extract_filename(framelist, ISAAC_CALIB_STARTRACE);
    if (startrace == NULL)
        startrace = isaac_extract_filename(framelist, ISAAC_CALIB_LW_STARTRACE);
    sed = isaac_extract_filename(framelist, ISAAC_CALIB_SED);
    stdstars = isaac_extract_filename(framelist, ISAAC_CALIB_STDSTARS);
    oh = isaac_extract_filename(framelist, ISAAC_CALPRO_OH_CAT);

    /* If arc calibration requested, arc file must be provided */
    error_if (arc == NULL && isaac_spc_jitter_config.wavecal_in == 2,
              CPL_ERROR_DATA_NOT_FOUND, "Missing an arc file for the "
              "arc-based wavelength calibration");

    cpl_msg_info(cpl_func, "Create the combined image");
    combined = isaac_spc_jitter_combine(rawframes, oh, flat, arc,
                                        startrace);
    skip_if (combined == NULL);

    cpl_msg_info(cpl_func, "Extract the spectrum");
    extracted = isaac_spc_jitter_extract(combined[0]);
    if (extracted == NULL || !cpl_errorstate_is_equal(prestate)) {
        cpl_msg_warning(cpl_func, "Cannot extract the spectrum");
        if (!cpl_errorstate_is_equal(prestate)) {
            cpl_errorstate_dump(prestate, CPL_FALSE,
                                irplib_errorstate_dump_warning);
            cpl_errorstate_set(prestate);
        }
    } else if (isaac_spc_jitter_config.std_mode) {
        /* In standard star mode, compute the efficiency and the conversion */
        cpl_msg_info(cpl_func, "Compute flux calibration");
        cur_frame = cpl_frameset_get_position_const(framelist, 0);
        if (isaac_spc_jitter_std(sed, stdstars, cur_frame, extracted, 
                                 isaac_spc_jitter_config.std_magnitude)) {
            cpl_msg_warning(cpl_func, "Cannot compute the flux calibration");
            if (!cpl_errorstate_is_equal(prestate)) {
                cpl_errorstate_dump(prestate, CPL_FALSE,
                                    irplib_errorstate_dump_warning);
                cpl_errorstate_set(prestate);
            }
        }
    }
    
    /* Write the products */
    cpl_msg_info(cpl_func, "Save the products");
    skip_if (isaac_spc_jitter_save(framelist, combined[0], combined[1],
                                   extracted, parlist));

    end_skip;

    if (combined != NULL) {
        cpl_image_delete(combined[0]);
        cpl_image_delete(combined[1]);
        cpl_free(combined);
    }

    cpl_table_delete(extracted);

    cpl_vector_delete(isaac_spc_jitter_config.intensities);
    cpl_vector_delete(isaac_spc_jitter_config.throws);

    cpl_free(isaac_spc_jitter_config.starname);
    cpl_free(isaac_spc_jitter_config.sptype);

    cpl_propertylist_delete(plist);

    cpl_frameset_delete(rawframes);

    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    The image combination routine
  @param    rawframes   the spectrum frames
  @param    oh          the oh lines catalog
  @param    flat        the flat frame or NULL
  @param    arc_dist    the arc frame for distortion or NULL
  @param    startrace   the startrace frame or NULL
  @return   the combined image or NULL
 */
/*----------------------------------------------------------------------------*/
static cpl_image ** isaac_spc_jitter_combine(const cpl_frameset * rawframes,
                                             const char         * oh,
                                             const char         * flat,
                                             const char         * arc,
                                             const char         * startrace)
{
    int                     nilist;
    cpl_imagelist       *   ilist = NULL;
    cpl_vector          *   offsets = NULL;
    int                 *   groups = NULL;
    int                     ngroups = 0; /* Avoid uninit warning */
    cpl_imagelist       *   abba = NULL;
    cpl_vector          *   abba_off = NULL;
    cpl_imagelist       *   nodded = NULL;
    cpl_vector          *   nodded_off_y = NULL;
    double                  throw;
    cpl_vector          *   nodded_off_x = NULL;
    cpl_bivector        *   nodded_offsets = NULL;
    cpl_image           **  combined = NULL;
    int                     nima;
    int                     i;
    
    bug_if (rawframes == NULL);

    /* Load the images */
    cpl_msg_info(cpl_func, "Load the data"); 
    ilist = isaac_spc_jitter_load(rawframes);
    skip_if (ilist == NULL);

    nilist = cpl_imagelist_get_size(ilist);
   
    /* Apply the odd-even correction */
    if (isaac_spc_jitter_config.oddeven) {
        cpl_errorstate prestate = cpl_errorstate_get();
        cpl_imagelist * corrected = cpl_imagelist_new();

        cpl_msg_info(cpl_func, "Apply the odd-even effect correction");
        for (i=0; i < nilist; i++) {
            cpl_image * cur_im
                = isaac_oddeven_correct(cpl_imagelist_get(ilist, i));

            cpl_msg_info(cpl_func, "Correcting frame %d/%d", i+1, nilist);

            if (cur_im == NULL) break;
            (void)cpl_imagelist_set(corrected, cur_im, i);
        }

        if (i == nilist) {
            /* Odd/even correction successful */
            cpl_imagelist_delete(ilist);
            ilist = corrected;
        } else {
            cpl_msg_warning(cpl_func, "Odd/even correction failed for "
                            "image %d/%d", i+1, nilist);
            cpl_imagelist_delete(corrected);
            if (!cpl_errorstate_is_equal(prestate)) {
                cpl_errorstate_dump(prestate, CPL_FALSE, 
                                    irplib_errorstate_dump_warning);
                cpl_errorstate_set(prestate);
            }
        }
    }


    if (isaac_spc_jitter_config.plot > 0) {
        cpl_plot_image("", "", "", cpl_imagelist_get(ilist, 0));
    }

    /* Apply the flatfield */
    if (flat != NULL) {
        cpl_errorstate prestate = cpl_errorstate_get();
        cpl_msg_info(cpl_func, "Apply the flatfield correction"); 
        if (isaac_spc_jitter_flat(ilist, flat)) {
            cpl_msg_warning(cpl_func, "cannot apply the flat field");
            if (!cpl_errorstate_is_equal(prestate)) {
                cpl_errorstate_dump(prestate, CPL_FALSE, 
                                    irplib_errorstate_dump_warning);
                cpl_errorstate_set(prestate);
            }
        }
    }
        
    if (isaac_spc_jitter_config.plot > 0) {
        cpl_plot_image("", "", "", cpl_imagelist_get(ilist, 0));
    }
    /* Get the offsets */
    cpl_msg_info(cpl_func, "Get the %d offsets", nilist);
    offsets = isaac_spc_jitter_get_offsets(rawframes);

    error_if (offsets == NULL, cpl_error_get_code(),
              "Could not get the offsets");

    /* Classify in groups the a and b images sequence */
    cpl_msg_info(cpl_func, "Classify in groups"); 
    groups = isaac_spc_jitter_classif(offsets, &ngroups);
    error_if (groups == NULL, cpl_error_get_code(),
              "Could not classify the data");
  
    /* Shift and add each group to one image */
    cpl_msg_info(cpl_func, "Shift and add each group to one image"); 
    abba = isaac_spc_jitter_saa_groups(ilist, offsets, groups, 
                                       ngroups, &abba_off);
    error_if (abba == NULL, cpl_error_get_code(), "Could not shift and add "
              "groups");

    /* Compute the wavelength calibration */
    cpl_msg_info(cpl_func, "Compute the wavelength calibration"); 
    error_if (isaac_spc_jitter_wavecal(arc, cpl_imagelist_get(abba, 0), oh,
                                       rawframes),
              cpl_error_get_code(), "Wavelength calibration failed");

   /* Create the nodded images */
    cpl_msg_info(cpl_func, "Create the nodded images"); 
    nodded = isaac_spc_jitter_nodded(abba, abba_off, &nodded_off_y);
    cpl_imagelist_delete(abba);
    abba = NULL;

    skip_if (nodded == NULL);

    /* Get the throw offsets from abba_off */
    nima = cpl_imagelist_get_size(nodded);
    isaac_spc_jitter_config.throws = cpl_vector_new(nima);
    if (isaac_spc_jitter_config.chopping == 0) {
        for (i=0; i<nima/2; i++) {
            throw = fabs(   (cpl_vector_get(abba_off, 2*i))-
                            (cpl_vector_get(abba_off, 2*i+1)));
            cpl_vector_set(isaac_spc_jitter_config.throws, 2*i, throw);
            cpl_vector_set(isaac_spc_jitter_config.throws, 2*i+1, throw);
        }
    } else {
        for (i=0; i<nima; i++) {
            throw = fabs(   (cpl_vector_get(abba_off, 2*i))-
                            (cpl_vector_get(abba_off, 2*i+1)));
            cpl_vector_set(isaac_spc_jitter_config.throws, i, throw);
        }
    }
    cpl_vector_delete(abba_off);
    abba_off = NULL;
    
    /* Get the spectra intensities in the nodded images */
    cpl_msg_info(cpl_func, "Compute the spectra intensities");
    cpl_msg_indent_more();
    isaac_spc_jitter_config.intensities = cpl_vector_new(nima);
    for (i=0; i<nima; i++) {
        cpl_errorstate prestate = cpl_errorstate_get();
        cpl_table    * extracted
            = isaac_spc_jitter_extract(cpl_imagelist_get(nodded, i));
        double         intensity;


        if (extracted == NULL || !cpl_errorstate_is_equal(prestate)) {
            cpl_msg_warning(cpl_func,"Cannot extract the spectrum from "
                            "nodded image %d", i+1);
            cpl_errorstate_dump(prestate, CPL_FALSE, 
                                irplib_errorstate_dump_warning);
            cpl_errorstate_set(prestate);
            intensity = -1.0;
        } else {
            intensity = cpl_table_get_column_mean(extracted, 
                    "Extracted_spectrum_value");
            intensity *= cpl_table_get_nrow(extracted);
            cpl_table_delete(extracted);
            cpl_msg_info(cpl_func, "Spectrum intensity %d: %g", i+1,
                         intensity);
        }
        cpl_vector_set(isaac_spc_jitter_config.intensities, i, intensity);
    }
    cpl_msg_indent_less();

    if (arc || startrace) {
        cpl_imagelist * nodded_warped;
        cpl_msg_info(cpl_func, "Correct the distortion on nodded images"); 
        cpl_msg_indent_more();
        nodded_warped = isaac_spc_jitter_distor(nodded, arc, startrace);
        cpl_msg_indent_less();

        skip_if (nodded_warped == NULL);

        cpl_imagelist_delete(nodded);
        nodded = nodded_warped;
        nima = cpl_imagelist_get_size(nodded);
    }


    /* Refine the offsets if requested */
    if (isaac_spc_jitter_config.saa_refine) {
        double * pnodded_off_y = cpl_vector_get_data(nodded_off_y);
        cpl_msg_info(cpl_func, "Refine the %d offsets", nima);
        for (i=0; i < nima; i++) {
            const double new_offset
                = isaac_spc_jitter_refine_offset(cpl_imagelist_get(nodded, 0),
                                                 cpl_imagelist_get(nodded, i));
            if (new_offset > 5000) {
                cpl_msg_debug(cpl_func, "cannot refine the offset - keep %g", 
                              pnodded_off_y[i]);
            } else if (fabs(new_offset-pnodded_off_y[i]) < 
                       ISAAC_SPC_JITTER_OFFSET_ERR) {
                cpl_msg_debug(cpl_func, "refined offset : %g (old was %g)",
                              new_offset, pnodded_off_y[i]);
                pnodded_off_y[i] = new_offset;
            } else { 
                cpl_msg_debug(cpl_func, 
                              "refined offset %g too different - keep %g",
                              new_offset, pnodded_off_y[i]);
            }
        }
    }

    /* Images Shift and add */

    /* Get the offsets in a bivector */
    nodded_off_x = cpl_vector_new(nima);
    bug_if(cpl_vector_fill(nodded_off_x, 0.0));
    nodded_offsets = cpl_bivector_wrap_vectors(nodded_off_x, nodded_off_y);

    if (nima == 1) {

        combined = cpl_malloc(2 * sizeof(cpl_image*)); 
        combined[1] = NULL; /* In case the unset fails */

        combined[0] = cpl_imagelist_unset(nodded, 0);
        bug_if (combined[0] == NULL);

        combined[1] = cpl_image_new(cpl_image_get_size_x(combined[0]),
                                    cpl_image_get_size_y(combined[0]),
                                    CPL_TYPE_INT);
        bug_if (combined[1] == NULL);

        /* Set all pixel values to 1 */
        bug_if(cpl_image_threshold(combined[1], 1.0, 1.0, 1.0, 1.0));

    } else {
        /* FIXME: Round-off instead */
        const int nrejlo = (int)(nima * isaac_spc_jitter_config.saa_rej_low);
        const int nrejhi = (int)(nima * isaac_spc_jitter_config.saa_rej_high);

        cpl_msg_info(cpl_func, "Apply the shift and add on the %d nodded "
                     "frames (Rejecting %d lowermost, %d uppermost)", nima,
                     nrejlo, nrejhi);
        combined = cpl_geom_img_offset_saa(nodded, nodded_offsets, 
                                           CPL_KERNEL_DEFAULT, nrejlo, nrejhi,
                                           CPL_GEOM_FIRST, NULL, NULL);
    }

    skip_if(combined == NULL);

    end_skip;

    cpl_free(groups);
    cpl_imagelist_delete(ilist);
    cpl_vector_delete(offsets);
    cpl_imagelist_delete(abba);
    cpl_vector_delete(abba_off);
    cpl_imagelist_delete(nodded);
    cpl_bivector_unwrap_vectors(nodded_offsets);
    cpl_vector_delete(nodded_off_x);
    cpl_vector_delete(nodded_off_y);

    return combined;
}
 
/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Apply the flatfield
  @param    self    the input images list
  @param    flat    the flat field image
  @return   0 if ok, -1 in error case
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code isaac_spc_jitter_flat(cpl_imagelist * self,
                                            const char    * flat)
{
    cpl_image        * fim = NULL;
    cpl_image        * big_fim = NULL;
    const int          mx = cpl_image_get_size_x(cpl_imagelist_get(self, 0));
    const int          my = cpl_image_get_size_y(cpl_imagelist_get(self, 0));
    int                nx, ny;

    bug_if (self == NULL);
    bug_if (flat == NULL);
   
    /* Load the flat */
    fim = cpl_image_load(flat, CPL_TYPE_UNSPECIFIED, 0, 0);
    error_if(fim == NULL, cpl_error_get_code(), "Could not load the flat "
             "field from %s", flat);
   
    /* Get flat size */
    nx = cpl_image_get_size_x(fim);
    ny = cpl_image_get_size_y(fim);

    error_if(ny != my, CPL_ERROR_INCOMPATIBLE_INPUT, "Flat field Y-size "
             "%d is incompatible with image Y-size %d", ny, my);

    /* Handle the case where nx is smaller */
    if (nx != mx) {
        cpl_propertylist * plist
            = cpl_propertylist_load_regexp(flat, 0, "ESO DET WIN STARTX", 0);
        const int lx = (int)isaac_pfits_get_win_startx(plist);

        cpl_propertylist_delete(plist);
        skip_if (0);

        cpl_msg_info(cpl_func, "Windowing mode: flat between pixels %d "
                     "and %d", lx, ISAAC_MIN(mx, lx + nx - 1));

        if (lx > mx || lx < 1) {
            cpl_msg_warning(cpl_func, "Cannot apply out of bounds flat field, "
                            "lx=%d > mx=%d", lx, mx);
            cpl_image_delete(fim);
            fim = NULL;
        } else {

            /* Create the big flat image */
            big_fim = cpl_image_new(mx, my, cpl_image_get_type(fim));

            /* Set all pixel values to 1 */
            bug_if(cpl_image_threshold(big_fim, 1.0, 1.0, 1.0, 1.0));

            /* Copy the small flat in the big one */
            bug_if (cpl_image_copy(big_fim, fim, lx, 1));

            cpl_image_delete(fim);
            fim = big_fim;
            big_fim = NULL;
        }
    }

    if (fim != NULL) {

        /* Threshold all pixels below the "mid-point" of FLT_MIN and 1, and
           all pixels above the "mid-point" of 1 and FLT_MAX. */
        bug_if(cpl_image_threshold(fim, sqrt(FLT_MIN), sqrt(FLT_MAX),
                                   1.0, 1.0));

        /* Apply the division */
        error_if (cpl_imagelist_divide_image(self, fim),
                  cpl_error_get_code(), "Could not apply the flat field");
    }

    end_skip;

    cpl_image_delete(fim);
    cpl_image_delete(big_fim);

    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Load the imagelist
  @param    rawframes   The input frameset
  @return   The image list or NULL on error
 */
/*----------------------------------------------------------------------------*/
static cpl_imagelist * isaac_spc_jitter_load(const cpl_frameset * rawframes)
{
    /* Load the propertylist of the first frame to get the data type */
    const cpl_frame  * cur_frame = cpl_frameset_get_position_const(rawframes, 0);
    cpl_propertylist * plist
        = cpl_propertylist_load(cpl_frame_get_filename(cur_frame), 0);
    const char          *   sval;
    cpl_imagelist       *   self = NULL;


    skip_if (plist == NULL);

    sval = isaac_pfits_get_frame_type(plist);
    skip_if (sval == NULL);

    if (!strcmp(sval, "INT")) isaac_spc_jitter_config.chopping = 0;
    else if (!strcmp(sval, "CUBE1")) isaac_spc_jitter_config.chopping = 1;
    else {
        error_if(1, CPL_ERROR_UNSUPPORTED_MODE, "Unsupported frame type: %s",
                 sval);
    }

    cpl_propertylist_empty(plist);

    /* Load the data */
    self = cpl_imagelist_load_frameset(rawframes, CPL_TYPE_FLOAT, 1, 0);

    if (isaac_spc_jitter_config.chopping == 1) {
        cpl_imagelist * self_tmp
            = cpl_imagelist_load_frameset(rawframes, CPL_TYPE_FLOAT, 2, 0);
        const cpl_error_code error = cpl_imagelist_subtract(self, self_tmp);

        cpl_imagelist_delete(self_tmp);
        skip_if(error);
    }

    end_skip;

    cpl_propertylist_delete(plist);

    if (cpl_error_get_code()) {
        cpl_imagelist_delete(self);
        self = NULL;
    }

    return self;
}

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Get the offsets
  @param    rawframes   the input rawframes
  @return   the offsets in a vector
 */
/*----------------------------------------------------------------------------*/
static
cpl_vector * isaac_spc_jitter_get_offsets(const cpl_frameset * rawframes)
{
    const int          nraw = cpl_frameset_get_size(rawframes);
    cpl_vector       * offsets = cpl_vector_new(nraw);
    cpl_propertylist * plist = NULL;
    int                i;

    bug_if (rawframes == NULL);
    bug_if (nraw == 0);
    

    /* Get the rawframes Y offsets */
    for (i=0; i < nraw; i++) {
        const cpl_frame * cur_frame
            = cpl_frameset_get_position_const(rawframes, i);
        const char * filename = cpl_frame_get_filename(cur_frame);
        double yoff;

        skip_if(filename == NULL);

        cpl_propertylist_delete(plist);
        plist = cpl_propertylist_load_regexp(filename, 0,
                                             "ESO SEQ CUMOFFSETY", 0);

        any_if("Could not load propertylist from %s", filename);

        yoff = cpl_propertylist_get_double(plist, "ESO SEQ CUMOFFSETY");

        any_if("Could not get the offset from %s", filename);

        bug_if(cpl_vector_set(offsets, i, yoff));
    }

    end_skip;

    cpl_propertylist_delete(plist);

    if (cpl_error_get_code()) {
        cpl_vector_delete(offsets);
        offsets = NULL;
    }

    return offsets;
}
    
/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    The abba classification
  @param    offsets     the images offsets
  @param    pngroups    the computed number of groups (must be even)
  @return   the groups array
 
  The classification is done according the offsets read in header of the 
  rawframes.
  
  The frames are first classified in two different categories A or B depending 
  in which zone their offset is.

  Example
  frame1 - offset: -250.4    ---> cat. A
  frame2 - offset: -120.6    ---> cat. B
  frame3 - offset: -150.6    ---> cat. B
  frame4 - offset: -220.4    ---> cat. A

  These frames are then grouped in cubes according the category order.
  Each group is then averaged in a single image.

  Example 
  frame1 -  cat A    -> group 1  
  frame2 -  cat A    -> group 1
  frame3 -  cat B    -> group 2 
  frame4 -  cat B    -> group 2
  frame5 -  cat B    -> group 3
  frame6 -  cat B    -> group 3
  frame7 -  cat A    -> group 4
  frame8 -  cat A    -> group 4
  frame9 -  cat A    -> group 5
  frame10 - cat A    -> group 5
  frame11 - cat B    -> group 6
  frame12 - cat B    -> group 6
  frame13 - cat B    -> group 6
  frame14 - cat B    -> group 6
 */
/*----------------------------------------------------------------------------*/
static int * isaac_spc_jitter_classif(const cpl_vector * offsets,
                                      int              * pngroups)
{
    const int      nraw   = cpl_vector_get_size(offsets);
    const double * pvect  = cpl_vector_get_data_const(offsets);
    const double   offmin = cpl_vector_get_min(offsets);
    const double   offmax = cpl_vector_get_max(offsets);
    const double   offset_thresh = 0.5 * (offmin + offmax);
    int          * groups = cpl_calloc(nraw, sizeof(int));
    int            last_group;
    int            i = 0;


    bug_if(offsets == NULL);
    bug_if(pngroups == NULL);
    *pngroups = 0;

    /* Separate the offsets in 2 categories */
    error_if (offmin >= offmax, CPL_ERROR_DATA_NOT_FOUND,
              "The %d offset(s) all have the same value: %g", nraw, pvect[0]);

    /* Identify the different A and B groups */

    /* Create a look up table to associate the ith obj with the jth frame */
    while (i < nraw) {
        int j = 0;
        /* Count the number of successive '+' or '-' (j) */
        while ((i+j < nraw) &&
                (!off_comp(pvect[i], pvect[i+j], offset_thresh))) j++;

        if (i+j >= nraw) i = nraw;
        else {
            int l, k = 0;
            /* Check if there are j '-' or '+' (k) */
            while ((i+j+k < nraw)
                    && (!off_comp(pvect[i+j], pvect[i+j+k], offset_thresh))
                    && (k<j)) k++;
            last_group = 1;
            if (i+j+k < nraw) {
                for (l=i+j+k; l<nraw; l++) {
                    if (off_comp(pvect[i+j], pvect[l], offset_thresh)) {
                        last_group = 0;
                        break;
                    }
                }
            }
            if (last_group == 0) {
                for (l=0; l<j; l++) groups[i+l]   = *pngroups + 1;
                for (l=0; l<k; l++) groups[i+j+l] = *pngroups + 2;
                *pngroups += 2;
                i += j+k;
            } else {
                for (l=0; l<j; l++)             groups[i+l]   = *pngroups + 1;
                for (l=0; l<nraw - (i+j); l++)  groups[i+j+l] = *pngroups + 2;
                *pngroups += 2;
                i = nraw;
            }
        }
    }

    /* Nb of groups found should be even */
    error_if (*pngroups & 1, CPL_ERROR_ILLEGAL_INPUT, "Found an odd number "
              "(%d) of groups in the %d offsets", *pngroups, nraw);

    end_skip;

    if (cpl_error_get_code()) {
        cpl_free(groups);
        groups = NULL;
    }

    return groups;
}
    
/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Shift and add groups
  @param    ilist       the input image list
  @param    offsets     the offsets
  @param    groups      the groups ids
  @param    ngroups     the number of groups
  @param    abba_off    the output list offsets
  @return   the new abba image list
 
  The frames in each group are combined together. In principle, there
  should be no offset difference between the frames in the same group,
  so an average should be enough. But to support more general cases
  (generic offset) a shift and add is applied if the offsets differ.

  Example 
  frame1 -  cat A    -> group 1  ---> shift and add 1
  frame2 -  cat A    -> group 1
  frame3 -  cat B    -> group 2  ---> shift and add 2
  frame4 -  cat B    -> group 2
  frame5 -  cat B    -> group 3  ---> shift and add 3
  frame6 -  cat B    -> group 3
  frame7 -  cat A    -> group 4  ---> shift and add 4
  frame8 -  cat A    -> group 4
  frame9 -  cat A    -> group 5  ---> shift and add 5
  frame10 - cat A    -> group 5
  frame11 - cat B    -> group 6  ---> shift and add 6
  frame12 - cat B    -> group 6
  frame13 - cat B    -> group 6
  frame14 - cat B    -> group 6
 */
/*----------------------------------------------------------------------------*/
static cpl_imagelist * isaac_spc_jitter_saa_groups(
        cpl_imagelist   *   ilist,
        cpl_vector      *   offsets,
        int             *   groups,
        int                 ngroups,
        cpl_vector      **  abba_off)
{
    cpl_imagelist       *   abba; 
    cpl_imagelist       *   group_list; 
    cpl_image           *   tmp_ima;
    cpl_image           **  combined;
    cpl_bivector        *   group_off;
    double              *   pgroup_off;
    double              *   poffsets;
    double              *   pabba_off;
    int                     nima;
    int                     saa;
    int                     i, j, k;

    /* Test entries */
    if ((ilist == NULL) || (offsets == NULL) || (groups == NULL)) return NULL;
    
    /* Initialise */
    nima = cpl_imagelist_get_size(ilist);
    poffsets = cpl_vector_get_data(offsets);

    /* Create the output image list */
    abba = cpl_imagelist_new();
    *abba_off = cpl_vector_new(ngroups);
    pabba_off = cpl_vector_get_data(*abba_off);
    
    /* Loop on the groups */
    for (i=0; i<ngroups; i++) {
        /* Initialise */
        saa = 0;
        /* Create the group list of images */
        group_list = cpl_imagelist_new();
        k = 0;
        for (j=0; j<nima; j++) {
            if (i+1 == groups[j]) {
                /* Get the first offset of the group in abba_off */
                if (k==0) pabba_off[i] = poffsets[j];
                /* To know if we need the saa (shift and add) */
                if (fabs(pabba_off[i]-poffsets[j]) > 1e-3) saa = 1;
                /* Copy the images of the group in group_list */
                tmp_ima = cpl_image_duplicate(cpl_imagelist_get(ilist, j));
                cpl_imagelist_set(group_list, tmp_ima, k);
                tmp_ima = NULL;
                k++;
            }
        }

        if (saa) {
            /* Get the offsets of the group in group_off */
            group_off = cpl_bivector_new(k);
            cpl_vector_fill(cpl_bivector_get_x(group_off), 0.0);
            pgroup_off = cpl_bivector_get_y_data(group_off);
            k = 0;
            for (j=0; j<nima; j++) {
                if (i+1 == groups[j]) {
                    pgroup_off[k] = poffsets[j];
                    k++;
                }
            }
            cpl_vector_subtract_scalar(cpl_bivector_get_y(group_off), 
                    pabba_off[i]);
            /* Shift and add */
            cpl_msg_debug(cpl_func, "Apply shift-and-add for group %d", i+1);
            if ((combined = cpl_geom_img_offset_saa(group_list,
                           group_off, CPL_KERNEL_DEFAULT, 0, 0,
                           CPL_GEOM_FIRST, NULL, NULL)) == NULL) {
                cpl_msg_error(cpl_func, "Cannot shift and add group nb %d", i+1);
                cpl_imagelist_delete(group_list);
                cpl_bivector_delete(group_off);
                cpl_imagelist_delete(abba);
                cpl_vector_delete(*abba_off);
                return NULL;
            }
            cpl_bivector_delete(group_off);
            cpl_image_delete(combined[1]);
            cpl_imagelist_set(abba, combined[0], i);
            cpl_free(combined);
        } else {
            /* Averaging */
            cpl_msg_debug(cpl_func, "Apply averaging for group %d", i+1);
            if ((tmp_ima = cpl_imagelist_collapse_create(group_list)) == NULL) {
                cpl_msg_error(cpl_func, "Cannot average group nb %d", i+1);
                cpl_imagelist_delete(group_list);
                cpl_imagelist_delete(abba);
                cpl_vector_delete(*abba_off);
                return NULL;
            }
            cpl_imagelist_set(abba, tmp_ima, i);
        }
        cpl_imagelist_delete(group_list);
    }
    return abba;
}

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    The wavelength calibration
  @param    arc     the arc file 
  @param    ima     the image with the sky lines 
  @param    oh      the oh lines catalog
  @param    raw     the frameset for the header
  @return   int 0 if ok, -1 otherwise
 */
/*----------------------------------------------------------------------------*/
static int isaac_spc_jitter_wavecal(const char         * arc,
                                    const cpl_image    * ima,
                                    const char         * oh,
                                    const cpl_frameset * raw)
{
    cpl_table           *   arc_tab;
    double              *   phdisprel;
    const cpl_frame     *   cur_frame;
    const char          *   cur_fname;
    computed_disprel    *   disprel;
    int                     order;
    double                  slit_width;

    /* Get the wavelength from the arc file */
    if (isaac_spc_jitter_config.wavecal_in == 2) {
        if (arc == NULL) {
            cpl_msg_error(cpl_func, 
                    "Missing arc for the wavelength calibration");
            return CPL_ERROR_UNSPECIFIED;
        }
        cpl_msg_info(cpl_func, "Get the wavelength from the ARC file");
        if ((arc_tab = cpl_table_load(arc, 1, 0)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot load the arc table");
            isaac_spc_jitter_config.wavecal_out = -1;
            return CPL_ERROR_UNSPECIFIED;
        }
        isaac_spc_jitter_config.wavecal_a0 =
            cpl_table_get_double(arc_tab, "WL_coefficients", 0, NULL);
        isaac_spc_jitter_config.wavecal_a1 = 
            cpl_table_get_double(arc_tab, "WL_coefficients", 1, NULL);
        isaac_spc_jitter_config.wavecal_a2 = 
            cpl_table_get_double(arc_tab, "WL_coefficients", 2, NULL);
        isaac_spc_jitter_config.wavecal_a3 = 
            cpl_table_get_double(arc_tab, "WL_coefficients", 3, NULL);
        cpl_table_delete(arc_tab);
        isaac_spc_jitter_config.wavecal_out = 2;
        isaac_spc_jitter_config.wavecal_cc = -1.0;
        return 0;
    }

    /* Get the reference frame */
    cur_frame = cpl_frameset_get_position_const(raw, 0);
    cur_fname = cpl_frame_get_filename(cur_frame);
    
    /* Get the physical model */
    cpl_msg_info(cpl_func, "Compute the physical model");
    cpl_msg_indent_more();
    if ((phdisprel = isaac_get_disprel_estimate(cur_fname, 3)) == NULL) {
        cpl_msg_error(cpl_func, "cannot compute the physical model");
        isaac_spc_jitter_config.wavecal_out = -1;
        cpl_msg_indent_less();
        return CPL_ERROR_UNSPECIFIED;
    }
    cpl_msg_info(cpl_func, "f(x)=%g + %g*x + %g*x^2 + %g*x^3",
            phdisprel[0], phdisprel[1], phdisprel[2], phdisprel[3]);
    isaac_spc_jitter_config.wavecal_a0 = phdisprel[0];
    isaac_spc_jitter_config.wavecal_a1 = phdisprel[1];
    isaac_spc_jitter_config.wavecal_a2 = phdisprel[2];
    isaac_spc_jitter_config.wavecal_a3 = phdisprel[3];
    isaac_spc_jitter_config.wavecal_cc = -1.0;
    isaac_spc_jitter_config.wavecal_out = 0;
    cpl_msg_indent_less();

    /* Compute the wavelength using the sky lines */
    if (isaac_spc_jitter_config.wavecal_in == 1) {
        /* Compute the slit_width */
        if ((slit_width = isaac_get_slitwidth(cur_fname)) == -1) {
            cpl_msg_warning(cpl_func, "cannot get the slit width");
            cpl_free(phdisprel);
            return 0;
        }
        /* Get the order */
        if ((order = isaac_find_order(cur_fname)) == -1) {
            cpl_msg_warning(cpl_func, "cannot get the order");
            cpl_free(phdisprel);
            return 0;
        }
        /* Compute the wavelength */
        cpl_msg_info(cpl_func, "Compute the wavelength with the sky lines");
        cpl_msg_indent_more();
        if ((disprel = spectro_compute_disprel(ima,
                        isaac_spc_jitter_config.wavecal_rej_bottom,
                        isaac_spc_jitter_config.wavecal_rej_top,
                        isaac_spc_jitter_config.wavecal_rej_left,
                        isaac_spc_jitter_config.wavecal_rej_right,
                        isaac_spc_jitter_config.max_offset,
                        isaac_has_thermal(cur_fname) > 0,
                        "oh", oh, NULL, NULL, slit_width, order, 
                        (int)(cpl_msg_get_level() == CPL_MSG_DEBUG), 
                        phdisprel)) == NULL) {
            cpl_msg_error(cpl_func, "cannot compute the dispersion relation");
            cpl_free(phdisprel);
            cpl_msg_indent_less();
            return 0;
        }
        cpl_msg_info(cpl_func, "Cross correlation factor: %g", disprel->cc);
        cpl_msg_info(cpl_func, "f(x)=%g + %g*x + %g*x^2 + %g*x^3",
                disprel->poly[0], disprel->poly[1], disprel->poly[2],
                disprel->poly[3]);
        isaac_spc_jitter_config.wavecal_a0 = disprel->poly[0];
        isaac_spc_jitter_config.wavecal_a1 = disprel->poly[1];
        isaac_spc_jitter_config.wavecal_a2 = disprel->poly[2];
        isaac_spc_jitter_config.wavecal_a3 = disprel->poly[3];
        isaac_spc_jitter_config.wavecal_cc = disprel->cc;
        isaac_spc_jitter_config.wavecal_out = 1;
        if (disprel->poly != NULL) cpl_free(disprel->poly);
        cpl_free(disprel);
        cpl_msg_indent_less();
    }
    cpl_free(phdisprel);
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Create the nodded images
  @param    abba        the abba images list
  @param    abba_off    the abba offsets
  @param    nodded_off  the offsets of the output image list
  @return   the nodded images

  Depending if we are in chopping or non-chopping mode, the frames combination 
  is not exactly the same. In both cases, the frames are reduced by pairs. 
  At this stage, we have an even number of frames.

  In non-chopping mode, each pair (frame n, frame n+1) will generate two 
  nodded frames (frame n-frame n+1) and (frame n+1-frame n).
  Example 
  abba1 - cat A  ---> abba1 - abba2 = nodded1
  abba2 - cat B  ---> abba2 - abba1 = nodded2
  abba3 - cat B  ---> abba3 - abba4 = nodded3
  abba4 - cat A  ---> abba4 - abba3 = nodded4
  abba5 - cat A  ---> abba5 - abba6 = nodded5
  abba6 - cat B  ---> abba6 - abba5 = nodded6
  
  In chopping mode, as the difference is already done by the chopping, the 
  pairs are directly combined.
  Example
  abba1 - cat A ---> (abba1 - abba2)/2 = nodded1
  abba2 - cat B
  abba3 - cat B
  abba4 - cat A ---> (abba4 - abba3)/2 = nodded2
  abba5 - cat A ---> (abba5 - abba6)/2 = nodded3
  abba6 - cat B
 */
/*----------------------------------------------------------------------------*/
static cpl_imagelist * isaac_spc_jitter_nodded(cpl_imagelist   *   abba,
                                               cpl_vector      *   abba_off,
                                               cpl_vector      **  nodded_off)
{
    cpl_imagelist       *   nodded;
    cpl_image           *   tmp_ima;
    int                     nima;
    double              *   pabba_off;
    double              *   pnodded_off;
    double                  ref_off;
    int                     i;

    /* Test entries */
    if ((abba == NULL) || (abba_off == NULL)) return NULL;

    /* Initialise */
    nima = cpl_imagelist_get_size(abba);
    if (nima % 2) {
        cpl_msg_error(cpl_func, "Number of images should be even");
        return NULL;
    }
   
    /* Non-chopping case */
    if (isaac_spc_jitter_config.chopping == 0) {
        /* Create the offsets between the nodded images */
        *nodded_off = cpl_vector_duplicate(abba_off);
        /* The image list to contain the nodded images */
        nodded = cpl_imagelist_new();
        for (i=0; i<(nima/2); i++) {
            /* a-b */
            tmp_ima = cpl_image_duplicate(cpl_imagelist_get(abba, 2*i));
            cpl_image_subtract(tmp_ima, cpl_imagelist_get(abba, 2*i+1));
            cpl_imagelist_set(nodded, tmp_ima, 2*i);
            /* b-a */
            tmp_ima = cpl_image_duplicate(cpl_imagelist_get(abba, 2*i+1));
            cpl_image_subtract(tmp_ima, cpl_imagelist_get(abba, 2*i));
            cpl_imagelist_set(nodded, tmp_ima, 2*i+1);
        }
    /* Chopping case */
    } else if (isaac_spc_jitter_config.chopping == 1) {
        /* Create the offsets between the nodded images */
        *nodded_off = cpl_vector_new(nima/2);
        pnodded_off = cpl_vector_get_data(*nodded_off);
        pabba_off = cpl_vector_get_data(abba_off);
        /* The image list to contain the nodded images */
        nodded = cpl_imagelist_new();
        for (i=0; i<nima/2; i++) {
            if ((pabba_off[0]-pabba_off[1])*(pabba_off[2*i]-pabba_off[2*i+1])
                    > 0) {
                /* ab sequence */
                pnodded_off[i] = pabba_off[2*i];
                tmp_ima = cpl_image_duplicate(cpl_imagelist_get(abba, 2*i));
                cpl_image_subtract(tmp_ima, cpl_imagelist_get(abba, 2*i+1)); 
                cpl_image_divide_scalar(tmp_ima, 2.0);
                cpl_imagelist_set(nodded, tmp_ima, i);
            } else {
                /* ba sequence */
                pnodded_off[i] = pabba_off[2*i+1];
                tmp_ima = cpl_image_duplicate(cpl_imagelist_get(abba, 2*i+1));
                cpl_image_subtract(tmp_ima, cpl_imagelist_get(abba, 2*i)); 
                cpl_image_divide_scalar(tmp_ima, 2.0);
                cpl_imagelist_set(nodded, tmp_ima, i);
            }
        }
    } else {
        cpl_msg_error(cpl_func, "Unsupported chopping mode");
        return NULL;
    }

    /* Subtract the first offset to the others */
    ref_off = cpl_vector_get(*nodded_off, 0);
    cpl_vector_subtract_scalar(*nodded_off, ref_off);
    return nodded;
}

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Correct the distortion
  @param    ilist       the images list to correct
  @param    arc         the arc file containing the polynomial
  @param    startrace   the startrace file containing the polynomial
  @return   the distortion corrected image list or NULL
 */
/*----------------------------------------------------------------------------*/
static cpl_imagelist * isaac_spc_jitter_distor(
        cpl_imagelist   *   ilist,
        const char      *   arc,
        const char      *   startrace) 
{
    cpl_polynomial      *   arc_poly;
    cpl_polynomial      *   sttr_poly;
    cpl_table           *   tab;
    cpl_size                power[2];
    cpl_vector          *   profile;
    cpl_imagelist       *   warped_list;
    cpl_image           *   warped;
    int                     i;
    
    /* Test entries */
    if (ilist == NULL) return NULL;
    if ((arc == NULL) && (startrace == NULL)) return NULL;

    /* Get the arc polynomial */
    arc_poly = cpl_polynomial_new(2);
    if (arc != NULL) {
        cpl_msg_info(cpl_func, "Get the arc distortion from the file");
        if ((tab = cpl_table_load(arc, 1, 0)) == NULL) {
            cpl_msg_error(cpl_func, "cannot load the arc table");
            cpl_polynomial_delete(arc_poly);
            return NULL;
        }
        for (i=0; i<cpl_table_get_nrow(tab); i++) {
            if (cpl_table_get_column_type(tab, "Degree_of_x") == CPL_TYPE_INT) {
                power[0] = cpl_table_get_int(tab, "Degree_of_x", i, NULL); 
                power[1] = cpl_table_get_int(tab, "Degree_of_y", i, NULL); 
            } else {
                power[0] = cpl_table_get_double(tab, "Degree_of_x", i, NULL); 
                power[1] = cpl_table_get_double(tab, "Degree_of_y", i, NULL); 
            }
            cpl_polynomial_set_coeff(arc_poly, power,
                    cpl_table_get_double(tab, "poly2d_coef", i, NULL));
        }
        cpl_table_delete(tab);
    } else {
        cpl_msg_info(cpl_func, "Use the ID polynomial for the arc dist");
        power[0] = 1;
        power[1] = 0;
        cpl_polynomial_set_coeff(arc_poly, power, 1.0);
    }
 
    /* Get the startrace polynomial */
    sttr_poly = cpl_polynomial_new(2);
    if (startrace != NULL) {
        cpl_msg_info(cpl_func, "Get the startrace distortion from the file");
        if ((tab = cpl_table_load(startrace, 1, 0)) == NULL) {
            cpl_msg_error(cpl_func, "cannot load the startrace table");
            cpl_polynomial_delete(arc_poly);
            cpl_polynomial_delete(sttr_poly);
            return NULL;
        }
        for (i=0; i<cpl_table_get_nrow(tab); i++) {
            
            if (cpl_table_get_column_type(tab, "Degree_of_x") == CPL_TYPE_INT) {
                power[0] = cpl_table_get_int(tab, "Degree_of_x", i, NULL); 
                power[1] = cpl_table_get_int(tab, "Degree_of_y", i, NULL); 
            } else {
                power[0] = cpl_table_get_double(tab, "Degree_of_x", i, NULL); 
                power[1] = cpl_table_get_double(tab, "Degree_of_y", i, NULL); 
            }
            cpl_polynomial_set_coeff(sttr_poly, power,
                    cpl_table_get_double(tab, "poly2d_coef", i, NULL));
        }
        cpl_table_delete(tab);
    } else {
        cpl_msg_info(cpl_func, "Use the ID polynomial for the startrace dist");
        power[0] = 0;
        power[1] = 1;
        cpl_polynomial_set_coeff(sttr_poly, power, 1.0);
    }
   
    /* Create the kernel */
    profile = cpl_vector_new(CPL_KERNEL_DEF_SAMPLES);
    cpl_vector_fill_kernel_profile(profile, CPL_KERNEL_DEFAULT,
            CPL_KERNEL_DEF_WIDTH);

    /* Correct the images */
    warped_list = cpl_imagelist_new();
    for (i=0; i<cpl_imagelist_get_size(ilist); i++) {
        warped = cpl_image_duplicate(cpl_imagelist_get(ilist, i));
        if (cpl_image_warp_polynomial(warped, cpl_imagelist_get(ilist, i), 
                    arc_poly, sttr_poly, profile, CPL_KERNEL_DEF_WIDTH, profile,
                    CPL_KERNEL_DEF_WIDTH) != CPL_ERROR_NONE) {
            cpl_msg_error(cpl_func, "cannot correct the distortion");
            cpl_image_delete(warped);
            cpl_polynomial_delete(arc_poly);
            cpl_polynomial_delete(sttr_poly);
            cpl_vector_delete(profile);
            return NULL;
        }
        cpl_imagelist_set(warped_list, warped, i);
    }
    cpl_vector_delete(profile);
    cpl_polynomial_delete(arc_poly);
    cpl_polynomial_delete(sttr_poly);
    return warped_list;
}

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Refine the offsets
  @param    ima1    the first image
  @param    ima2    the second image
  @return   the computed offset or 10000.0 in error case
 */
/*----------------------------------------------------------------------------*/
static double isaac_spc_jitter_refine_offset(const cpl_image   *   ima1,
                                             const cpl_image   *   ima2)
{
    double                  pos1, pos2;

    /* Test entries */
    if (ima1 == NULL) return 10000.0;
    if (ima2 == NULL) return 10000.0;
    
    /* Detect the spectra */
    if (irplib_spectrum_find_brightest(ima1, 0.0, NO_SHADOW, 0.0, 0, 
                &pos1) == -1){
        return 10000.0;
    }
    if (irplib_spectrum_find_brightest(ima2, 0.0, NO_SHADOW, 0.0, 0,
                &pos2) == -1){
        return 10000.0;
    }
    return pos1-pos2;
}

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    The spectrum extraction
  @param    combined    the combined image
  @return   the extracted table or NULL on error
 */
/*----------------------------------------------------------------------------*/
static cpl_table * isaac_spc_jitter_extract(const cpl_image * combined)
{
    int                     lo_dist, hi_dist, lo_width, hi_width, spec_pos;
    int                     nx, ny;
    double                  pos;
    int                     low_side, up_side;
    int                     sky_pos[4];
    cpl_vector          *   sky;
    cpl_vector          *   spec;
    cpl_vector          *   wl;
    cpl_image           *   colfluximg;
    double              *   pspec;
    double              *   psky;
    double              *   pwl;
    cpl_table           *   out;
    cpl_bivector        *   toplot;
    int                     i;

    /* Test entries */
    if (combined == NULL) return NULL;

    /* Initialise */
    nx = cpl_image_get_size_x(combined);
    ny = cpl_image_get_size_y(combined);
    lo_dist = isaac_spc_jitter_config.extr_sky_lo_dist;
    hi_dist = isaac_spc_jitter_config.extr_sky_hi_dist;
    lo_width = isaac_spc_jitter_config.extr_sky_lo_width;
    hi_width = isaac_spc_jitter_config.extr_sky_hi_width;
    spec_pos = isaac_spc_jitter_config.extr_spec_pos;

    /* Detect the spectrum position if not passed */
    if (spec_pos < 0) {
        const int nthrow = cpl_vector_get_size(isaac_spc_jitter_config.throws);

        if (nthrow < 1) {
            cpl_msg_error(cpl_func, 
                    "Need a throw value to detect the spectra !!");
            cpl_ensure(0, CPL_ERROR_DATA_NOT_FOUND, NULL);
        }
        
        for (i=0; i < nthrow; i++){
            const int throw = (int)cpl_vector_get(isaac_spc_jitter_config.throws,
                                                  i);
            if (irplib_spectrum_find_brightest(combined, throw, TWO_SHADOWS,
                                               0.0, 0, &pos) == 0) break;
            if (irplib_spectrum_find_brightest(combined, throw, ONE_SHADOW,
                                               0.0, 0, &pos) == 0) break;
        }
        if (i == nthrow) {
            cpl_msg_error(cpl_func, "Could not detect the spectrum using %d "
                          "throw(s):", i);
#if 0
            cpl_vector_dump(isaac_spc_jitter_config.throws, stderr);
            cpl_plot_image("", "", "", combined); 
#endif
            cpl_ensure(0, CPL_ERROR_DATA_NOT_FOUND, NULL);
        }
        spec_pos = (int)pos;
        cpl_msg_info(cpl_func, "Spectrum detected at y = %d", spec_pos);
    }

    /* Set the parameters for the extraction */

    /* Spectrum position */
    low_side = spec_pos - (int)(isaac_spc_jitter_config.extr_spec_width/2);
    up_side = low_side + isaac_spc_jitter_config.extr_spec_width;
    if (up_side < low_side) {
        (void)cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT,
                                    "Spectrum zone invalid: low=%d > up=%d",
                                    low_side, up_side);
        return NULL;
    }
    if (low_side < 1) {
        (void)cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT,
                                    "Spectrum zone lower limit too low: "
                                    "%d < 1", low_side);
        return NULL;
    }
    if (up_side > ny) {
        (void)cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT,
                                    "Spectrum zone upper limit too high: "
                                    "%d > ny=%d", up_side, ny);
        return NULL;
    }

    /* Form a 1D single-row-image comprising the column fluxes of the zone */
    colfluximg = cpl_image_collapse_window_create(combined, 1, low_side,
                                                  nx, up_side, 0);

    if (cpl_image_count_rejected(colfluximg) == up_side - low_side + 1) {
        (void)cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                                    "Combined image has undefined flux "
                                    "from column %d to %d", low_side, up_side);
        cpl_image_delete(colfluximg);
        return NULL;
    }

    /* Residual Sky position */
    if (lo_dist < 0) lo_dist = 2*isaac_spc_jitter_config.extr_spec_width;
    if (hi_dist < 0) hi_dist = 2*isaac_spc_jitter_config.extr_spec_width;
    sky_pos[1] = spec_pos - lo_dist;
    sky_pos[0] = sky_pos[1] - lo_width;
    sky_pos[2] = spec_pos + hi_dist;
    sky_pos[3] = sky_pos[2] + hi_width;

    /* Get the sky */
    sky = cpl_vector_new(nx);
    psky = cpl_vector_get_data(sky);
    if (((sky_pos[0] < 1) || (lo_width == 0)) && 
            ((sky_pos[3] <= ny) && (hi_width > 0))) {
        for (i=0; i<nx; i++) {
            psky[i] = cpl_image_get_median_window(combined, i+1,
                    sky_pos[2], i+1, sky_pos[3]);
        }
    } else if (((sky_pos[3] > ny) || (hi_width == 0))
            && ((sky_pos[0] > 0) && (lo_width > 0))) {
        for (i=0; i<nx; i++) {
            psky[i] = cpl_image_get_median_window(combined, i+1,
                    sky_pos[0], i+1, sky_pos[1]);
        }
    } else if ((lo_width != 0) && (hi_width != 0)
            && (sky_pos[0] > 0) && (sky_pos[3] <= ny)) {
        for (i=0; i<nx; i++) {
            psky[i] = cpl_image_get_median_window(combined, i+1,
                    sky_pos[2], i+1, sky_pos[3]);
            psky[i] += cpl_image_get_median_window(combined, i+1,
                    sky_pos[0], i+1, sky_pos[1]);
            psky[i] /= 2.0;
        }
    } else {
        for (i=0; i<nx; i++) psky[i] = 0.0;
    }

    /* Estimate the spectrum */
    spec = cpl_vector_new(nx);
    pspec = cpl_vector_get_data(spec);
    for (i=0; i<nx; i++) {
        int is_bad = 0;
        const double colflux = cpl_image_get(colfluximg, i+1, 1, &is_bad);

        if (is_bad) {
            pspec[i] = 0.0;
            cpl_msg_warning(cpl_func, "Combined image has undefined flux in "
                            "column %d", i+1);
        } else {
            pspec[i] = colflux
                - psky[i] * isaac_spc_jitter_config.extr_spec_width;
        }
    } 
    cpl_image_delete(colfluximg);

    /* Get the wavelength */
    wl = cpl_vector_new(nx);
    pwl = cpl_vector_get_data(wl);
    for (i=0; i<nx; i++) {
        pwl[i] = isaac_spc_jitter_config.wavecal_a0 +
            isaac_spc_jitter_config.wavecal_a1 * (i+1) +
            isaac_spc_jitter_config.wavecal_a2 * (i+1) * (i+1) +
            isaac_spc_jitter_config.wavecal_a3 * (i+1) * (i+1) * (i+1);
    }

    /* Plot the spectrum if requested */
    if (isaac_spc_jitter_config.plot > 0) {
        toplot = cpl_bivector_wrap_vectors(wl, spec);
        cpl_plot_bivector(NULL, "t 'Spectrum' w lines", NULL, toplot);
        cpl_bivector_unwrap_vectors(toplot);
        toplot = cpl_bivector_wrap_vectors(wl, sky);
        cpl_plot_bivector(NULL, "t 'Sky' w lines", NULL, toplot);
        cpl_bivector_unwrap_vectors(toplot);
    }
    
    /* Create and fill the output table */
    out = cpl_table_new(nx);
    cpl_table_new_column(out, "X_coordinate", CPL_TYPE_DOUBLE);
    cpl_table_new_column(out, "Extracted_spectrum_value", CPL_TYPE_DOUBLE);
    cpl_table_new_column(out, "Sky_spectrum", CPL_TYPE_DOUBLE);
    for (i=0; i<nx; i++) {
        cpl_table_set_double(out, "X_coordinate", i, pwl[i]);
        cpl_table_set_double(out, "Extracted_spectrum_value", i, pspec[i]);
        cpl_table_set_double(out, "Sky_spectrum", i, psky[i]);
    }
    cpl_vector_delete(wl);
    cpl_vector_delete(spec);
    cpl_vector_delete(sky);
    return out;
}

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Compute the flux calibration
  @param    seds_file   the file with the SEDs
  @param    stdstars    the file with the standard stars
  @param    frame       the ref frame for header infos
  @param    tab         the table with the extracted spectrum
  @param    mag         the standard star magnitude
  @return   0 iff everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_spc_jitter_std(const char      * seds_file,
                                const char      * stdstars,
                                const cpl_frame * frame,
                                cpl_table       * tab,
                                double          mag)
{
    double                  surface = CPL_MATH_PI * 400 * 400;
    cpl_propertylist    *   plist;
    double                  dit, ra, dec, magnitude, cent_wl;
    isaac_band              band;
    const int               nelem = cpl_table_get_nrow(tab);
    cpl_vector          *   wl;
    cpl_vector          *   spec;
    cpl_bivector        *   spec_biv;
    cpl_bivector        *   sed;
    cpl_vector          *   efficiency;
    cpl_vector          *   mag_zero;
    cpl_vector          *   conversion;
    int                     i;

    /* Test entries */
    if (seds_file == NULL) return CPL_ERROR_UNSPECIFIED;
    if (stdstars == NULL) return CPL_ERROR_UNSPECIFIED;
    if (frame == NULL) return CPL_ERROR_UNSPECIFIED;
    if (tab == NULL) return CPL_ERROR_UNSPECIFIED;
    if (cpl_error_get_code()) return CPL_ERROR_UNSPECIFIED;

    /* Get the propertylist */
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(frame),0))==NULL) {
        cpl_msg_error(cpl_func, "Cannot load the property list");
        return CPL_ERROR_UNSPECIFIED;
    }
    
    /* Get DIT */
    dit = isaac_pfits_get_dit(plist);
    if (cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "Cannot get DIT");
        cpl_propertylist_delete(plist); 
        return CPL_ERROR_UNSPECIFIED;
    }
    cpl_msg_info(cpl_func, "DIT:   %g", dit);
    
    /* Get band */
    if ((band = isaac_get_associated_filter(isaac_spc_jitter_config.filter)) 
            == ISAAC_BAND_UNKNOWN) {
        cpl_msg_error(cpl_func, "Cannot associate the filter to a BB one");
        cpl_propertylist_delete(plist); 
        return CPL_ERROR_UNSPECIFIED;
    }
    cpl_msg_info(cpl_func, "Band : %s", isaac_std_band_name(band));

    /* Get RA / DEC */
    ra = isaac_pfits_get_ra(plist);
    dec = isaac_pfits_get_dec(plist);
    if (cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "Cannot get star position");
        cpl_propertylist_delete(plist); 
        return CPL_ERROR_UNSPECIFIED;
    }
    cpl_msg_info(cpl_func, "RA:  %g  DEC:  %g", ra, dec);

    /* Get magnitude */
    if (mag < 0) {
        /* Reference star */
        if (irplib_stdstar_find_star(stdstars, ra, dec,
        isaac_std_band_name(isaac_get_bbfilter(isaac_spc_jitter_config.filter)),
                                     "all", &magnitude,
                                     &isaac_spc_jitter_config.starname,
                                     &isaac_spc_jitter_config.sptype,
                                     NULL, NULL, NULL, 2.0)) {
            cpl_msg_error(cpl_func, "Cannot find the star in catalogs");
            cpl_propertylist_delete(plist); 
            return CPL_ERROR_UNSPECIFIED;
        }
    } else { 
        magnitude = mag;
    }
    cpl_propertylist_delete(plist); 
    (void)strncpy(isaac_spc_jitter_config.filter_ref, 
                  isaac_std_band_name(isaac_get_bbfilter
                                      (isaac_spc_jitter_config.filter)),
                  KEYSIZE-1);
    cpl_msg_info(cpl_func, "Magnitude: %g", magnitude);
    isaac_spc_jitter_config.std_magnitude = magnitude;
    
    /* Get band center */
    switch (band) {
        case ISAAC_BAND_Z:        cent_wl = CENT_WL_BAND_Z;  break;
        case ISAAC_BAND_SZ:       cent_wl = CENT_WL_BAND_SZ; break;
        case ISAAC_BAND_J:        cent_wl = CENT_WL_BAND_J;  break;
        case ISAAC_BAND_JBLOCK:   cent_wl = CENT_WL_BAND_J;  break;
        case ISAAC_BAND_SH:       cent_wl = CENT_WL_BAND_H;  break;
        case ISAAC_BAND_SK:       cent_wl = CENT_WL_BAND_K;  break;
        case ISAAC_BAND_SL:       cent_wl = CENT_WL_BAND_SL; break;
        case ISAAC_BAND_M:        cent_wl = CENT_WL_BAND_M;  break;
        default:
            cpl_msg_error(cpl_func, "Unsupported band: %s", 
                    isaac_std_band_name(band));
            return CPL_ERROR_UNSPECIFIED;
    }
    cpl_msg_info(cpl_func, "Center of band : %g microns", cent_wl);
   
    /* Get the SED */
    sed = irplib_stdstar_get_sed(seds_file, isaac_spc_jitter_config.sptype);
    if (sed == NULL) {
        cpl_msg_error(cpl_func, "Cannot get the SED");
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Put the spectrum in a bivector */
    wl = cpl_vector_wrap(nelem, cpl_table_get_data_double(tab,
                "X_coordinate"));
    spec = cpl_vector_wrap(nelem, cpl_table_get_data_double(tab,
                "Extracted_spectrum_value"));
    spec_biv = cpl_bivector_wrap_vectors(wl, spec);

    /* Get the conversion */
    if ((conversion = irplib_stdstar_get_conversion(spec_biv, dit, surface,
                    4.5, magnitude)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot get the conversion");
        cpl_bivector_unwrap_vectors(spec_biv);
        cpl_vector_unwrap(spec);
        cpl_vector_unwrap(wl);
        cpl_bivector_delete(sed);
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Get the 0 magnitude spectrum */
    if ((mag_zero = irplib_stdstar_get_mag_zero(sed, 
                    cpl_bivector_get_x(spec_biv), cent_wl)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot get the 0 magnitude spectrum");
        cpl_bivector_unwrap_vectors(spec_biv);
        cpl_vector_unwrap(spec);
        cpl_vector_unwrap(wl);
        cpl_bivector_delete(sed);
        cpl_vector_delete(conversion);
        return CPL_ERROR_UNSPECIFIED;
    }
    cpl_bivector_unwrap_vectors(spec_biv);
    cpl_vector_unwrap(spec);
    cpl_vector_unwrap(wl);
    cpl_bivector_delete(sed);

    /* Set the mag_zero 0 values to 1 */
    for (i=0; i<cpl_vector_get_size(mag_zero); i++) {
        if (cpl_vector_get(mag_zero, i) < 1e-19) {
            cpl_vector_set(mag_zero, i, 1.0);
        }
    }

    /* Get the efficiency */
    efficiency = cpl_vector_duplicate(conversion);
    if (cpl_vector_divide(efficiency, mag_zero) != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "Missing SED info in this wavelength range");
        cpl_vector_delete(conversion);
        cpl_vector_delete(mag_zero);
        cpl_vector_delete(efficiency);
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Set the efficiency to 0 where the mag_zero is 1 */
    for (i=0; i<cpl_vector_get_size(mag_zero); i++) {
        if (cpl_vector_get(mag_zero, i) == 1.0) {
            cpl_vector_set(efficiency, i, 0.0);
        }
    }
    cpl_vector_delete(mag_zero);

    /* Create extra columns for the results */
    cpl_table_wrap_double(tab, cpl_vector_get_data(efficiency), "Efficiency");
    cpl_table_wrap_double(tab, cpl_vector_get_data(conversion), "Conversion");
    cpl_vector_unwrap(efficiency);
    cpl_vector_unwrap(conversion);

    /* Plot the spectra if requested */
    if (isaac_spc_jitter_config.plot > 0) {
        wl = cpl_vector_wrap(nelem, cpl_table_get_data_double(tab,
                    "X_coordinate"));
        spec = cpl_vector_wrap(nelem, cpl_table_get_data_double(tab,
                    "Conversion"));
        spec_biv = cpl_bivector_wrap_vectors(wl, spec);
        cpl_plot_bivector(NULL, "t 'Conversion' w lines", 
                NULL, spec_biv);
        cpl_bivector_unwrap_vectors(spec_biv);
        cpl_vector_unwrap(spec);
        spec = cpl_vector_wrap(nelem, cpl_table_get_data_double(tab,
                    "Efficiency"));
        spec_biv = cpl_bivector_wrap_vectors(wl, spec);
        cpl_plot_bivector(NULL, "t 'Efficiency' w lines", NULL, spec_biv);
        cpl_bivector_unwrap_vectors(spec_biv);
        cpl_vector_unwrap(spec);
        cpl_vector_unwrap(wl);
    }
    
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Save the recipe products on disk
  @param    ima     image to save
  @param    contrib Contribution map to save
  @param    tab     table to save
  @param    parlist the input list of parameters
  @param    set     the input frame set
  @return   CPL_ERROR_NONE iff everything is ok
 */
/*----------------------------------------------------------------------------*/
static
cpl_error_code isaac_spc_jitter_save(cpl_frameset            * set,
                                     const cpl_image         * ima,
                                     const cpl_image         * contrib,
                                     const cpl_table         * tab,
                                     const cpl_parameterlist * parlist)
{
    /* Get the QC params in qclist */
    cpl_propertylist    *   qclist = cpl_propertylist_new();
    cpl_propertylist    *   paflist = NULL;
    const cpl_frame     *   ref_frame;
    char                *   qc_str = NULL;
    const char          *   procat;
    const int               nintens
        = cpl_vector_get_size(isaac_spc_jitter_config.intensities);
    int                     i;
    cpl_propertylist * xtlist = cpl_propertylist_new();

    bug_if(cpl_propertylist_append_string(xtlist, "EXTNAME",
                                          "Contribution Map"));
    bug_if(contrib == NULL);

    /* Add QC parameters */
    /* FIXME: Fill this list where the numbers are available */
    if (isaac_spc_jitter_config.filter[0] != (char)0) 
        bug_if(cpl_propertylist_append_string(qclist, "ESO QC FILTER OBS",
                                              isaac_spc_jitter_config.filter));

    bug_if(cpl_propertylist_append_double(qclist, "ESO QC DISPCO1",
                                          isaac_spc_jitter_config.wavecal_a0));
    cpl_propertylist_append_double(qclist, "ESO QC DISPCO2",
                                   isaac_spc_jitter_config.wavecal_a1);
    cpl_propertylist_append_double(qclist, "ESO QC DISPCO3",
                                   isaac_spc_jitter_config.wavecal_a2);
    cpl_propertylist_append_double(qclist, "ESO QC DISPCO4",
                                   isaac_spc_jitter_config.wavecal_a3);
    cpl_propertylist_append_double(qclist, "ESO QC WLEN",
                                   isaac_spc_jitter_config.wavecal_a0 +
                                   isaac_spc_jitter_config.wavecal_a1 * 512 +
                                   isaac_spc_jitter_config.wavecal_a2 * 512 *
                                   512 +
                                   isaac_spc_jitter_config.wavecal_a3 * 512 *
                                   512 * 512);
    bug_if(cpl_propertylist_append_double(qclist, "ESO QC DISP XCORR",
                                          isaac_spc_jitter_config.wavecal_cc));
    if (isaac_spc_jitter_config.wavecal_out == 0) {
        cpl_propertylist_append_string(qclist, "ESO QC WLMETHOD",
                "physical model");
    } else if (isaac_spc_jitter_config.wavecal_out == 1) {
        cpl_propertylist_append_string(qclist, "ESO QC WLMETHOD",
                "sky lines");
    } else if (isaac_spc_jitter_config.wavecal_out == 2) {
        cpl_propertylist_append_string(qclist, "ESO QC WLMETHOD",
                "arc file");
    }
    for (i = 0; i < nintens; i++) {
        const double intens
            = cpl_vector_get(isaac_spc_jitter_config.intensities, i);

        qc_str = cpl_sprintf("ESO QC SPEC INTENS%d", i+1);
        cpl_propertylist_append_double(qclist, qc_str, intens);
        cpl_free(qc_str);
        qc_str = NULL;
    }
    
    /* In std mode */
    if (isaac_spc_jitter_config.std_mode == 1) {
        bug_if(cpl_propertylist_append_string(qclist, "ESO QC STDNAME",
                                              isaac_spc_jitter_config.starname ?
                                              isaac_spc_jitter_config.starname :
                                              ""));
        bug_if(cpl_propertylist_append_string(qclist, "ESO QC SPECTYPE",
                                              isaac_spc_jitter_config.sptype ?
                                              isaac_spc_jitter_config.sptype :
                                              ""));
        cpl_propertylist_append_double(qclist, "ESO QC STARMAG",
                                       isaac_spc_jitter_config.std_magnitude);
        if (isaac_spc_jitter_config.filter_ref[0] != (char)0) 
            cpl_propertylist_append_string(qclist, "ESO QC FILTER REF",
                                           isaac_spc_jitter_config.filter_ref);
    }

    /* Change WCS keywords to the computed wavelength solution */
    cpl_propertylist_update_double(qclist, "CRVAL1",
                                   isaac_spc_jitter_config.wavecal_a0);
    cpl_propertylist_update_double(qclist, "CRVAL2", 1.0);
    cpl_propertylist_update_double(qclist, "CRPIX1", 1.0);
    cpl_propertylist_update_double(qclist, "CRPIX2", 1.0);
    cpl_propertylist_update_double(qclist, "CDELT1", 
                                   isaac_spc_jitter_config.wavecal_a1);
    cpl_propertylist_update_double(qclist, "CDELT2", 1.0);
    cpl_propertylist_update_string(qclist, "CTYPE1", "LINEAR");
    cpl_propertylist_update_string(qclist, "CTYPE2", "LINEAR");
    if (cpl_propertylist_has(qclist, "CD1_1")) {
        cpl_propertylist_update_double(qclist, "CD1_1",
                                       isaac_spc_jitter_config.wavecal_a1);
    } else {
        cpl_propertylist_insert_after_double(qclist, "CTYPE2", "CD1_1",
                                             isaac_spc_jitter_config.wavecal_a1);
    }
    if (cpl_propertylist_has(qclist, "CD2_2")) {
        cpl_propertylist_update_double(qclist, "CD2_2", 1.0);
    } else {
        cpl_propertylist_insert_after_double(qclist, "CD1_1", "CD2_2", 1.0);
    }

    /* Write the image  */
    procat = isaac_spc_jitter_config.std_mode == 1 ?
        ISAAC_SPC_JITTER_COMB_STD : ISAAC_SPC_JITTER_COMB;

    irplib_dfs_save_image(set, parlist, set, ima, CPL_BPP_IEEE_FLOAT,
                          RECIPE_STRING, procat, qclist, NULL,
                          PACKAGE "/" PACKAGE_VERSION,
                          RECIPE_STRING "_combined" CPL_DFS_FITS);

    /* Append the contribution map */
    skip_if (cpl_image_save(contrib, RECIPE_STRING "_combined" CPL_DFS_FITS,
                            CPL_BPP_16_UNSIGNED, xtlist, CPL_IO_EXTEND));

    /* Write the table  */
    if (tab != NULL) {
        const char * procatt = isaac_spc_jitter_config.std_mode == 1 ?
            ISAAC_SPC_JITTER_EXTR_STD : ISAAC_SPC_JITTER_EXTR;

        irplib_dfs_save_table(set, parlist, set, tab, NULL, RECIPE_STRING,
                              procatt, qclist, NULL,
                              PACKAGE "/" PACKAGE_VERSION,
                              RECIPE_STRING "_extracted"  CPL_DFS_FITS);
    }

    /* Get the reference frame */
    ref_frame = irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW);

    /* Get the keywords for the paf file from the reference file */
    paflist = cpl_propertylist_load_regexp(cpl_frame_get_filename(ref_frame),
                                           0, "^(ARCFILE|MJD-OBS|INSTRUME"
                                           "|ESO TPL ID|ESO TPL NEXP"
                                           "|ESO DPR CATG|ESO DPR TECH"
                                           "|ESO DPR TYPE|DATE-OBS"
                                           "|ESO INS GRAT NAME"
                                           "|ESO INS GRAT WLEN"
                                           "|ESO INS OPTI1 ID|ESO OBS ID"
                                           "|ESO OBS TARG NAME)$", 0);
    skip_if (paflist == NULL);

    /* Copy the QC in paflist */
    bug_if(cpl_propertylist_append(paflist, qclist));
    cpl_propertylist_empty(qclist);

    /* PRO.CATG */
    cpl_propertylist_update_string(paflist, CPL_DFS_PRO_CATG, procat);

    /* Save the PAF file */
    skip_if(cpl_dfs_save_paf("ISAAC", RECIPE_STRING, paflist,
                             RECIPE_STRING CPL_DFS_PAF));

    end_skip;

    cpl_propertylist_delete(xtlist);
    cpl_propertylist_delete(paflist);
    cpl_propertylist_delete(qclist);
    cpl_free(qc_str);

    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    compares two doubles in regard to a given threshold
  @param    off1 first double
  @param    off2 second double
  @param    thresh  threshold
  @return   1 if threshold is between offsets, 0 otherwise
 */
/*----------------------------------------------------------------------------*/
static int off_comp(double off1, double off2, double thresh)
{
    return (off1 > thresh && off2 < thresh) || (off1 < thresh && off2 > thresh)
        ? 1 : 0;
}

