/* $Id: isaac_util_genlines.c,v 1.9 2013-03-12 08:06:48 llundin Exp $
 *
 * This file is part of the ISAAC Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2013-03-12 08:06:48 $
 * $Revision: 1.9 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>

#include "irplib_utils.h"

#include "isaac_utils.h"
#include "isaac_pfits.h"
#include "isaac_dfs.h"

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int isaac_util_genlines_create(cpl_plugin *);
static int isaac_util_genlines_exec(cpl_plugin *);
static int isaac_util_genlines_destroy(cpl_plugin *);
static int isaac_util_genlines(cpl_parameterlist *, cpl_frameset *);
static int isaac_util_genlines_save(cpl_table *, cpl_parameterlist *, 
        cpl_frameset *);

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs */
    int             fill_blanks; 
    int             display;
    int             mode;
    double          wl_factor;
    /* Outputs */
} isaac_util_genlines_config;

static char isaac_util_genlines_description[] =
"This recipe is used to generate spectrum calibration tables.\n"
"The sof file contains the names of the input ASCII file\n"
"tagged with "ISAAC_UTIL_GENLINES_RAW".\n"
"The ASCII file must contain two columns:\n"
"1st: Wavelengths in increasing order (the unit is corrected by\n"
"     the factor option to obtain nanometers).\n"
"2nd: The atmospheric emission.\n"
"The ASCII files are in the catalogs/ directory of the ISAAC distribution.\n"
"This recipe produces 1 file:\n"
"First product:     the table with the lines.\n"
"                   (PRO CATG = "ISAAC_UTIL_GENLINES_OH_CAT") or\n"
"                   (PRO CATG = "ISAAC_UTIL_GENLINES_XE_CAT") or\n"
"                   (PRO CATG = "ISAAC_UTIL_GENLINES_AR_CAT")\n";

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module. 
  @param    list    the plugin list
  @return   0 iff everything is ok

  This function is exported.
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist * list)
{
    cpl_recipe  *   recipe = cpl_calloc(1, sizeof(*recipe));
    cpl_plugin  *   plugin = &recipe->interface;

    cpl_plugin_init(plugin,
                    CPL_PLUGIN_API,
                    ISAAC_BINARY_VERSION,
                    CPL_PLUGIN_TYPE_RECIPE,
                    "isaac_util_genlines",
                    "Generate spectrum calibration FITS tables",
                    isaac_util_genlines_description,
                    "Lars Lundin",
                    PACKAGE_BUGREPORT,
                    isaac_get_license(),
                    isaac_util_genlines_create,
                    isaac_util_genlines_exec,
                    isaac_util_genlines_destroy);

    cpl_pluginlist_append(list, plugin);
    
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options    
  @param    plugin  the plugin
  @return   0 iff everything is ok

  Create the recipe instance and make it available to the application using the 
  interface. 
 */
/*----------------------------------------------------------------------------*/
static int isaac_util_genlines_create(cpl_plugin * plugin)
{
    cpl_recipe      * recipe;
    cpl_parameter   * p;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    /* Create the parameters list in the cpl_recipe object */
    recipe->parameters = cpl_parameterlist_new();

    /* Fill the parameters list */
    /* --fill_blanks */
    p = cpl_parameter_new_value("isaac.isaac_util_genlines.fill_blanks",
            CPL_TYPE_BOOL, "flag to fill blanks", "isaac.isaac_util_genlines",
            FALSE);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "fill_blanks");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);
    /* --display */
    p = cpl_parameter_new_value("isaac.isaac_util_genlines.display",
            CPL_TYPE_BOOL, "flag to plot", "isaac.isaac_util_genlines",
            FALSE);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "display");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);
    /* --mode */
    p = cpl_parameter_new_value("isaac.isaac_util_genlines.mode",
            CPL_TYPE_INT, "1-OH, 2-XE, 3-AR", 
            "isaac.isaac_util_genlines", 1);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "mode");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);
    /* --wl_factor */
    p = cpl_parameter_new_value("isaac.isaac_util_genlines.wl_factor",
            CPL_TYPE_DOUBLE, "The factor used to multiply the wl", 
            "isaac.isaac_util_genlines", 1.0);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "wl_factor");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);
 
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 iff everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_util_genlines_exec(cpl_plugin * plugin)
{
    cpl_recipe  *   recipe;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    return isaac_util_genlines(recipe->parameters, recipe->frames);
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 iff everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_util_genlines_destroy(cpl_plugin * plugin)
{
    cpl_recipe  *   recipe;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    cpl_parameterlist_delete(recipe->parameters);
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The FITS file creation occurs here 
  @param    framelist   the frames list
  @return   0 iff everything is ok

  The recipe expects a text file with two columns, and will create a FITS file
  out of it. 
 */
/*----------------------------------------------------------------------------*/
static int isaac_util_genlines(
        cpl_parameterlist   *   parlist,
        cpl_frameset        *   framelist)
{
    cpl_bivector    *   bivec;
    double          *   pbivec_x;
    double          *   pbivec_y;
    cpl_bivector    *   bivec_fill;
    double          *   pbivec_fill_x;
    double          *   pbivec_fill_y;
    cpl_frame       *   cur_frame;
    int                 nvals, nb_new_vals;
    double              wavel;
    cpl_table       *   tab;
    cpl_parameter   *   par;
    int                 i;

    /* Initialise */
    par = NULL;
    
    /* Retrieve input parameters */
    /* Fill blanks */
    par = cpl_parameterlist_find(parlist, 
            "isaac.isaac_util_genlines.fill_blanks");
    isaac_util_genlines_config.fill_blanks = cpl_parameter_get_bool(par);
    /* Fill blanks */
    par = cpl_parameterlist_find(parlist,"isaac.isaac_util_genlines.display");
    isaac_util_genlines_config.display = cpl_parameter_get_bool(par);
    /* Mode */
    par = cpl_parameterlist_find(parlist,"isaac.isaac_util_genlines.mode");
    isaac_util_genlines_config.mode = cpl_parameter_get_int(par);
    /* Wavelength factor */
    par=cpl_parameterlist_find(parlist,"isaac.isaac_util_genlines.wl_factor");
    isaac_util_genlines_config.wl_factor = cpl_parameter_get_double(par);
   
    /* Identify the RAW and CALIB frames in the input frameset */
    if (isaac_dfs_set_groups(framelist)) {
        cpl_msg_error(cpl_func, "Cannot identify RAW and CALIB frames");
        return CPL_ERROR_UNSPECIFIED;
    }
   
    /* Load the file */
    cur_frame = cpl_frameset_get_position(framelist, 0);
    if ((bivec=cpl_bivector_read(cpl_frame_get_filename(cur_frame)))==NULL) {
        cpl_msg_error(cpl_func, "Cannot load the file in the bivector");
        return CPL_ERROR_UNSPECIFIED;
    }
    nvals = cpl_bivector_get_size(bivec);
    
     /* If fill_blanks is requested */
    if (isaac_util_genlines_config.fill_blanks) {
        nb_new_vals = 3 * nvals;
        bivec_fill = cpl_bivector_new(nb_new_vals);
        pbivec_fill_x = cpl_bivector_get_x_data(bivec_fill);
        pbivec_fill_y = cpl_bivector_get_y_data(bivec_fill);
        pbivec_x = cpl_bivector_get_x_data(bivec);
        pbivec_y = cpl_bivector_get_y_data(bivec);
        for (i=0; i<nvals; i++) {
            wavel = pbivec_x[i] * isaac_util_genlines_config.wl_factor;
            pbivec_fill_x[3*i] = wavel - 0.01;
            pbivec_fill_y[3*i] = 0.0;
            pbivec_fill_x[3*i+1] = wavel;
            pbivec_fill_y[3*i+1] = pbivec_y[i];
            pbivec_fill_x[3*i+2] = wavel + 0.01;
            pbivec_fill_y[3*i+2] = 0.0;
        }
        cpl_bivector_delete(bivec);
        bivec = bivec_fill;
        bivec_fill = NULL;
        nvals = cpl_bivector_get_size(bivec);
    } else {
        cpl_vector_multiply_scalar(cpl_bivector_get_x(bivec),
                isaac_util_genlines_config.wl_factor);
    }

    /* Display if requested */
    if (isaac_util_genlines_config.display) {
        cpl_plot_bivector(
                "set grid;set xlabel 'Wavelength (A)';set ylabel 'Emission';",
                "t 'Catalog lines' w lines", "", bivec);
    }
    
    /* Allocate the data container */
    tab = cpl_table_new(nvals);
    cpl_table_wrap_double(tab, cpl_bivector_get_x_data(bivec),
            ISAAC_COL_WAVELENGTH);
    cpl_table_wrap_double(tab, cpl_bivector_get_y_data(bivec), 
            ISAAC_COL_EMISSION);
   
    /* Save the table */
    cpl_msg_info(cpl_func, "Saving the table with %d rows", nvals);
    if (isaac_util_genlines_save(tab, parlist, framelist) == -1) {
        cpl_msg_error(cpl_func, "Cannot write the table");
        cpl_bivector_delete(bivec);
        cpl_table_unwrap(tab, ISAAC_COL_WAVELENGTH);
        cpl_table_unwrap(tab, ISAAC_COL_EMISSION);
        cpl_table_delete(tab);
        return CPL_ERROR_UNSPECIFIED;
    }
    cpl_bivector_delete(bivec);
    cpl_table_unwrap(tab, ISAAC_COL_WAVELENGTH);
    cpl_table_unwrap(tab, ISAAC_COL_EMISSION);
    cpl_table_delete(tab);
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Save the product of the recipe
  @param    out_table   the table 
  @param    parlist     the input list of parameters
  @param    set         the input frame set
  @return   0 if everything is ok, -1 otherwise
 */
/*----------------------------------------------------------------------------*/
static int isaac_util_genlines_save(
        cpl_table           *   out_table,
        cpl_parameterlist   *   parlist,
        cpl_frameset        *   set)
{
    char                    name_o[512];
    cpl_propertylist    *   plist;
    cpl_frame           *   product_frame;

    /* Set the file name */
    sprintf(name_o, "isaac_util_genlines_save.fits");
    cpl_msg_info(cpl_func, "Writing %s" , name_o);

    /* Get FITS header from reference file */
    plist = cpl_propertylist_new();
    cpl_propertylist_append_string(plist, "INSTRUME", "ISAAC");
            
    /* Create product frame */
    product_frame = cpl_frame_new();
    cpl_frame_set_filename(product_frame, name_o);
    if (isaac_util_genlines_config.mode == 1) 
        cpl_frame_set_tag(product_frame, ISAAC_UTIL_GENLINES_OH_CAT);
    else if (isaac_util_genlines_config.mode == 2)
        cpl_frame_set_tag(product_frame, ISAAC_UTIL_GENLINES_XE_CAT);
    else if (isaac_util_genlines_config.mode == 3)
        cpl_frame_set_tag(product_frame, ISAAC_UTIL_GENLINES_AR_CAT);
    else 
        cpl_frame_set_tag(product_frame, "UNKNOWN");

    cpl_frame_set_type(product_frame, CPL_FRAME_TYPE_TABLE);
    cpl_frame_set_group(product_frame, CPL_FRAME_GROUP_PRODUCT);
    cpl_frame_set_level(product_frame, CPL_FRAME_LEVEL_FINAL);

    /* Add DataFlow keywords */
    if (cpl_dfs_setup_product_header(plist, product_frame, set, parlist,
            "isaac_util_genlines", PACKAGE "/" PACKAGE_VERSION,
	    "PRO-1.15", NULL)!=CPL_ERROR_NONE) {
        cpl_msg_warning(cpl_func, "Problem in the product DFS-compliance");
        cpl_error_reset();
    }

    /* Save the file */
    if (cpl_table_save(out_table, plist, NULL, name_o,
                CPL_IO_DEFAULT) != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "Cannot save the product");
        cpl_frame_delete(product_frame);
        cpl_propertylist_delete(plist);
        return CPL_ERROR_UNSPECIFIED;
    }
    cpl_propertylist_delete(plist);

    /* Log the saved file in the input frameset */
    cpl_frameset_insert(set, product_frame);
    /* Return */
    return 0;
}
